; Assembles on Sinclair QL with Gunther Strube's Z80ASM cross-assembler

; Note, stack order is high/low downwards (bytes in words or words in doubles)

; keys:  *** = TREE SYSTEMS USA (60fps) changes to HUSBAND UK (50fps)
;        !!! = bug fix
;        .X  = unaddressed label

MODULE   forth
ORG      0

; if RAM/ROM at $2000 to $3FFF

defc     RHEAD = $2000        ; $A4 or $A5 in header if set up
defc     RBOOT = $2002        ; boot routine else NUL (return)
defc     RDICT = $2004        ; pointer latest dictionary entry
defc     RHERE = $2006        ; pointer end of dictionary
defc     RTSKL = $2008        ; task list link
defc     RST28 = $200A        ; RST $28 call point
defc     RST30 = $200D        ; RST $30 call point
defc     RRAM  = $2010        ; new start of RAM if no header set

; constant addresses

defc     SRAM  = $4000        ; start of RAM
defc     SRAMH = $40          ; start of RAM high address
defc     SPTOP = $FB80        ; stack pointer top
defc     PAD   = $FBC0        ; location of scratchpad
defc     CBUF  = $FCC0        ; location of CBUF character buffer
defc     SPIDT = $FC36        ; idle stack pointer top
defc     IDLE  = $FC3E        ; idle routine
defc     SYSI  = $FC40        ; system integrity byte
defc     ORGB  = $FC50        ; base address of origin parameter area
defc     PAGE  = $FC52        ; current block number
defc     ICTR  = $FC54        ; idle counter
defc     RAMS  = $FC56        ; size of RAM
defc     UCTR  = $FC58        ; user counter double
defc     TIME  = $FC5C        ; time counter double
defc     UINC  = $FC60        ; user increment double
defc     TINC  = $FC64        ; time increment double
defc     PER   = $FC68        ; period of TIME before wrap-around double
defc     RAMB  = $FC6C        ; bottom of RAM
defc     DISB  = $FC6E        ; base of display
defc     TSKS  = $FC70        ; start of linked list of tasks
defc     FNCE  = $FC72        ; dictionary boundary set by FENCE
defc     LPTR  = $FC74        ; printer routine
defc     BLKB  = $FC76        ; block base address
defc     TIBO  = $FC7E        ; TIB output index (FIFO)
defc     TIBI  = $FC7F        ; TIB input index (FIFO)
defc     SYSF  = $FC78        ; system set-up byte
defc     NMBR  = $FC7A        ; number parse routine
defc     BTSK  = $FC7C        ; background task
defc     NACT  = $FC80        ; next action when NMI counter zeroed
defc     DISR  = $FC82        ; display routine
defc     CBFP  = $FC84        ; character LIFO buffer pointer
defc     BASE  = $FC86        ; current number base
defc     DICT  = $FC88        ; pointer to latest dictionary entry
defc     HERE  = $FC8A        ; pointer to end of dictionary
defc     XDICT = $FC8C        ; pointer to accepted latest dictionary entry
defc     XHERE = $FC8E        ; pointer to accepted end of dictionary
defc     IYSP  = $FC90        ; IY stack pointer (empty)
defc     TSKE  = $FC92        ; end of linked list of tasks
defc     STSP  = $FC94        ; stack pointer store whilst idle
defc     IDSP  = $FC96        ; idle stack pointer store
defc     TSKI  = $FC98        ; internal task linkage
defc     SFLG  = $FCA5        ; system flags
defc     INPB  = $FCA6        ; process byte as input routine
defc     PUTB  = $FCA8        ; put character into TIB routine
defc     KEYD  = $FCAA        ; key debounce/auto-repeat
defc     KEYL  = $FCAB        ; last key read
defc     KEYN  = $FCAC        ; new key event
defc     DFLG  = $FCAD        ; device flags
defc     EDWP  = $FCAE        ; edit window parameters
defc     COWP  = $FCB6        ; command window parameters
defc     PFLG  = $FCBE        ; print flags
defc     CCNT  = $FCBF        ; count for string in CBUF
defc     DFRAM = $BD00        ; display file if 48K RAM
defc     DFILE = $FD00        ; display file (fixed 768 bytes)

; constant data
                              ;                                 *** 60fps
defc     LTOP  =    30        ; top raster lines (FAST & SLOW)         30
defc     LFST  =   255        ; FAST remaining raster lines           228
defc     LSLW  =    74        ; SLOW bottom raster lines               35
defc     FPS   = $0032        ; frames per second                   $003C
defc     FPM   = $0BB8        ; frames per minute                   $0E10
defc     FPHH  = $0002        ; frames per hour high word           $0003
defc     FPHL  = $BF20        ; frames per hour low word            $4BC0
defc     FPDH  = $0041        ; frames per day high word            $004F
defc     FPDL  = $EB00        ; frames per day low word             $1A00
defc     FPWH  = $01CD        ; frames per week high word           $0229
defc     FPWL  = $6D00        ; frames per week low word            $B600
defc     FPYH  = $5DFC        ; frames per year high word           $70C8
defc     FPYL  = $0F00        ; frames per year low word            $1200

defc     RSCN  =    22        ; M1 events offset to start of RH border
defc     SCAN  =    35        ; M1 events offset for whole scan line

;  With horizontal scan of 15625Hz and pixels every 6.5MHz a line consists
;  of 416 pixels (or 52 character columns 8 pixels wide). A 32 character
;  row therefore leaves 20 columns for left and right borders.
;
;  RSCN represents +12 on the horizontal column offset of the display file
;  within the screen. To match the original ZX81 placement make RSCN = 20.


; =============================== START OF ROM ================================

; RST $00 - power up, hardware reset (not necessarily same as COLD reset)

         out   ($FD),a        ; NMI generator off
         ld    sp,SPTOP       ; initial SP
         jp    L946           ; continuation

; RST $08 - push HL onto IY stack

.L8      dec   iy
         ld    (iy+$00),h
         jr    L4B
         rst   $38

; RST $10 - pop HL from IY stack

         ld    l,(iy+$00)
         inc   iy
         jr    L51
         rst   $38

; RST $18 - put character into CBUF

         push  hl
         push  af
         ld    hl,CBFP        ; current pointer
         jp    L201           ; continuation

; RST $20 - get character from CBUF

         push  hl
         ld    hl,CBFP        ; current pointer
         ld    a,(hl)         ; low byte of pointer
         jp    L20B           ; continuation

; RST $28 - not called

         jp    RST28          ; RAM/ROM extension
         nop
         nop
         nop
         nop
         nop

; RST $30 - not called

         jp    RST30          ; RAM/ROM extension
         nop
         nop
         nop
         nop
         nop

; RST $38 - INT Interrupt (scan line produced every 205 T)

                              ; INT = perform RST $38              11
         dec   c              ; all lines in row done ?             4
         jp    nz,L47         ; not yet                         10/10
         pop   hl             ; else point to next row             10
         dec   b              ; all rows done?                      4
         ret   z              ; yes, return to SLOW routine      5/11
         set   $03,c          ; else reset line counter and         8
.L41     ld    r,a            ; renew M1 counter for next line      9
         ei                   ; A6 going low causes INT             4
         nop                  ; dummy (fixed DFILE, no N/L's)       4
         nop                  ;   "      "     "    "   "           4
         jp    (hl)           ; send line to screen                 4

.L47     pop   de             ; discard INT return address         10
         ret   z              ; dummy, never Z                   5/11
         jr    L41            ; next line                          12

; continuation of RST $08 - push HL onto IY stack

.L4B     dec   iy
         ld    (iy+$00),l
.L50     ret

; continuation of RST $10 - pop HL from IY stack

.L51     ld    h,(iy+$00)
         inc   iy
         ret

; FAST routine

.L57     ld    a,LFST         ; fast mode remaining lines
         ex    af,af'         ; produce no display (and restore A)
         ld    hl,L98         ; read keyboard and tasking is
         ld    (NACT),hl      ; next action when NMI counter zeroed
         pop   hl             ; restore
         ret                  ; and resume main process

         nop
         nop
         nop

; jump to (HL)

.L65     jp    (hl)

; NMI Interrupt (adds overhead of 40T = 12.3uS every 64uS)

                              ; NMI = perform RST $66              11
         ex    af,af'         ; get line counter                    4
         dec   a              ; one less                            4
         jr    z,L6C          ; exhausted, do next action        7/12
         ex    af,af'         ; else update                         4
         ret                  ; and resume main process            10

.L6C     push  hl             ; preserve
         ld    hl,(NACT)      ; next action
         jp    (hl)           ; do it

; SLOW routine

.L71     ld    a,LSLW         ; slow mode bottom lines
         ex    af,af'         ; store (and restore A)
         push  af             ; preserve
         push  bc
         push  de
         ld    bc,$1809       ; rows / scan lines per row +1
         ld    hl,L98         ; read keyboard and tasking is
         ld    (NACT),hl      ; next action when NMI counter zeroed
         ld    hl,(DISB)      ; base of display memory
         ld    a,-RSCN        ; M1 events to start of RH border
         halt                 ; wait for next NMI
         out   ($FD),a        ; NMI generator off                  11
         call  L92            ; produce display                    17
         out   ($FE),a        ; NMI generator on
         pop   de             ; restore
         pop   bc
         pop   af
         pop   hl
         ret                  ; and resume main process

.L92     ld    r,a            ; R counts M1 events (instr fetch)    9
         ld    a,-SCAN        ; M1 events in each full line         9
         ei                   ; A6 going low causes INT             4
         halt                 ; wait until last character column    4
                              ; produce display (at RST $38)
                              ; (which discards INT return address)

; read keyboard

.L98     out   ($FD),a        ; NMI generator off
         in    a,($FE)        ; VSYNC on, LCNTR held reset and HSYNC off
         ld    a,LTOP         ; number of top blank lines (SLOW or FAST)
         ex    af,af'         ; to be done before
         ld    hl,(DISR)      ; display routine is
         ld    (NACT),hl      ; next action when NMI counter zeroed
         push  bc
         push  af
         ld    hl,$0000       ; result
         ld    bc,$FEFE       ; rotating bitmask / port number
.LAD     in    a,(c)          ; read keyboard port (et al)
         or    $E0            ; mask out all but KBD0-4
         cpl
         or    a              ; test result
         jr    z,LB7          ; no key pressed
         set   $00,l          ; else record on which address line
.LB7     rrc   l
         or    h              ; compound KBD0-4 reads
         ld    h,a
         rlc   b              ; next address line low
         jr    c,LAD          ; until A8 to A15 done
         ld    a,l
         and   h              ; A8.KBD0
         rrca                 ; to carry
         ld    bc,$0800       ; counter / character index
         jr    nc,LC9         ; SHIFT not pressed
         set   $06,c          ; else offset index by $40
.LC9     srl   h              ; lose KBD0
         srl   l              ; lose A8
         ld    a,h
         neg                  ; two's complement
         and   h
         sub   h
         jr    nz,LD9         ; more than one KBD1-4
         ld    a,l
         neg
         and   l
         sub   l
.LD9     jr    nz,LF4         ; more than one A9-A15
.LDB     rlc   l
         jr    nc,LE0
         add   a,b            ; becomes 1 to 7 for A9 to A15 with key
.LE0     djnz  LDB
         add   a,c            ; $00 or $40 if shift
         ld    c,a
         ld    b,$05          ; KBD1-4 (bogus KBD0)
         ld    l,$08          ; eight keys per KBD line
         xor   a              ; zero running total
.LE9     add   a,l            ; each KBD line
         rrc   h              ; next KBD bit
         jr    nc,LF0         ; no key
         add   a,c
         ld    c,a            ; form keycode index
.LF0     djnz  LE9
         jr    LFA

; if more than one key pressed (excluding shift) then C=$00 or $40 (shift alone)

.LF4     ld    b,$22          ; (? $18 for correct delay)
.LF6     djnz  LF6            ; delay  (ensure VSYNC 400uS)
         set   $07,c          ; bad key

; one key pressed (with/out shift) C = keycode $00 to $27 (or $40 to $67 shifted)

.LFA     ld    hl,KEYL
         ld    a,c            ; keycode
         and   $BF            ; mask out shift bit
         out   ($FF),a        ; VSYNC off, LCNTR running and HSYNC on
         out   ($FE),a        ; NMI generator on
         jr    nz,L109        ; key or keys pressed
         ld    (hl),a         ; else zero if no key (or shift alone)
         jr    L119

.L109    ld    a,c            ; keycode
         bit   $07,(hl)       ; previous key
         jr    nz,L119        ; was invalid
         xor   (hl)           ; same key again ?
         jr    z,L11E         ; yes
         cp    c              ; last keycode zero ?
         jr    nz,L117        ; no
         ld    (hl),c         ; new keycode
         jr    L119

.L117    set   $07,(hl)       ; make invalid
.L119    dec   hl             ; point to KEYD
         ld    (hl),$03       ; reset debounce value
         jr    L12E

; same key as last pressed so perform debounce timeout and auto-delay/repeat
; ( debounce 3 frames, auto-delay 31 frames, auto-rep 4 frames)

.L11E    dec   hl
         dec   (hl)           ; KEYD debounce/auto-rep
         ld    a,(hl)
         and   $1F            ; mask out debounce/auto-delay timeout bits
         jr    nz,L12E        ; not expired
         bit   $05,(hl)       ; auto-delay expired ?
         jr    z,L12B         ; no
         ld    (hl),$24       ; else flag done and reset auto-rep value
.L12B    inc   hl
         inc   hl             ; KEYN
         ld    (hl),c         ; store new key event

; scan linked list of tasks - update parameters

.L12E    ld    hl,(TSKS)      ; start of linked list of tasks
         push  hl             ; preserve copy
         jr    L139

.L134    ex    (sp),ix        ; preserve
.L136    pop   ix             ; retrieve
         pop   hl             ; next link in list
.L139    push  hl             ; preserve copy
         ld    a,(hl)
         inc   hl
         ld    h,(hl)
         ld    l,a            ; point to next in list
         xor   a              ; zero
         ex    (sp),hl        ; swap next with previous
         cp    h              ; already end of list ?
         jr    z,L178         ; yes

; check task timeout

         inc   hl
         inc   hl             ; point to task parameters
         ld    bc,$04FF       ; loop counter / not timed out
         push  hl             ; preserve
.L149    cp    (hl)           ; zero ?
         jr    nz,L16B        ; no, timeout not expired
         inc   hl
         djnz  L149           ; until double value scanned

; timeout is zero, copy task period to timeout

.L14F    ex    (sp),ix        ; preserve / pointer to parameters
         ld    b,4            ; loop counter
.L153    ld    a,(ix+$04)     ; copy task period
         ld    (ix+$00),a     ; to timeout
         inc   ix
         djnz  L153           ; until double value copied

; if task timed out then check for action

         inc   c              ; timed out ?
         jr    z,L136         ; no, next task in list
         bit   $06,(ix+$04)   ; STOP'ped ?
         jr    nz,L136        ; yes, next task in list
         inc   (ix+$04)       ; else accumulate RUN's
         jr    L136           ; next task in list

; timeout not expired, increment towards zero

.L16B    pop   hl             ; pointer to parameters
         push  hl             ; preserve copy
         ld    bc,$0400       ; loop counter / timed out
.L170    inc   (hl)           ; timeout zero ?
         jr    nz,L134        ; no, next task in list
         inc   hl
         djnz  L170
         jr    L14F           ; timeout now zero

; scan linked list of tasks - call execution vectors of any tasks due to RUN

.L178    pop   hl             ; discard false link
.L179    pop   hl             ; start of linked list of tasks (TSKS)
         push  hl             ; preserve copy
         ld    a,(hl)
         inc   hl
         ld    h,(hl)
         ld    l,a            ; point to next in list
         ex    (sp),hl        ; swap next with previous
         ld    a,l
         add   a,10           ; offset to action byte
         ld    l,a
         ld    a,0
         adc   a,h
         ld    h,a            ; point to task action byte
         or    a              ; already end of list ?
         jr    z,L1DB         ; yes

; check action byte for task RUN

.L18B    ld    a,(hl)         ; get task action byte
         bit   $07,a          ; LOCK'ed ?
         jr    nz,L1DB        ; yes, ignore rest of list
         bit   $06,a          ; STOP'ped ?
         jr    nz,L179        ; yes, next task in list
         or    a              ; START'ed ?
         jr    z,L179         ; yes, next task in list
         dec   (hl)           ; task is to RUN, less one
         set   $07,(hl)       ; lock task list here onwards

; prepare to call execution vector for task

         push  hl             ; preserve pointer to action byte
         push  de             ; preserve
         exx
         push  hl             ; preserve alternate registers
         push  de
         push  bc
         push  ix
         exx
         inc   hl             ; point to task execution vector
         ld    a,(hl)
         inc   hl
         ld    h,(hl)
         ld    l,a            ; execution vector
         push  hl             ; preserve
         ld    hl,DFLG
         bit   $04,(hl)       ; idle state ?
         jr    nz,L1B6        ; yes
         pop   hl             ; execution vector
         call  L65            ; jump to (HL)
         jr    L1CF           ; unstack

; idle state was active when frame interrupt occured 

.L1B6    res   $04,(hl)       ; exit idle state stack
         set   $05,(hl)       ; atomic stack change started
         ex    (sp),hl        ; execution vector
         ld    (IDSP),sp      ; preserve limited idle stack pointer
         ld    sp,(STSP)      ; restore to deeper main stack
         call  L65            ; jump to (HL)
         ld    sp,(IDSP)      ; restore idle stack pointer
         pop   hl             ; discard pointer to DFLG
         res   $05,(hl)       ; atomic stack change complete
         set   $04,(hl)       ; back to idle state stack

; task execution call complete

.L1CF    pop   ix             ; restore alternate registers
         pop   bc
         pop   de
         pop   hl
         exx
         pop   de             ; restore
         pop   hl             ; retrieve pointer to action byte
         res   $07,(hl)       ; unlock task list
         jr    L18B           ; check for more RUN's of this task

; deal with any new key event

.L1DB    pop   hl             ; discard false link
         ld    hl,KEYN        ; new key event
         ld    a,(hl)         ; get key value
         bit   $07,a          ; key event ?
         jr    nz,L1FD        ; no
         set   $07,(hl)       ; now dealt with
         push  hl             ; preserve
         ld    hl,L1D38       ; keycode table
         add   a,l
         ld    l,a
         ld    a,(hl)         ; get keycode
         bit   $07,a          ; SHIFT + SPACE ?
         jp    nz,L94D        ; yes, if safe warm reset else cold
         ld    hl,(INPB)      ; process character as input
         call  L65            ; jump to (HL)
         pop   hl             ; retrieve pointer to KEYN
         jr    nc,L1FD        ; key was succesfully put into TIB
         res   $07,(hl)       ; else try again next call
.L1FD    pop   af             ; restore
         pop   bc
         pop   hl
         ret                  ; and resume main process

; continuation of rst $18 - put character into CBUF

.L201    ld    a,(hl)         ; low byte of pointer
         dec   a              ; backwards filled
         or    $C0            ; modulo 64
         ld    (hl),a         ; update low byte of pointer
         ld    l,a            ; index to buffer
         pop   af
         ld    (hl),a         ; store character
         pop   hl
         ret

; continuation of rst $20 - get character from CBUF

.L20B    ld    l,a            ; low byte of pointer
         inc   a              ; forwards emptied
         or    $C0            ; modulo 64
         ld    (CBFP),a       ; update low byte of pointer
         ld    a,(hl)         ; fetch character
         pop   hl
         ret

; copy (HL+DE)+ to (HL)+  B times

.L215    push  hl
         push  de
         push  bc
         ex    de,hl          ; copy backwards
         add   hl,de          ; displacement
.L21A    ld    a,(hl)         ; source
         ld    (de),a         ; destination
         inc   de
         inc   hl
         djnz  L21A           ; until done
         pop   bc
         pop   de
         pop   hl
         ret

; copy A to (HL)+  B times

.L224    push  hl
         push  bc
.L226    ld    (hl),a         ; copy
         inc   hl
         djnz  L226           ; until done
         pop   bc
         pop   hl
         ret

; convert col/row co-ords pointed to by IX to offset returned in HL

.L22D    push  af
         ld    a,(ix+$01)     ; row position
         rrca
         rrca
         rrca
         ld    h,a
         and   $E0            ; row b2-0
         or    (ix+$00)       ; column b4-0
         ld    l,a            ; forms low offset
         ld    a,h
         and   $03            ; row b4-3
         ld    h,a            ; forms high offset
         pop   af
         ret

; return in HL address of window origin within display

.L241    push  ix
         push  de
         ld    de,$0002       ; offset to window co-ords
         add   ix,de
         ex    de,hl          ; (superfluous instruction)
         ld    hl,(DISB)
         ex    de,hl          ; preserve
         call  L22D           ; convert origin co-ords to offset
         add   hl,de          ; add to display base address
         pop   de
         pop   ix
         ret

; return in HL cursor address

.L256    push  de
         call  L241           ; address of window origin
         ex    de,hl          ; preserve
         call  L22D           ; convert cursor co-ords to offset
         add   hl,de          ; add to origin address
         pop   de
         ret

; return BC as window column width / row depth

.L261    ld    bc,$0000
         ld    a,(ix+$04)     ; column limit
         sub   (ix+$02)       ; column origin
         ret   c              ; bad column
         ld    b,a
         ld    a,(ix+$05)     ; row limit
         sub   (ix+$03)       ; row origin
         ret   c              ; bad row
         ld    c,a
         inc   b              ; column width
         inc   c              ; row depth
         ret

; invert character at cursor

.L277    push  af
         push  hl
         call  L256           ; cursor address
         ld    a,(hl)         ; read character
         xor   $80            ; invert bit
         ld    (hl),a         ; write back
         pop   hl
         pop   af
         ret

; ascii $0C (SHIFT A) - clear screen

.L283    call  L261           ; window columns / rows to BC
         ret   c              ; bad window origin
         call  L241           ; address of window origin to HL
         ld    de,$0020       ; row offset
         ld    a,(ix+$07)     ; byte to clear with
.L290    call  L224           ; copy A to (HL)+  B times
         add   hl,de          ; next row
         dec   c
         jr    nz,L290

; ascii $01 (SHIFT n/l) - home cursor

.L297    call  L29F           ; zero cursor column
         ld    (ix+$01),$00   ; zero cursor row
         ret

; ascii $0D - CR

.L29F    ld    (ix+$00),$00   ; zero cursor column
         ret

; scroll window up one row and clear bottom row

.L2A4    push  hl
         push  de
         push  bc
         call  L261           ; window columns / rows to BC
         jr    c,L2C2         ; bad window origin
         call  L241           ; address of window origin to HL
         ld    de,$0020       ; row offset
.L2B2    dec   c              ; window depth
         jr    z,L2BC         ; only 1 row !
.L2B5    call  L215           ; copy (HL+DE)+ to (HL)+  B times
         add   hl,de          ; next row
         dec   c
         jr    nz,L2B5        ; until done
.L2BC    ld    a,(ix+$07)     ; byte to clear new line with
         call  L224           ; copy A to (HL)+  B times
.L2C2    pop   bc
         pop   de
         pop   hl
         ret

; ascii $0A (SHIFT 6) - cursor down

.L2C6    push  hl
         push  de
         push  bc
.L2C9    call  L261           ; window columns / rows to BC
         jr    c,L2DF         ; bad window origin
         ld    a,(ix+$01)     ; cursor row
         inc   a              ; down one row
         cp    c              ; bottom of window ?
         jr    c,L2DC         ; no
         dec   c              ; cursor stays still
         ld    a,c            ; but window is...
         push  af
         call  L2A4           ; scrolled up one row
         pop   af
.L2DC    ld    (ix+$01),a     ; update cursor row
.L2DF    pop   bc
         pop   de
         pop   hl
         ret

; ascii $09 (SHIFT 8) - cursor right

.L2E3    push  hl
         push  de
         push  bc
         call  L261           ; window columns / rows to BC
         jr    c,L2FB         ; bad window origin
         ld    a,(ix+$00)     ; cursor column
         inc   a              ; right one column
         cp    b              ; end of row ?
         jr    c,L2F8         ; no
         call  L29F           ; zero cursor column
         jp    L2C9           ; down one row

.L2F8    ld    (ix+$00),a     ; update cursor column
.L2FB    pop   bc
         pop   de
         pop   hl
         ret

; ascii $0B (SHIFT 7) - cursor up

.L2FF    ld    a,(ix+$01)     ; cursor row
         or    a              ; test value
         ret   z              ; already at top
         dec   (ix+$01)       ; back one row
         ret

; ascii $08 (SHIFT 5) - cursor left

.L308    ld    a,(ix+$00)     ; cursor column
         or    a              ; test value
         ld    b,a
         jr    nz,L317        ; not at start of row
         call  L2FF           ; back one row
         nop
         call  L261           ; window columns / rows to BC
         ret   c              ; bad window origin
.L317    dec   b              ; backspace
         ld    (ix+$00),b     ; update cursor column
         ret

; routine never called - cursor to end of row

.X31C    call  L261           ; window columns / rows to BC
         dec   c
         ld    (ix+$01),c     ; cursor to last column
         ret

; reprint character at cursor position using XOR byte

.L324    push  af
         push  hl
         call  L256           ; cursor address
         ld    a,(hl)         ; read character at cursor
         and   $3F            ; lose b7 (b6 never set)
         xor   (ix+$06)       ; XOR byte
         ld    (hl),a         ; reprint
         pop   hl
         pop   af
         ret

; ascii character in A to display

.L333    push  af
         push  hl
         push  de
         push  bc
         call  L324           ; reprint cursor character using XOR byte
         and   $7F            ; mask out invert bit
         cp    $60            ; lowercase ?
         jr    c,L342         ; no
         add   a,-$20         ; else make uppercase
.L342    cp    $20            ; control code ?
         jr    nc,L357        ; no
         ld    hl,L1CC0       ; table of 32 address pointers
         add   a,a            ; form index
         add   a,l            ; into table
         ld    l,a
         call  L365           ; jump to address held at (HL)
.L34F    call  L277           ; invert character at cursor
         pop   bc
         pop   de
         pop   hl
         pop   af
         ret

; printable character $20-$60, convert to index into character bitmaps $00-$40

.L357    add   a,-$20         ; convert to internal code
         call  L256           ; cursor address
         xor   (ix+$06)       ; XOR byte
         ld    (hl),a         ; write character to display file
         call  L2E3           ; cursor right
         jr    L34F

; jump to address held at (HL)

.L365    ld    a,(hl)
         inc   hl
         ld    h,(hl)
         ld    l,a
         jp    (hl)

; move BC bytes from HL to DE (avoiding overlap)

.L36A    push  hl
         push  de
         push  bc
         push  hl
         or    a              ; clear carry
         sbc   hl,de          ; move direction
         jr    nc,L37E        ; is downwards
         ex    de,hl
         add   hl,bc          ; end of destination
         ex    de,hl
         pop   hl
         add   hl,bc          ; end of source
         dec   hl             ; last in destination
         dec   de             ; last in source
         lddr                 ; HL to DE decrementing
         jr    L381

.L37E    pop   hl
         ldir                 ; HL to DE incrementing
.L381    pop   bc
         pop   de
         pop   hl
         ret

; ascii $1B (SHIFT 0) - delete right

.L385    call  L256           ; cursor address to HL
         call  L261           ; window columns / rows to BC
         ld    a,b
         ld    b,$00
         sub   (ix+$00)       ; cursor column
         dec   a
         ld    c,a            ; characters to shift left
         jr    z,L39B         ; none
         push  hl
         pop   de             ; destination
         inc   hl             ; source
         call  L36A           ; move BC bytes from HL to DE
.L39B    ld    a,c            ; bytes shifted
         add   a,e            ; last shifted
         ld    l,a            ; point to end of row
         ld    a,(ix+$07)     ; byte to clear
         ld    (hl),a         ; end of row character
         ret

; ascii $1A (SHIFT 4) - delete cursor line, by scrolling bottom up one row

.L3A3    ld    a,(ix+$03)     ; window origin row
         push  af             ; preserve copy
         add   a,(ix+$01)     ; cursor row
         ld    (ix+$03),a     ; temporarily change row origin
         call  L2A4           ; scroll window up one row
         pop   af             ; restore
         ld    (ix+$03),a     ; origin row
         ret

; check DFLG bit $00 - current window status edit=1, command=0

.L3B5    push  hl
         ld    hl,DFLG
         bit   $00,(hl)       ; current window ?
         pop   hl
         ret

; ascii $1D (no key) - delete cursor line, by scrolling top down one row

.L3BD    push  hl
         push  de
         push  bc
         call  L261           ; window columns / rows to BC
         ld    a,(ix+$03)     ; window origin row
         push  af             ; preserve
         ld    a,(ix+$05)     ; row limit
         ld    (ix+$03),a     ; temporarily collapse window to last row
         call  L241           ; address of window origin to HL
         pop   af             ; restore
         ld    (ix+$03),a     ; origin row
         ld    de,$FFE0       ; -1 row offset
         jp    L2B2           ; scroll window down one row

; ascii $1C (SHIFT 9) - insert blank line

.L3DA    ld    a,(ix+$03)     ; window origin row
         push  af             ; preserve copy
         add   a,(ix+$01)     ; cursor row
         ld    (ix+$03),a     ; temporarily change row origin
         call  L3BD           ; effectively insert blank line
         pop   af             ; restore
         ld    (ix+$03),a     ; origin row
         jp    L29F           ; zero cursor column

; ascii $11 (SHIFT 3) - copy line to line buffer  <edit window only>

.L3EE    call  L3B5           ; edit window ?
         ret   z              ; no, ignore key
         call  L29F           ; zero cursor column
         call  L256           ; cursor address to HL
         push  hl             ; preserve
         ld    hl,(DISB)
         ld    de,512         ; offset to row 16
         add   hl,de          ; address of line buffer
         ex    de,hl
         pop   hl             ; address of row containing cursor
         call  L261           ; window columns / rows to BC
         ld    c,b            ; columns per row
         ld    b,$00
         call  L36A           ; move BC bytes from HL to DE
         ex    de,hl
         ld    b,c            ; columns copied
         call  L411           ; invert characters copied to line buffer
         ret

; invert B characters at (HL)+

.L411    push  hl
         push  bc
.L413    ld    a,(hl)
         xor   $80            ; invert
         ld    (hl),a
         inc   hl
         djnz  L413           ; until done
         pop   bc
         pop   hl
         ret

; ascii $12 (no key) -  overwrite line from line buffer  <edit window only>

.L41D    call  L3B5           ; edit window ?
         ret   z              ; no, ignore key
         call  L261           ; window columns / rows to BC
         ld    hl,(DISB)
         ld    de,512         ; offset to row 16
         add   hl,de          ; address of line buffer
         call  L411           ; invert B characters at (HL)+
         ex    de,hl
         call  L29F           ; zero cursor column
         call  L256           ; cursor address to HL
         ex    de,hl
         ld    c,b            ; columns
         ld    b,$00
         call  L36A           ; move BC bytes from HL to DE
         ld    b,c
         call  L411           ; invert B characters at (HL)+
         ret

; ascii $13 (no key) - extract line to line buffer  <edit window only>

.L441    call  L3B5           ; edit window ?
         ret   z              ; no, ignore key
         call  L3EE           ; copy line to line buffer
         jp    L3A3           ; delete line (scroll bottom up)

; ascii $14 (SHIFT 2) - insert line from line buffer  <edit window only>

.L44B    call  L3B5           ; edit window ?
         ret   z              ; no, ignore key
         call  L3DA           ; insert blank line
         jp    L41D           ; overwrite from line buffer

; ascii $1E (SHIFT 1) - enter edit mode or switch windows

.L455    push  hl
         ld    hl,DFLG
         bit   $00,(hl)       ; edit window current ?
         jr    nz,L478        ; yes
         set   $00,(hl)       ; make it current
         bit   $06,(hl)       ; edit mode ?
         jr    nz,L476        ; already on
         set   $06,(hl)       ; edit mode is now on
         ld    l,$B9          ; command window row origin
         ld    (hl),$11       ; now starts at row 17
         ld    a,$11          ; control code for copy line to line buffer
         call  L4CF           ; process character as input
         ld    a,$0C          ; control code for clear screen
         call  L4CF           ; process character as input
         call  L333           ; emit to display
.L476    pop   hl
         ret

.L478    res   $00,(hl)       ; back to command window
         pop   hl
         ret

; ascii $1F - get key input (invert cursor every other call)

.L47C    push  hl
         ld    hl,DFLG
         bit   $07,(hl)       ; last call inverted cursor ?
         set   $07,(hl)       ; (assume bit was reset)
         jr    z,L48B         ; yes, return
         res   $07,(hl)       ; else it will be now
         call  L277           ; invert character at cursor
.L48B    pop   hl
         ret

; request input

.L48D    ld    a,$1F          ; control code for get key input
         ld    hl,(INPB)      ; process character as input
         jp    (hl)           ; jump to it

; emit character in A, return carry set if emit to display not done (or busy)
; ascii $1F never sent to printer, $1E sent to display even if busy flag set

.L493    push  hl
         ld    hl,PFLG        ; print flags
         cp    $1F            ; code for get key input ?
         jr    z,L4AD         ; yes
         bit   $00,(hl)       ; emit to printer also ?
         jr    z,L4AD         ; no
         push  hl
         push  af
         ld    l,a
         ld    h,$00
         rst   $08            ; push HL onto IY stack
         ld    hl,(LPTR)      ; printer routine
         call  L65            ; jump to (HL)
         pop   af
         pop   hl
.L4AD    bit   $01,(hl)       ; emit to display ?
         jr    nz,L4BC        ; no
         ld    hl,DFLG
         cp    $1E            ; code for edit mode ?
         jr    z,L4BF         ; yes
         bit   $02,(hl)       ; display busy ?
         set   $02,(hl)       ; make busy
.L4BC    scf                  ; carry set, emit not done
         jr    nz,L4CD        ; display was busy, return
.L4BF    push  ix
         ld    ix,COWP        ; command window parameters
         call  L333           ; emit to display
         pop   ix
         or    a              ; clear carry, emit done
         res   2,(hl)         ; display no longer busy
.L4CD    pop   hl
         ret

; process character as input

.L4CF    push  hl
         ld    hl,DFLG
         bit   $00,(hl)       ; edit window current ?
         scf                  ; prime for not done
         jr    nz,L4E0        ; yes, deal with immediately
         ld    hl,(PUTB)      ; put character into TIB
         call  L65            ; jump to (HL)
.L4DE    pop   hl
         ret

; edit window active, process input immediately

.L4E0    bit   $01,(hl)       ; edit window busy ?
         set   $01,(hl)       ; make busy
         jr    nz,L4DE        ; already busy, return
         push  ix
         ld    ix,EDWP        ; edit window parameters
         cp    $20            ; control code ?
         jr    c,L4FA         ; yes
         call  L511           ; insert character into row at cursor position
.L4F3    res   $01,(hl)       ; no longer busy
         pop   ix
         or    a              ; clear carry, done
         jr    L4DE           ; done

.L4FA    cp    $0D            ; CR ?
         jr    nz,L508        ; no
         call  L333           ; emit to display
         ld    a,$0A          ; with LF also
.L503    call  L333           ; emit to display
         jr    L4F3

.L508    cp    $18            ; SHIFT Q ?
         jr    nz,L503        ; no, rest dealt with otherwise
         call  L539           ; stuff edit window cursor line into command window
         jr    L4F3

; insert character into row at cursor position

.L511    push  hl
         push  de
         push  bc
         push  af
         call  L324           ; reprint cursor character using XOR byte
         call  L256           ; cursor address to HL
         call  L261           ; window columns / rows to BC
         ld    a,b            ; columns per row
         ld    b,$00
         sub   (ix+$00)       ; cursor column
         dec   a
         ld    c,a            ; columns to end of row to be moved
         jr    z,L52E         ; none
         push  hl
         pop   de
         inc   de
         call  L36A           ; move BC bytes from HL to DE
.L52E    call  L277           ; invert character at cursor
         pop   af
         call  L333           ; emit to display
         pop   bc
         pop   de
         pop   hl
         ret

; stuff edit window cursor line into command window

.L539    push  hl
         push  de
         push  bc
         call  L324           ; reprint cursor character using XOR byte
         call  L29F           ; zero cursor column
         call  L256           ; cursor address to HL
         call  L261           ; window columns / rows to BC
         dec   b
         push  hl
         ld    a,b            ; last column in row
         add   a,l
         ld    l,a            ; point to last character in cursor line
         jr    nc,L550
         inc   h
.L550    ld    a,(hl)         ; fetch last character in row
         and   $7F            ; mask out invert bit
         jr    nz,L568        ; not space, whole line to do
.L555    dec   hl             ; else backtrack along row
         ld    a,(hl)         ; fetch character
         and   $7F            ; mask out invert bit
         jr    nz,L55D        ; not space
         djnz  L555           ; until non-space character found
.L55D    pop   hl
         call  L574           ; copy partial line into TIB
         ld    a,$0D          ; CR (force line to be parsed)
         call  L581           ; put character into TIB
         jr    L56D           ; done

.L568    pop   hl
         inc   b              ; whole line
         call  L574           ; to be copied into TIB
.L56D    call  L277           ; invert character at cursor
         pop   bc
         pop   de
         pop   hl
         ret

.L574    xor   a              ; zero
         cp    b              ; characters in line ?
         ret   z              ; yes, ignore empty line
.L577    ld    a,(hl)         ; fetch each character
         add   a,$20          ; convert to ascii
         call  L581           ; put into TIB
         inc   hl
         djnz  L577           ; until line done
         ret

.L581    push  hl
         ld    hl,(PUTB)      ; put character into TIB
         call  L65            ; jump to (HL)
         pop   hl
         ret

; stuff whole edit window into command window

.L58A    call  L3B5           ; edit window ?
         ret   nz             ; yes, return
         ld    hl,SFLG        ; system flags
         bit   $01,(hl)       ; compilation ?
         res   $01,(hl)       ; dealt with now
         ret   z              ; no, return
         ld    b,15           ; lines to process (less last done post loop)
         ld    a,$1E          ; control code to invoke edit window
         call  L5B5           ; wait until command complete
         ld    a,$01          ; control code for home cursor
         call  L4CF           ; done immediately (were in edit mode)
.L5A2    ld    a,$18          ; control code for copy edit line to command
         call  L4CF           ; done immediatley
         ld    a,$0A          ; control code for cursor down
         call  L5B5           ; wait for command to accept last line
         djnz  L5A2           ; until all lines done except last
         ld    a,$18          ; control code for copy edit line to command
         call  L5B5           ; wait for command to accept last line
         ld    a,$1E          ; control code to switch back to command window
.L5B5    call  L4CF           ; process character as input
         jp    L8EF           ; wait until command complete and return

; change state from idle to main (or vv) by means of swapping stacks

.L5BB    push  hl             ; preserve current context
         push  de
         push  bc
         push  ix
         ld    hl,DFLG
         bit   $05,(hl)       ; atomic stack change occuring ?
         jr    nz,L5E5        ; yes, abort
         set   $05,(hl)       ; atomic stack change started
         bit   $04,(hl)       ; idle state ?
         jr    z,L5D9         ; no
         res   $04,(hl)       ; to main state
         ld    (IDSP),sp      ; preserve idle stack
         ld    sp,(STSP)      ; restore main stack
         jr    L5E3

.L5D9    set   $04,(hl)       ; to idle state
         ld    (STSP),sp      ; preserve main stack
         ld    sp,(IDSP)      ; restore idle stack
.L5E3    res   $05,(hl)       ; atomic stack change complete
.L5E5    pop   ix             ; restore opposite context
         pop   bc
         pop   de
         pop   hl
         ret

; reset idle state conditions (warm reset)

.L5EB    ld    hl,DFLG
         res   $04,(hl)       ; main state
         res   $05,(hl)       ; no atomic stack change occuring
         ld    hl,LA1B        ; idle routine is only ever
         ld    (IDLE),hl      ; called by being popped from
         ld    hl,SPIDT       ; idle stack pointer
         ld    (IDSP),hl      ; reset to top
         ret

; character from TIB, return from idle routine

.L5FF    cp    $1E            ; invoke edit mode ? (or switch windows)
         jp    z,L493         ; yes, emit character in A register
         set   $07,a          ; tag character as unaccepted (as yet)
         push  hl
         ld    hl,DFLG
         bit   $04,(hl)       ; idle state ?
         jr    z,L61E         ; no, return in error (never ?)
         bit   $03,(hl)       ; busy otherwise ?
         set   $03,(hl)       ; assume not and make busy
         jr    nz,L61E        ; locked out, return in error
         call  L5BB           ; change state back to main
         res   $03,(hl)       ; no longer busy
         or    a              ; clear carry
         bit   $07,a          ; character accepted ?
         jr    z,L61F         ; yes
.L61E    scf                  ; error
.L61F    pop   hl
         ret

; fetch character from input stream

.L621    ld    a,(DFLG)
         bit   $04,a          ; idle state ?
         ret   nz             ; yes, return in error (never ?)
         sbc   a,a            ; zero (always called with carry clear)
         call  L5BB           ; fetch character
         ret

; emit CR/LF and flag occurence

.L62C    push  hl
         ld    hl,SFLG        ; system flags
         set   $00,(hl)       ; CR issued, flag occurence
         call  L6DA           ; emit CR/LF
         pop   hl
         ret

; ( -- c ) copy HL counted string to CBUF, return characters copied

.L637    push  hl             ; pointer to string
         ld    a,(hl)         ; length byte
         ld    h,$00
         and   $3F            ; restrict string length
         ld    l,a            ; return length in word
         rst   $08            ; push HL onto IY stack
         jr    z,L650         ; string length is zero
         pop   hl
         push  hl             ; retrieve pointer to string
         push  bc
         ld    b,a            ; string length
         add   a,l
         ld    l,a            ; point to end of string
         jr    nc,L64A
         inc   h
.L64A    ld    a,(hl)         ; get character from string
         rst   $18            ; put character into CBUF
         dec   hl
         djnz  L64A           ; until done
         pop   bc
.L650    pop   hl
         ret

; ( c -- ) emit characters from string in CBUF

.L652    push  af
         push  hl
         rst   $10            ; pop HL from IY stack
         ld    a,l
         and   $3F
         jr    z,L662         ; string length is zero (or >63)
         ld    l,a            ; counter
.L65B    rst   $20            ; get character from CBUF
         call  L493           ; emit character in A register
         dec   l
         jr    nz,L65B        ; until string done
.L662    pop   hl
         pop   af
         ret

; print string pointed to by HL

.L665    push  af
         call  L637           ; copy string to CBUF, return count on IY stack
         call  L652           ; emit string from CBUF, count on IY stack
         pop   af
         ret

; print ' OK' prompt

.L66E    push  hl
         ld    hl,L1D60       ; ' OK' string
         call  L665           ; print string pointed to by HL
         pop   hl
         ret

; convert hex ascii character 0-F, a-f to 0-15

.L677    and   $7F            ; limit range
         cp    $60            ; 0-9 or A-F ?
         jr    c,L67F         ; yes
         add   a,-$20         ; else make a-f = A-F
.L67F    add   a,-$30         ; 0-9 = $00-09, A-F = $11-16
         cp    $10            ; number ?
         ret   c              ; yes, done
         add   a,-7           ; A-F = $0A-0F
         ret

; ( c -- ) extract c characters from CBUF, store at HL as counted string

.L687    push  af
         push  hl
         push  de
         ex    de,hl          ; destination
         rst   $10            ; pop HL from IY stack
         ld    a,l            ; count
         and   $3F            ; modulo 64
         ld    (de),a
         jr    z,L699         ; done, empty string
         ld    l,a
.L693    inc   de
         rst   $20            ; get character from CBUF
         ld    (de),a         ; copy string
         dec   l
         jr    nz,L693        ; until done
.L699    pop   de
         pop   hl
         pop   af
         ret

; ( addr1 addr2 -- )  compare counted strings

.L69D    push  hl
         push  de
         push  bc
         rst   $10            ; pop HL from IY stack
         ex    de,hl
         rst   $10            ; pop HL from IY stack
         ld    a,(hl)         ; string 1 count
         inc   hl
         jr    L6B1

; ( c -- c )  compare counted string with CBUF

.L6A7    push  hl
         push  de
         push  bc
         ex    de,hl          ; counted string address
         rst   $10            ; pop HL from IY stack
         rst   $08            ; push HL onto IY stack
         ld    a,l
         ld    hl,(CBFP)
.L6B1    and   $3F            ; limit
         ld    b,a
         ld    a,(de)         ; string 2 count
         and   $3F            ; limit
         cp    b
         jr    nz,L6C5        ; length mismatch
         or    a
         jr    z,L6C5         ; zero length
.L6BD    inc   de
         ld    a,(de)         ; each character in strings
         cp    (hl)           ; same ?
         jr    nz,L6C5        ; no
         inc   hl
         djnz  L6BD           ; until end of strings
.L6C5    pop   bc
         pop   de
         pop   hl
         ret

; print error report in A

.L6C9    push  hl
         ld    hl,L1D66       ; ' ERROR ' string
         call  L665           ; print string pointed to by HL
         call  L493           ; emit character in A register
         ld    a,' '          ; space
         call  L493           ; emit character in A register
         pop   hl
         ret

; emit CR/LF

.L6DA    push  hl
         ld    hl,L1D34       ; string for CR with LF
         call  L665           ; print string pointed to by HL
         pop   hl
         ret

; return HL holding execution vector of next word from input

.L6E3    call  L782           ; next word found in dictionary ?
         ld    a,'U'          ; error code, undefined word
         jp    c,L878         ; no, report error and do warm reset
         call  L82A           ; point to xv of dictionary entry
         jp    LD6B           ; perform CDROP

; ( -- c ) copy next word from input to CBUF, emitting characters accordingly

.L6F1    push  hl
         ld    hl,CCNT        ; precedes CBUF, used as
         ld    (hl),0         ; string count
.L6F7    or    a              ; clear carry
         call  L621           ; fetch character from input stream
         and   $7F            ; (clear idle state tag)
         cp    $1E            ; control key ?
         jr    c,L717         ; yes
         call  L493           ; emit character in A register
         cp    ' '            ; space ?
         jr    z,L70D         ; yes
         call  nc,L737        ; put character in A into HL buffer
         jr    L6F7           ; next character

.L70D    ld    a,(hl)         ; count
         or    a              ; zero ?
         jr    z,L6F7         ; yes, get next character
.L711    call  L637           ; copy string to CBUF, return count on IY stack
         pop   hl
         or    a              ; clear carry
         ret

.L717    cp    $0D            ; CR ?
         jr    nz,L720        ; no
         call  L62C           ; emit CR/LF and flag occurence
         jr    L711           ; done

.L720    cp    $1B            ; SHIFT 0, delete ?
         jr    z,L72E         ; yes
         call  L737           ; put character in A into HL buffer
         ld    a,'.'          ; period for other control characters
.L729    call  L493           ; emit character in A register
         jr    L6F7           ; get next character

.L72E    ld    a,(hl)         ; count
         or    a              ; zero ?
         jr    z,L6F7         ; yes, get next character
         dec   (hl)           ; else reduce
         ld    a,$08          ; cursor left
         jr    L729           ; emit

; put character in A into counted string buffer pointed to by HL

.L737    push  hl
         push  af
         ld    a,(hl)         ; string length
         inc   a              ; to be added to
         and   $3F            ; limited to modulo 64 bytes
         ld    (hl),a         ; updated length
         add   a,l            ; also index
         ld    l,a            ; point to next entry
         pop   af             ; retrieve character
         ld    (hl),a         ; store it
         pop   hl
         ret

; RAM found at $2000 with $A4 or $A5 - update TSKI and HERE

.L744    ld    hl,(RTSKL)     ; task list linkage
         ld    (TSKI),hl      ; add link to internal task
         ld    hl,(RHERE)     ; pointer to end of dictionary
         ld    (HERE),hl      ; reassign HERE pointer

; ROM found at $2000 with $A4 or $A5 - introduce new dictionary

.L750    ld    hl,(RDICT)     ; pointer to latest dictionary entry
         ld    (DICT),hl      ; reassign DICT pointer

; include lastest entry in accepted dictionary

.L756    push  hl
         ld    hl,(DICT)      ; pointer to latest dictionary entry
         ld    (XDICT),hl     ; is now accepted
         ld    hl,(HERE)      ; pointer to end of dictionary
         ld    (XHERE),hl     ; is now accepted
         pop   hl
         ret

; reset DICT and HERE to accepted counterparts

.L765    push  hl
         ld    hl,(XDICT)     ; restore accepted latest dictionary entry
         ld    (DICT),hl      ; to DICT pointer
         ld    hl,(XHERE)     ; restore accepted end of dictionary
         ld    (HERE),hl      ; to HERE pointer
         pop   hl
         ret

; put offset to previous entry into dictionary

.L774    push  de
         ex    de,hl          ; previous entry position
         ld    hl,(DICT)      ; pointer to latest dictionary entry
         or    a              ; clear carry
         sbc   hl,de          ; calculate offset
         call  L7C4           ; put <HL> into dictionary
         ex    de,hl          ; restore
         pop   de
         ret

; return carry set if next word from input was not found in dictionary

.L782    call  L6F1           ; copy next word from input to CBUF
.L785    ld    hl,(DICT)      ; pointer to latest dictionary entry
.L788    call  L6A7           ; compare counted string with CBUF
         ret   z              ; match found
         call  L795           ; update HL to point to next entry in dictionary
         xor   a              ; zero
         cp    h              ; test value
         jr    nz,L788        ; until end of dictionary
         scf                  ; not found
         ret

; update HL to point to next entry in dictionary

.L795    push  de
         ld    d,h
         ld    e,l            ; copy HL to DE
         ld    a,(de)         ; string length
         and   $3F            ; restrict
         add   a,l
         ld    l,a
         jr    nc,L7A0
         inc   h
.L7A0    inc   hl             ; point to offset to next entry
         ld    a,(hl)
         inc   hl
         ld    h,(hl)
         ld    l,a
         add   hl,de          ; point to next entry
         pop   de
         ret

; ( c -- )  copy c characters from CBUF into dictionary as counted string

.L7A8    ld    hl,(HERE)      ; pointer to end of dictionary
         call  LD8D           ; perform DUP
         call  L687           ; extract from CBUF to HL as counted string
         ex    de,hl
         rst   $10            ; pop HL from IY stack
         add   hl,de          ; point to end of string
         inc   hl             ; plus count byte
         ld    (HERE),hl      ; update pointer to end of dictioanry
         ret

; store byte in A at next dictionary location

.L7B9    push  hl
         ld    hl,(HERE)      ; pointer to end of dictionary
         ld    (hl),a
         inc   hl
         ld    (HERE),hl      ; update pointer to end of dictionary
         pop   hl
         ret

; store word in HL at next dictionary location

.L7C4    push  af
         ld    a,l
         call  L7B9           ; put <A> into dictionary
         ld    a,h
         call  L7B9           ; put <A> into dictionary
         pop   af
         ret

; put JP <HL> into dictionary

.L7CF    ld    a,$C3
         jr    L7D5

; put CALL <HL> into dictionary

.L7D3    ld    a,$CD          ; opcode CALL nn
.L7D5    call  L7B9           ; put <A> into dictionary
         jp    L7C4           ; put <HL> into dictionary

; put LD HL,<HL> and RST $08 into dictionary

.L7DB    ld    a,$21          ; opcode LD HL,nn
         call  L7D5           ; put <A> and <HL> into dictionary
         ld    a,$CF          ; opcode RST $08 (push HL to IY)
         jp    L7B9           ; put <A> into dictionary

; put RET into dictionary

.L7E5    ld    a,$C9          ; opcode RET
         jp    L7B9           ; put <A> into dictionary

; parse input and perform action

.L7EA    push  hl
         call  L785           ; word in CBUF in dictionary ?
         jr    nc,L805        ; yes, parse word
         ld    hl,(NMBR)      ; parse as number routine
         call  L65            ; jump to (HL), returns carry if not a number
         jr    nc,L803        ; ok, done
         call  L6DA           ; emit CR/LF
         call  L652           ; emit string from CBUF, count on IY stack
         ld    a,'U'          ; error code, undefined word
         call  L6C9           ; print error report in A
.L803    pop   hl
         ret

.L805    call  LD6B           ; perform CDROP
         bit   $07,(hl)       ; immediate word ?
         push  af             ; preserve answer
         call  L82A           ; point to xv of dictionary entry
         pop   af             ; was it an immediate word ?
         jr    nz,L819        ; yes, perform immediate action and clean up
         call  L65            ; jump to (HL)
         call  L756           ; include lastest entry in accepted dictionary
         pop   hl
         ret

.L819    call  L65            ; jump to (HL), perform immediate action
         call  L7E5           ; put $C9 into dictionary
         ld    hl,(XHERE)     ; now call the code put down into dictionary
         call  L65            ; jump to (HL)
         call  L765           ; reset DICT and HERE to accepted counterparts
         pop   hl
         ret

; return HL pointing to execution vector of dictionary entry

.L82A    ld    a,(hl)         ; dictionary entry name length
         and   $3F            ; mask out flags
         bit   $06,(hl)       ; task entry ?
         jr    z,L833         ; no
         add   a,$02          ; else skip word holding pfa
.L833    add   a,$03          ; skip length byte and word offset
         add   a,l
         ld    l,a            ; point to xv of entry
         ret   nc
         inc   h
         ret

; return HL pointing to pfa of dictionary entry, carry set if not task

.X83A    ld    a,(hl)         ; dictionary entry name length
         bit   $06,a          ; task entry ?
         scf
         ret   z              ; no, return carry set
         and   $3F            ; mask out flags
         add   a,$03          ; skip length byte and word offset
         add   a,l
         ld    l,a            ; point to pfa pointer of entry
         jr    nc,L848
         inc   h
.L848    ld    a,(hl)
         inc   hl
         ld    h,(hl)
         ld    l,a            ; return pfa pointer
         ret

; print last word with error report in A

.L84D    push  af
         call  L6DA           ; emit CR/LF
         call  LD48           ; perform CDUP
         call  L652           ; emit string from CBUF, count on IY stack
         pop   af
         jp    L6C9           ; print error report in A

; if next word from input is in dictionary report error R and return

.L85B    call  L782           ; next word found in dictionary ?
         ret   c              ; no
         ld    a,'R'          ; error report, word redefinition
         call  L84D           ; print last word with error report in A
         or    a              ; clear carry
         ret

; set up new dictionary entry with name as next word from input

.L866    ld    hl,(HERE)      ; pointer to end of dictionary
         push  hl             ; preserve whilst
         call  L85B           ; report error if next word already in dictioanry
         call  L7A8           ; copy from CBUF into current dictionary location
         pop   hl             ; retrieve pointer to new entry
         call  L774           ; put offset to previous entry into dictionary
         ld    (DICT),hl      ; update pointer to latest dictionary entry
         ret

; report last word in error (error code in A) and do warm reset

.L878    call  L84D           ; print last word with error report in A
         jp    L99A           ; do warm reset

.L87E    call  L7D3           ; put CALL <HL> into dictionary

; parse input during construct ( DO : IF ELSE BEGIN )

.L881    call  L782           ; next word found in dictionary ?
         jr    c,L89C         ; no, try to parse as a number
         call  LD6B           ; perform CDROP
         bit   $07,(hl)       ; immediate word ?
         push  af             ; preserve answer
         call  L82A           ; point to xv of dictionary entry
         pop   af             ; was it an immediate word ?
         jr    z,L87E         ; no, just put call to it into dictionary
         ld    de,L878        ; else prime to report an error
         push  de             ; as return address
         call  L65            ; jump to (HL), perform immediate action
         pop   de             ; succesful, discard
         jr    L881           ; next word from input

.L89C    rst   $10            ; pop HL from IY stack
         xor   a              ; zero, condition where
         or    l              ; CR returned ? 
         jr    z,L881         ; yes, next word from input
         rst   $08            ; push HL onto IY stack
         ld    hl,(NMBR)      ; parse as number routine
         call  L65            ; jump to (HL), returns carry if not a number
         call  L8AD           ; compile into dictionary, else abort with error
         jr    L881           ; next word from input

.L8AD    bit   $07,a          ; double number input ?
         ld    a,'U'          ; error code, undefined word
         jp    c,L878         ; report last word in error and do warm reset
         jr    z,L8BB         ; number not a double
         rst   $10            ; pop HL from IY stack
         call  L8BB           ; else do high word first, then low
         rst   $08            ; push HL onto IY stack
.L8BB    push  hl             ; preserve
         rst   $10            ; pop HL from IY stack
         call  L7DB           ; put LD HL,<HL> and RST $08 into dictionary
         pop   hl             ; restore
         ret

; reset flags (on warm reset)

.L8C2    ld    hl,DFLG        ; device flags
         ld    a,(hl)
         and   $70            ; b4-0 cleared
         ld    (hl),a
         ld    hl,SFLG        ; system flags
         ld    a,(hl)
         and   $C0            ; b5 & b3-0 cleared   *** $D0 !!!
         ld    (hl),a
         ld    hl,PFLG        ; print flags
         ld    a,(hl)
         and   $FC            ; b1-0 cleared
         ld    (hl),a
         ret

; interpret routine

.L8D8    call  L6F1           ; copy next word from input to CBUF
         rst   $10            ; pop HL from IY stack
         xor   a              ; zero
         cp    l              ; word ?
         jr    z,L8E4         ; no
         rst   $08            ; push HL onto IY stack
         call  L7EA           ; parse input and perform action
.L8E4    ld    hl,SFLG        ; system flags
         bit   $00,(hl)       ; CR occured ?
         res   $00,(hl)       ; dealt with now
         call  nz,L66E        ; print ' OK' prompt
         ret

; wait until command complete (idle task, do auto-complie)

.L8EF    call  L910           ; get character from TIB
         ret   c              ; none
         call  L5FF           ; return from idle with character
         jr    L8EF           ; continue until TIB empty

; put character into TIB

.L8F8    push  hl
         push  af
         ld    hl,TIBI
         ld    a,(hl)         ; input index
         inc   a              ; one character to be added
         and   $BF            ; modulo 64
         ld    (hl),a         ; update
         ld    l,a
         dec   h              ; point to position within TIB
         pop   af
         ld    (hl),a         ; store character
         or    a              ; clear carry, done
         pop   hl
         ret

; return Z if TIB empty

.L909    ld    hl,TIBO
         ld    a,(hl)         ; output index
         inc   hl
         cp    (hl)           ; input index, same ?
         ret

; get character from TIB, return carry set if none

.L910    call  L909           ; return Z if TIB empty
         scf                  ; prime for not done
         ret   z              ; TIB empty
         dec   hl             ; point to output index
         inc   a              ; one character to be removed
         and   $BF            ; modulo 64 (clear carry, done)
         ld    (hl),a         ; update output index
         ld    l,a
         dec   h              ; point to position within TIB
         ld    a,(hl)         ; fetch character
         ret

; convert 0-15 to ascii hex character

.L91E    add   a,-$70
         daa                  ; 0-9 = $90-99, 10-15 = $00-05
         adc   a,$40
         daa                  ; 0-9 = $30-39, 10-15 = $41-46
         ret

; check for IY stack underflow

.L925    ld    hl,(IYSP)      ; empty IY stack position
         push  iy             ; get copy of
         pop   de             ; current IY
         or    a              ; clear carry
         sbc   hl,de          ; underflow ?
         ret   nc             ; no
         ld    a,'S'          ; error code, stack underflow
         jp    L878           ; report last word in error and do warm reset

; if TOFF turn off background task, lock out tasks and SLOW (warm reset)

.L934    ld    hl,SFLG        ; system flags
         bit   $06,(hl)       ; TON/TOFF ?
         ret   nz             ; is TON
         ld    hl,L50         ; no action for
         ld    (BTSK),hl      ; background task
         call  L1128          ; perform LOCK
         jp    L106E          ; perform SLOW

; continuation of initialisation

.L946    ld    a,$7F          ; A15 low when
         in    a,($FE)        ; read keyboard port
         rrca                 ; KBD0 to carry
         jr    nc,L953        ; SPACE key held down (COLD start)
.L94D    ld    hl,SYSF        ; system set-up byte
         inc   (hl)           ; valid ?  (internal task re-validates)
         jr    z,L99A         ; yes, system already set up (WARM start)

; COLD restart entry point

.L953    out   ($FD),a        ; NMI generator off
         ld    hl,SRAM+$FF    ; start of RAM check
.L958    ld    (hl),h         ; write upper address byte to
         inc   h              ; every 256th byte
         jr    nz,L958        ; until $FFFF (ie, over shadowed areas)
         ld    h,SRAMH        ; back to start
.L95E    ld    a,(hl)         ; search memory for first
         cp    h              ; unique match
         jr    z,L965         ; in last shadowed area of RAM
         inc   h              ; else
         jr    nz,L95E        ; until $FFFF (fatal ?)
.L965    cpl                  ; invert (shadow address upper byte)
         ld    h,a            ; yields true
         inc   hl             ; size of RAM
         exx                  ; preserve registers
         ld    hl,SRAM        ; assumed start of RAM
         ld    de,1           ; increment
         xor   a              ; zero
.L970    ld    (hl),a
         add   hl,de          ; increment (to carry)
         jr    nc,L970        ; clear all RAM 4000 to FFFF
         ld    hl,L1DA0       ; initial system data in rom
         ld    de,UINC        ; to be copied here
         ld    bc,$60
         ldir
         exx                  ; restore registers
         ld    (RAMS),hl      ; size of RAM
         ld    a,$BF
         cp    h              ; 48K RAM ?
         jr    nc,L991        ; no
         ld    hl,DFRAM       ; else new display address
         ld    (DISB),hl      ; for FBUF, SLOW & window operations
         ld    (BLKB),hl      ; for BLK, STORE & LOAD
.L991    call  L9CA           ; check for ROM/RAM at $2000
         ld    hl,L1D00       ; copyright string
         call  L665           ; print string pointed to by HL

; WARM restart entry point

.L99A    ld    sp,SPTOP       ; reset SP
         ld    hl,CBUF        ; character buffer start
         ld    (CBFP),hl      ; reset pointer
         ld    hl,$8080       ; out/in indexes
         ld    (TIBO),hl      ; for TIB reset
.L9A9    call  L934           ; if TOFF stop tasks and SLOW
         call  L5EB           ; reset idle state conditions
         call  L765           ; reset DICT and HERE to accepted counterparts
         call  L8C2           ; reset flags
         ld    iy,(IYSP)      ; reset IY data stack pointer
         ld    a,$1E          ; high byte address of character bitmaps
         ld    i,a            ; used during display generation
         out   ($FE),a        ; NMI generator on
         ld    a,(SYSI)       ; system integrity byte
         or    a              ; valid ?
         jr    nz,L953        ; no, do COLD restart
.L9C5    call  L8D8           ; interpret routine
         jr    L9C5           ; indefinitely

; check for ROM/RAM at $2000

.L9CA    ld    hl,RHEAD
         ld    a,(hl)         ; read byte
         cpl                  ; invert
         ld    (hl),a         ; write back
         cp    (hl)           ; successful write ?
         cpl                  ; reinvert and
         ld    (hl),a         ; restore if RAM
         jr    nz,LA0F        ; ROM found, check for header
         push  af             ; else preserve value read
         ld    hl,RAMS+1      ; size of RAM upper byte
         ld    a,$20          ; 8K more RAM found so
         ld    (RAMB+1),a     ; lower bottom of RAM
         add   a,(hl)         ; and increase
         ld    (hl),a         ; size of RAM
         ld    hl,SFLG        ; system flags
         set   $04,(hl)       ; flag RAM found at $2000
         pop   af             ; restore value read
         xor   $A5            ; RAM header ?
         jr    z,LA0C         ; if $A5 introduce RAM parameters and return
         dec   a              ; else
         jr    nz,L9F4        ; if not $A4 then create header
         call  L744           ; else introduce RAM parameters
.L9F0    ld    hl,(RBOOT)     ; ROM/RAM boot routine
         jp    (hl)           ; call it (return to caller L991)

; RAM found at $2000, assume 8K in size, no header found so create one

.L9F4    ld    hl,RHEAD
         ld    (hl),$A5       ; mark with header token
         ld    hl,RRAM
         ld    (HERE),hl      ; update pointer to end of dictionary
         ld    (XHERE),hl
         call  L1B70          ; perform UPDATE
         ld    hl,L19B4       ; NUL (ret)
         ld    (RBOOT),hl     ; boot routine (only when $A4 in header)
         ret

.LA0C    jp    L744           ; jump assist

; ROM found at $2000, check for header

.LA0F    xor   $A5            ; ROM header ?
.LA11    jp    z,L750         ; if $A5 introduce ROM parameters and return
         dec   a              ; else
         ret   nz             ; if not $A4 then return
         call  L750           ; else introduce ROM parameters
         jr    L9F0           ; and also call ROM routine

; idle loop, whilst waiting on input do background tasks

.LA1B    call  L925           ; check for IY stack underflow
         call  L910           ; get character from TIB
         call  nc,L5FF        ; character returned, return from idle
         call  L58A           ; do any auto-compile, wait 
         ld    hl,(BTSK)      ; do any user background task
         call  L65            ; jump to (HL)
         ld    hl,(ICTR)
         inc   hl             ; increment idle counter
         ld    (ICTR),hl
         jr    LA1B           ; loop whilst no input

; ( c -- n | d ) return CBUF as number, if error then carry set and ( c -- c )

.LA36    push  hl
         push  de
         push  bc
         call  LD48           ; perform CDUP
         xor   a              ; zero
         ld    c,a            ; zero (later holds flags)
         rst   $10            ; pop HL from IY stack
         or    l              ; CBUF character count
         jr    z,LA79         ; nothing to do
         ld    b,a            ; loop counter is character count
         ld    h,c
         ld    l,c            ; zero final result
         ld    d,c            ; zero upper byte of value word
         rst   $08            ; push HL onto IY stack
         rst   $08            ; push HL onto IY stack
         ld    hl,(BASE)      ; current number base
         ex    de,hl          ; BASE,value
.LA4C    rst   $20            ; get character from CBUF
         call  L677           ; convert ascii character to number
         cp    e              ; greater than base ?
         jr    nc,LA7E        ; yes, illegal numeric character
         ex    de,hl          ; value,BASE
         rst   $08            ; push HL onto IY stack
         ex    de,hl          ; BASE,value
         ld    l,a            ; value of character from CBUF
         call  LC13           ; perform S->D
         call  LCD6           ; perform D*
         rst   $08            ; push HL onto IY stack
         call  LC13           ; perform S->D
         call  L1305          ; perform D+
         djnz  LA4C           ; next character
.LA66    call  LB9B           ; pop HL then DE from IY stack
         bit   $00,c          ; hyphen encounteredd ?
         call  nz,LB2D        ; two's complement double DEHL
         call  LD6B           ; perform CDROP (also clears carry)
         bit   $07,c          ; dot encountered ?
         ex    de,hl          ; if so return double
         call  nz,L8          ; RST $08 - push HL onto IY stack
         ex    de,hl
         rst   $08            ; push HL onto IY stack
.LA79    ld    a,c            ; b7 set if double on stack
         pop   bc
         pop   de
         pop   hl
         ret

; illegal numeric character encountered in CBUF

.LA7E    cp    $F7            ; dot ?
         jr    z,LA90         ; yes, ok
         inc   c              ; set bit 0, if
         cp    $F6            ; hyphen ?
         jr    z,LA92         ; yes, ok
         rst   $18            ; put character into CBUF
.LA88    rst   $20            ; get character from CBUF
         djnz  LA88           ; else discard rest of contents
         rst   $10            ; pop HL from IY stack
         rst   $10            ; pop HL from IY stack
         scf                  ; error condition
         jr    LA79           ; returned

.LA90    set   $07,c          ; flag return double result
.LA92    djnz  LA4C           ; next character
         jr    LA66           ; else exhausted

; ( d -- c )  convert double to number string of c characters put into CBUF

.LA96    push  hl
         push  de
         push  bc
         xor   a              ; zero
         ld    c,$01          ; character counter (initial space)
         call  LB9B           ; pop HL then DE from IY stack
         call  LB28           ; make DEHL absolute, record sign if negative
         push  af             ; preserve sign
.LAA3    call  LB39           ; push DE then HL onto IY stack
         ld    hl,(BASE)
         rst   $08            ; push HL onto IY stack
         call  LC13           ; perform S->D
         call  LCC5           ; perform D/
         rst   $10            ; pop HL from IY stack
         ld    a,l            ; remainder
         rst   $10            ; pop HL from IY stack
         call  L91E           ; convert 0-15 to ascii hex character
         rst   $18            ; put character into CBUF
         inc   c              ; one more charcter in string
         call  LB9B           ; pop HL then DE from IY stack
         call  LD93           ; double DEHL zero ?
         jr    nz,LAA3        ; no, not finished yet
         pop   af             ; retrieve sign
         rra                  ; to carry
         jr    nc,LAC8        ; value is positive
         ld    a,'-'          ; else preced with minus sign
         rst   $18            ; put character into CBUF
         inc   c              ; one more character in string
.LAC8    ld    a,' '          ; space
         rst   $18            ; put character into CBUF
         ld    h,$00
         ld    l,c            ; string length
         rst   $08            ; push HL onto IY stack
         pop   bc
         pop   de
         pop   hl
         ret

; ( n -- c )  convert n to number string of c characters put into CBUF

.LAD3    push  hl
         push  de
         push  bc
         ld    c,$01          ; character counter (initial space)
         rst   $10            ; pop HL from IY stack
         call  LB19           ; make HL absolute, record sign if negative
         push  af             ; preserve sign (Z if positive)
.LADD    rst   $08            ; push HL onto IY stack
         ld    hl,(BASE)
         rst   $08            ; push HL onto IY stack
         call  LBF2           ; perform /MOD
         rst   $10            ; pop HL from IY stack
         ld    a,l            ; remainder
         call  L91E           ; convert 0-15 to ascii hex character
         rst   $18            ; put character into CBUF
         inc   c              ; one more character in string
         rst   $10            ; pop HL from IY stack
         call  LD98           ; HL zero ?
         jr    nz,LADD        ; no, not finished yet
         pop   af             ; retrieve sign in Z
         jr    z,LAC8         ; value is positive
         ld    a,'-'          ; else precede with minus sign
         rst   $18            ; put character into CBUF
         inc   c              ; one more character in string
         jr    LAC8

; ================================ DICTIONARY =================================

; ------ M*  ( n1 n2 -- d )  multiply to double result

.LAFB    defb  $02
         defm  "M*"
         defw  LB5F-LAFB

.LB00    push  hl
         push  de
         xor   a              ; zero negative operands
         rst   $10            ; pop HL from IY stack
         call  LB19           ; make HL absolute, record sign if negative
         ex    de,hl          ; preserve n2
         rst   $10            ; pop HL from IY stack
         call  LB19           ; make HL absolute, record sign if negative
         call  LB3E           ; unsigned multiplication DE * HL = DEHL
         or    a              ; parity is odd if signs differed
         call  po,LB2D        ; two's complement double DEHL
         call  LB39           ; push DE then HL onto IY stack
         pop   de
         pop   hl
         ret

; make HL absolute, record sign if negative

.LB19    bit   $07,h          ; sign ?
         ret   z              ; return if positive
         scf
         rla                  ; rotate in operand sign

; two's complement HL

.LB1E    push  de
         ld    de,$0000
         or    a              ; clear carry
         ex    de,hl
         sbc   hl,de          ; two's complement HL
         pop   de
         ret

; make DEHL absolute, record sign if negative

.LB28    bit   $07,d          ; sign ?
         ret   z              ; return if positive
         scf
         rla                  ; rotate in operand sign

; two's complement double DEHL

.LB2D    call  LB1E           ; two's complement HL
         ex    de,hl
         jr    nc,LB34        ; was zero
         inc   hl             ; else update
.LB34    call  LB1E           ; two's complement HL
         ex    de,hl
         ret

; push DE then HL onto IY stack

.LB39    ex    de,hl
         rst   $08            ; push HL onto IY stack
         ex    de,hl
         rst   $08            ; push HL onto IY stack
         ret

; unsigned multiplication DE * HL = DEHL

.LB3E    push  ix
         push  bc
         ld    ix,$0000       ; low word running total
         ld    b,$10          ; 16 bits to do
         or    a              ; clear carry
         jr    LB4C

.LB4A    add   ix,ix          ; reflect bit weighting doubled
.LB4C    adc   hl,hl          ; next HL bit to carry
         jr    nc,LB55        ; no contribution to sum
         add   ix,de          ; else add
         jr    nc,LB55        ; no overflow
         inc   hl             ; else add carry to result upper word
.LB55    djnz  LB4A           ; until all bits in HL done
         push  ix
         pop   de
         ex    de,hl          ; result is double DEHL
         pop   bc
         pop   ix
         ret

; ------ *  ( n1 n2 -- n3 )  multiply

.LB5F    defb  $01
         defm  "*"
         defw  LB6E-LB5F

         push  hl
         call  LB00           ; perform M*
         rst   $10            ; pop HL from IY stack
         push  hl             ; preserve low word
         rst   $10            ; pop HL from IY stack
         pop   hl             ; discard high, retrieve low word
         rst   $08            ; push HL onto IY stack
         pop   hl
         ret

; ------ M/  ( d n1 -- n2 n3 ) division d/n1, quotient n2 remainder n3
;                              quotient is negative if d was negative
;                              remainder is negative if only d was negative

.LB6E    defb  $02
         defm  "M/"
         defw  LBC5-LB6E

.LB73    push  hl
         push  de
         push  bc
         xor   a              ; clear bits storing operands sign
         rst   $10            ; pop HL from IY stack
         call  LB19           ; make HL absolute, record sign if negative
         push  hl
         pop   bc             ; absolute value n1
         call  LB9B           ; pop HL then DE from IY stack
         call  LB28           ; make DEHL absolute, record sign if negative
         call  LBA0           ; perform ud / un
         or    a              ; signs of operands
         jr    z,LB94         ; both were positive
         inc   a
         rra                  ; carry set if only n1 was negative
         call  nc,LB1E        ; two's complement HL (quotient)
         ex    de,hl
         rra                  ; carry set if only d was negative
         call  c,LB1E         ; two's complement HL (remainder)
         ex    de,hl
.LB94    call  LB39           ; push DE then HL onto IY stack
         pop   bc
         pop   de
         pop   hl
         ret

; pop HL then DE from IY stack

.LB9B    rst   $10            ; pop HL from IY stack
         ex    de,hl
         rst   $10            ; pop HL from IY stack
         ex    de,hl
         ret

; perform ud / un where BC=un, DEHL=ud and return DE quotient, HL remainder

.LBA0    push  ix             ; preserve
         push  bc             ; un
         ld    b,17           ; 16 bits to do (+1 for djnz)
         push  hl             ; ud LSW
         ex    de,hl
         pop   ix
         pop   de
         jr    LBB5           ; DE=un, HL=ud MSW, IX=ud LSW

.LBAC    adc   hl,hl          ; left shift in ud LSW bit
         or    a              ; clear carry
         sbc   hl,de          ; - un
         ccf                  ; ( jr nc,LBB9 ? can't do adc ix,ix ! )
         jr    c,LBB9         ; ok to increase quotient
         add   hl,de          ; + un, else undo
.LBB5    add   ix,ix          ; ud LSW bit to carry, quotient x2
         jr    LBBD

.LBB9    add   ix,ix          ; ud LSW bit to carry, and
         inc   ix             ; quotient x2 +1
.LBBD    djnz  LBAC
         push  ix
         pop   de             ; final quotient
         pop   ix             ; restore
         ret

; ------ */MOD  ( n1 n2 n3 -- n4 n5 )  n1*n2/n3 with double intermediate
;                                      quotient n4 remainder n5

.LBC5    defb  $05
         defm  "*/MOD"
         defw  LBD6-LBC5

.LBCD    rst   $10            ; pop HL from IY stack
         call  LB00           ; perform M*
         rst   $08            ; push HL onto IY stack
         call  LB73           ; perform M/
         ret

; ------ */  ( n1 n2 n3 -- n4 )  n1*n2/n3 with double intermediate

.LBD6    defb  $02
         defm  "*/"
         defw  LBE0-LBD6

         call  LBCD           ; perform */MOD
         rst   $10            ; pop HL from IY stack
         ret

; ------ /  ( n1 n2 -- n3 )  division

.LBE0    defb  $01
         defm  "/"
         defw  LBEB-LBE0

         push  hl
         call  LBF2           ; perform /MOD
         rst   $10            ; pop HL from IY stack
         pop   hl
         ret

; ------ /MOD  ( n1 n2 -- n3 n4 )  n1/n2 quotient n3 remainder n4

.LBEB    defb  $04
         defm  "/MOD"
         defw  LBFD-LBEB

.LBF2    push  hl
         rst   $10            ; pop HL from IY stack
         call  LC13           ; perform S->D
         rst   $08            ; push HL onto IY stack
         call  LB73           ; perform M/
         pop   hl
         ret

; ------ MOD  ( n1 n2 -- n3 )  n1/n2 remainder n3

.LBFD    defb  $03
         defm  "MOD"
         defw  LC0C-LBFD

         call  LBF2           ; perform /MOD
         call  LE41           ; perform SWAP
         jp    L137A          ; perform DROP

; ------ S->D  ( n -- d )  convert to double

.LC0C    defb  $04
         defm  "S->D"
         defw  LC22-LC0C

.LC13    push  de
         push  hl
         rst   $10            ; pop HL from IY stack
         ld    a,h
         rla                  ; b7 to carry
         ex    de,hl
         sbc   hl,hl          ; sign extend from carry
         ex    de,hl
         call  LB39           ; push DE then HL onto IY stack
         pop   hl
         pop   de
         ret

; ------ ROT  ( n1 n2 n3 -- n2 n3 n1 )  rotate

.LC22    defb  $03
         defm  "ROT"
         defw  LC34-LC22


         rst   $10            ; pop HL from IY stack
         ex    de,hl          ; n3 to DE
         rst   $10            ; pop HL from IY stack
         push  hl             ; n2 to stack
         rst   $10            ; pop HL from IY stack
         ex    (sp),hl        ; n1 to stack, n2 to HL
         rst   $08            ; push HL onto IY stack
         ex    de,hl          ; n3
         rst   $08            ; push HL onto IY stack
         pop   hl             ; n1
         rst   $08            ; push HL onto IY stack
         ret

; ------ MD/  ( uq ud1 -- ud2 ud3 )  quadruple division uq/ud1 quotient ud2 remainder ud3
;                                    where uq is a number represented in eight bytes

.LC34    defb  $03
         defm  "MD/"
         defw  LC75-LC34

; similar algorithm as used at LBA0 to perform a division, but on larger scale

.LC3A    push  ix
         push  hl
         push  de
         push  bc
         push  iy
         pop   de
         ld    hl,$0008
         add   hl,de
         push  de
         pop   ix
         ld    bc,$2008       ; 32 bits to do / 8 bytes in quad word
.LC4C    push  bc
         push  ix
         or    a              ; clear carry
.LC50    rl    (ix+$04)       ; rotate left through carry
         inc   ix
         dec   c
         jr    nz,LC50
         pop   ix
         call  L12C3          ; compare doubles (HL)-(DE)
         jr    c,LC67
         call  L12E7          ; subtract doubles (HL)=(HL)-(DE)
         set   $00,(ix+$04)
.LC67    pop   bc 
         djnz  LC4C
         pop   bc
         pop   de
         rst   $10            ; pop HL from IY stack
         rst   $10            ; pop HL from IY stack
         call  LDA3           ; perform DSWAP
         pop   hl
         pop   ix
         ret

; ------ MD*  ( ud1 ud2 -- uq )  multiply to quadruple result

.LC75    defb  $03
         defm  "MD*"
         defw  LCC0-LC75

; similar algorithm as used at LB3E to perform a multiplication, but on a larger scale

.LC7B    push  ix
         push  hl
         push  de
         push  bc
         push  iy
         rst   $08            ; push HL onto IY stack
         rst   $08            ; push HL onto IY stack
         ld    bc,$0008       ; bytes to copy
         push  iy
         pop   de             ; pointer to empty double (TOS)
         pop   hl             ; pointer to original d2 (NOS)
         call  L36A           ; move BC bytes from HL to DE
         push  de             ; stack is now ( d1 d1 d2 )
         ld    b,$04          ; counter
         pop   ix             ; TOS
.LC93    ld    (ix+$08),$00   ; clear 3OS entry
         inc   hl
         inc   ix
         djnz  LC93           ; until ( 0 d1 d2 )

; HL points to 0, IX points to d1, DE points to d2

         ld    bc,$2008       ; 32 bit multiply / quad word result
.LC9F    push  bc
         or    a              ; clear carry
         bit   $00,(ix+$00)   ; reflect bit weighting
         call  nz,L12D7       ; add doubles (HL)=(HL)+(DE)
         ld    b,c
         push  ix
.LCAB    rr    (ix+$07)       ; rotate right through carry
         dec   ix
         djnz  LCAB           ; until quad word shifted right
         pop   ix
         pop   bc
         djnz  LC9F           ; until multiplication complete
         rst   $10            ; pop HL from IY stack
         rst   $10            ; pop HL from IY stack
         pop   bc
         pop   de
         pop   hl
         pop   ix
         ret

; ------ D/  ( ud1 ud2 -- ud3 ud4 ) ud1/ud2 quotient ud3 remainder ud4

.LCC0    defb  $02
         defm  "D/"
         defw  LCD1-LCC0

.LCC5    call  LB9B           ; pop HL then DE from IY stack
         call  LCFB           ; perform D->Q
         call  LB39           ; push DE then HL onto IY stack
         jp    LC3A           ; perform MD/

; ------ D*  ( d1 d2 -- d3 )  double multiplication

.LCD1    defb  $02
         defm  "D*"
         defw  LCE7-LCD1

.LCD6    push  hl
         push  de
         call  LC7B           ; perform MD*
         call  LB9B           ; pop HL then DE from IY stack
         call  LCEF           ; perform 2DROP
         call  LB39           ; push DE then HL onto IY stack
         pop   de
         pop   hl
         ret

; ------ 2DROP  ( d -- )  discard double

.LCE7    defb  $05
         defm  "2DROP"
         defw  LCF4-LCE7

.LCEF    push  hl
         rst   $10            ; pop HL from IY stack
         rst   $10            ; pop HL from IY stack
         pop   hl
         ret

; ------ D->Q  ( d -- q )  convert to quadruple

.LCF4    defb  $04
         defm  "D->Q"
         defw  LD13-LCF4


.LCFB    push  hl
         push  de
         call  LB9B           ; pop HL then DE from IY stack
         call  LB39           ; push DE then HL onto IY stack
         push  iy
         call  LB39           ; push DE then HL onto IY stack
         pop   hl
         push  hl             ; pointer to NOS double
         rl    d              ; sign of double to carry
         pop   de             ; as HL, forcing sign extension
         call  L12E7          ; subtract doubles (HL)=(HL)-(DE)
         pop   de
         pop   hl
         ret

; ------ +  ( n1 n2 -- n3 )  addition

.LD13    defb  $01
         defm  "+"
         defw  LD21-LD13

         push  hl
         push  de
         rst   $10            ; pop HL from IY stack
         ex    de,hl
         rst   $10            ; pop HL from IY stack
         add   hl,de          ; simple addition
         rst   $08            ; push HL onto IY stack
         pop   de
         pop   hl
         ret

; ------ -  ( n1 n2 -- n3 )  subtraction

.LD21    defb  $01
         defm  "-"
         defw  LD31-LD21

         push  hl
         push  de
         rst   $10            ; pop HL from IY stack
         ex    de,hl
         rst   $10            ; pop HL from IY stack
         or    a              ; clear carry
         sbc   hl,de          ; simple subtraction
         rst   $08            ; push HL onto IY stack
         pop   de
         pop   hl
         ret

; ------ MINUS  ( n1 -- n2 )  two's complement

.LD31    defb  $05
         defm  "MINUS"
         defw  LD41-LD31

         push  hl
         rst   $10            ; pop HL from IY stack
         call  LB1E           ; two's complement HL
         rst   $08            ; push HL onto IY stack
         pop   hl
         ret

; ------ CDUP  ( c -- c c )  duplicate characters in CBUF

.LD41    defb  $04
         defm  "CDUP"
         defw  LD63-LD41

.LD48    push  af
         push  hl
         rst   $10            ; pop HL from IY stack
         rst   $08            ; push HL onto IY stack
         rst   $08            ; push HL onto IY stack
         ld    a,l            ; count
         and   $3F            ; modulo 64
         jr    z,LD60         ; if zero then done
         push  bc
         ld    b,a            ; loop counter
         ld    hl,(CBFP)      ; current pointer
         dec   a
         add   a,l
         ld    l,a            ; point to last character
.LD5A    ld    a,(hl)         ; fetch
         rst   $18            ; put character into CBUF
         dec   hl             ; next character
         djnz  LD5A           ; until done
         pop   bc
.LD60    pop   hl
         pop   af
         ret

; ------ CDROP  ( c -- )  discard characters from CBUF

.LD63    defb  $05
         defm  "CDROP"
         defw  LD79-LD63

.LD6B    push  hl
         rst   $10            ; pop HL from IY stack
         ld    a,l            ; count
         and   $3F            ; modulo 64
         jr    z,LD77         ; if zero then done
         ld    l,a            ; loop counter
.LD73    rst   $20            ; get character from CBUF
         dec   l              ; discarded
         jr    nz,LD73        ; until done
.LD77    pop   hl
         ret

; ------ ?DUP  ( n -- n n | n )  duplicate if not zero

.LD79    defb  $04
         defm  "?DUP"
         defw  LD87-LD79

         rst   $10            ; pop HL from IY stack
         rst   $08            ; push HL onto IY stack
         ld    a,l
         or    h
         ret   z              ; value was zero
         rst   $08            ; push HL onto IY stack
         ret

; ------ DUP  ( n -- n n )  duplicate

.LD87    defb  $03
         defm  "DUP"
         defw  LD9B-LD87

.LD8D    push  hl
         rst   $10            ; pop HL from IY stack
         rst   $08            ; push HL onto IY stack
         rst   $08            ; push HL onto IY stack
         pop   hl
         ret

; double DEHL zero ?

.LD93    ld    a,l
         or    h
         or    e
         or    d
         ret

; HL zero ?

.LD98    ld    a,l
         or    h
         ret

; ------ DSWAP  ( d1 d2 -- d2 d1 )  swap double

.LD9B    defb  $05
         defm  "DSWAP"
         defw  LDBC-LD9B

.LDA3    push  hl
         push  de
         push  bc
         rst   $10            ; pop HL from IY stack
         push  hl             ; d2 low word
         rst   $10            ; pop HL from IY stack
         push  hl             ; d2 high word
         rst   $10            ; pop HL from IY stack
         push  hl             ; d1 low word
         rst   $10            ; pop HL from IY stack
         pop   bc             ; d1 low word
         pop   de             ; d2 high word
         ex    de,hl          ; d1 high word, d2 high word
         rst   $08            ; push HL onto IY stack
         pop   hl             ; d2 low word
         rst   $08            ; push HL onto IY stack
         ex    de,hl          ; d1 high word
         rst   $08            ; push HL onto IY stack
         push  bc
         pop   hl             ; d1 low word
         rst   $08            ; push HL onto IY stack
         pop   bc
         pop   de
         pop   hl
         ret

; ------ DO  ( n1 n2 -- )  start of DO construct with limit n1 index n2

.LDBC    defb  $82
         defm  "DO"
         defw  LDDD-LDBC

; set up DO loop dictionary parameters

         ld    hl,LDCE        ; execution vector
         call  L7D3           ; put CALL <HL> into dictionary
         ld    hl,(HERE)      ; loop back point
         push  hl             ; to stack (for LOOP set-up)
         call  L881           ; parse input loop during construct

; entry point at execution, convert and stack limit and index from IY stack

.LDCE    pop   de             ; next vector in dictionary
         ld    bc,$8000
         call  LE41           ; perform SWAP
         rst   $10            ; pop HL from IY stack
         add   hl,bc          ; convert limit
         push  hl             ; to stack (for LOOP)
         rst   $10            ; pop HL from IY stack
         add   hl,bc          ; convert index
         push  hl             ; to stack (for LOOP)
         ex    de,hl          ; next vector
         jp    (hl)           ; jump to it

; ------ +LOOP  ( n -- )  end of DO ... +LOOP construct with increment n

.LDDD    defb  $85
         defm  "+LOOP"
         defw  LE05-LDDD

         pop   hl             ; discard return to parse routine
         pop   hl             ;    "    error routine
         pop   hl             ;    "    return to DO
         ld    hl,LDEE        ; execution vector
         jp    LE12           ; set up as with LOOP

; entry point at execution, add increment to index and compare to limit

.LDEE    pop   bc             ; return address
         pop   de             ; index
         rst   $10            ; pop HL from IY stack
         bit   $07,h          ; sign of increment
         jr    nz,LDFF        ; is negative
         add   hl,de          ; increment index
         ex    de,hl
         pop   hl             ; get copy of limit
         push  hl             ; unchanged
         push  de             ; updated index
.LDFA    scf
         sbc   hl,de          ; limit - index -1
         push  bc             ; return address to
         ret                  ; perform JP NC,nn

.LDFF    add   hl,de          ; decrement index
         pop   de             ; get copy of limit
         push  de             ; unchanged
         push  hl             ; updated index
         jr    LDFA           ; (perform index - limit -1)

; ------ LOOP  ( -- )  end of DO ... LOOP construct

.LE05    defb  $84
         defm  "LOOP"
         defw  LE2C-LE05

         pop   hl             ; discard return to parse routine
         pop   hl             ;    "    error routine
         pop   hl             ;    "    return to DO
         ld    hl,LE21        ; execution vector
.LE12    call  L7D3           ; put CALL <HL> into dictionary
         pop   hl             ; loop back point in dictionary (post DO)
         ld    a,$D2          ; opcode JP NC,nn (index reached limit test)
         call  L7D5           ; put <A> and <HL> into dictionary
         ld    hl,$E1E1       ; opcodes POP HL & POP HL (discard index and limit)
         jp    L7C4           ; put <HL> into dictionary

; entry point at execution, increment index and compare to limit

.LE21    pop   de             ; return address
         pop   bc             ; index
         pop   hl             ; get copy of limit
         push  hl             ; unchanged
         inc   bc             ; increment index
         push  bc             ; updated index
         scf
         sbc   hl,bc          ; limit - index -1
         ex    de,hl          ; return address to
         jp    (hl)           ; perform JP NC,nn

; ------ LEAVE  ( -- ) force termination of DO LOOP construct (at next LOOP)

.LE2C    defb  $05
         defm  "LEAVE"
         defw  LE3A-LE2C

         pop   hl             ; next execution vector
         pop   de             ; index
         pop   bc             ; limit discarded
         push  de             ; index is now limit
         push  de             ; index
         jp    (hl)           ; to next vector

; ------ SWAP  ( n1 n2 -- n2 n1 )  swap

.LE3A    defb  $04
         defm  "SWAP"
         defw  LE4D-LE3A

.LE41    push  hl
         push  de
         rst   $10            ; pop HL from IY stack
         ex    de,hl          ; n2
         rst   $10            ; pop HL from IY stack
         ex    de,hl          ; n1,n2
         rst   $08            ; push HL onto IY stack
         ex    de,hl          ; n1
         rst   $08            ; push HL onto IY stack
         pop   de
         pop   hl
         ret

; ------ ' <word>  ( -- addr )  return execution vector of <word>

.LE4D    defb  $01
         defm  "'"
         defw  LE56-LE4D

         call  L6E3           ; execution vector of next word to HL
         rst   $08            ; push HL onto IY stack
         ret

; ------ : <name>  ( -- )  start of colon definition

.LE56    defb  $01
         defm  ":"
         defw  LE64-LE56

         call  L866           ; set up new dictionary entry with <name>
         ld    hl,L878        ; prime to report an error
         push  hl
         call  L881           ; parse input loop during construct

; ------ ;  ( -- )  end of colon definition (or CASE construct)

.LE64    defb  $81
         defm  ";"
         defw  LE70-LE64

.LE68    call  L7E5           ; put $C9 into dictionary
         pop   hl             ; discard return to parse routine
         pop   hl             ;    "    error routine
.LE6D    pop   hl             ;    "    return to :
         pop   hl             ;    "    error routine
         ret

; ------ ;C  ( -- )  end of CODE construct

.LE70    defb  $82
         defm  ";C"
         defw  LE77-LE70

         jr    LE6D           ; discard error routine and CODE return addresses

; ------ .  ( n -- )  print n according to current BASE

.LE77    defb  $01
         defm  "."
         defw   LE81-LE77

.LE7B    call  LAD3           ; convert n to string in CBUF
         jp    L652           ; emit string from CBUF, count on IY stack

; ------ D.  ( d -- )  print double according to current BASE

.LE81    defb  $02
         defm  "D."
         defw  LE8C-LE81

         call  LA96           ; convert d to string in CBUF
         jp    L652           ; emit string from CBUF, count on IY stack

; ------ I  ( -- n )  inner loop index within DO LOOP construct

.LE8C    defb  $01
         defm  "I"
         defw  LE9A-LE8C

         pop   de
         pop   hl
         push  hl             ; inner index is second item on stack
         push  de
.LE94    ld    de,$8000       ; clearing top bit
         add   hl,de          ; yields true value
         rst   $08            ; push HL onto IY stack
         ret

; ------ J  ( -- n )  outer loop index within DO LOOP constructs

.LE9A    defb  $01
         defm  "J"
         defw  LEA8-LE9A

         ld    hl,$0006
         add   hl,sp          ; outer index is fourth item on stack
.LEA2    ld    a,(hl)
         inc   hl
         ld    h,(hl)
         ld    l,a
         jr    LE94

; ------ NI  ( n1 -- n2 )  n'th outer loop index within DO LOOP constructs

.LEA8    defb  $02
         defm  "NI"
         defw  LEB4-LEA8

         rst   $10            ; pop HL from IY stack
         add   hl,hl
         inc   hl
         add   hl,hl
         add   hl,sp          ; point to n'th index on stack
         jr    LEA2

; ------ PICK  ( n - n'th )  pick n'th item to top

.LEB4    defb  $04
         defm  "PICK"
         defw  LEC7-LEB4

         push  iy             ; current IY stack base
         rst   $10            ; pop HL from IY stack
         add   hl,hl          ; make word index
         pop   de             ; point to
         add   hl,de          ; n'th item
         ld    e,(hl)
         inc   hl
         ld    d,(hl)
         ex    de,hl          ; copy to top
         rst   $08            ; push HL onto IY stack
         ret

; ------ CR  ( -- )  emit CR with LF

.LEC7    defb  $02
         defm  "CR"
         defw  LECF-LEC7

         jp    L6DA           ; emit CR/LF

; ------ EMIT  ( c -- )  emit ascii character

.LECF    defb  $04
         defm  "EMIT"
         defw  LEDB-LECF

         rst   $10            ; pop HL from IY stack
         ld    a,l
         jp    L493           ; emit character in A register

; ------ !  ( n addr -- )  store word n at address

.LEDB    defb  $01
         defm  "!"
         defw  LEE7-LEDB

         rst   $10            ; pop HL from IY stack
         ex    de,hl          ; addr
         rst   $10            ; pop HL from IY stack
         ex    de,hl          ; n,addr
         ld    (hl),e
         inc   hl
         ld    (hl),d
         ret

; ------ @  ( addr -- n )  fetch word contents of address

.LEE7    defb  $01
         defm  "@"
         defw  LEF2-LEE7

.LEEB    rst   $10            ; pop HL from IY stack
         ld    e,(hl)
         inc   hl
         ld    d,(hl)
         ex    de,hl          ; addr,n
         rst   $08            ; push HL onto IY stack
         ret

; ------ SP  ( -- )  emit space character

.LEF2    defb  $02
         defm  "SP"
         defw  LEFC-LEF2

         ld    a,' '          ; space
         jp    L493           ; emit character in A register

; ------ <BUILDS ( -- )  start of : <name> <BUILDS definition

.LEFC    defb  $07
         defm  "<BUILDS"
         defw  LF13-LEFC

         call  L866           ; set up new dictionary entry with name from input
         call  L7D3           ; put CALL <HL> into dictionary
         ld    hl,(HERE)
         dec   hl
         dec   hl             ; point to address in CALL
         ex    (sp),hl        ; stack it
         jp    (hl)           ; and return

; ------ DOES>  ( -- )  intermediate within : <name> <BUILDS ... DOES> ... ; definition

.LF13    defb  $85
         defm  "DOES>"
         defw  LF2D-LF13

         ld    hl,LF27        ; execution vector
         call  L7D3           ; put CALL <HL> into dictionary
         ld    hl,$CFE1       ; opcodes POP HL & RST $08
         jp    L7C4           ; put <HL> into dictionary

; entry point at execution, point CALL nn of <BUILDS to return address (POP HL & RST $08)

.LF27    pop   de             ; return address (POP HL & RST $08)
         pop   hl             ; pointer to nn in CALL nn set up by <BUILDS
         ld    (hl),e
         inc   hl
         ld    (hl),d         ; store return address as nn in CALL nn
         ret

; ------ ALLOT  ( n -- )  add n space to current dictionary pointer

.LF2D    defb  $05
         defm  "ALLOT"
         defw  LF3F-LF2D

.LF35    ld    hl,(HERE)      ; pointer to end of dictionary
         ex    de,hl
         rst   $10            ; pop HL from IY stack
         add   hl,de          ; offset pointer
         ld    (HERE),hl      ; update pointer to end of dictionary
         ret

; ------ CODE <hex>'s | <word>'s  ( -- )  start of CODE construct

.LF3F    defb  $84
         defm  "CODE"
         defw  LF79-LF3F

         call  LF7F           ; perform HEX
         jr    LF4E

.LF4B    call  L7D3           ; put CALL <HL> into dictionary
.LF4E    call  L782           ; next word found in dictionary ?
         jr    c,LF69         ; no, interpret as hex value
         call  LD6B           ; perform CDROP
         bit   $07,(hl)       ; immediate command ?
         push  af             ; preserve status
         call  L82A           ; point to xv of dictionary entry
         pop   af             ; was it immediate ?
         jr    z,LF4B         ; no, ok to put into dictionary
         ld    de,L878        ; else prime to report an error
         push  de             ; as return address
         call  L65            ; jump to (HL), perform immediate action
         pop   de             ; succesful, discard
         jr    LF4E           ; next word from input

.LF69    call  LA36           ; convert CBUF to number on IY stack
         ld    a,'H'          ; error code, bad hex value
         jp    c,L878         ; report last word in error and do warm reset
         rst   $10            ; pop HL from IY stack
         ld    a,l            ; hex value is machine code
         call  L7B9           ; put <A> into dictionary
         jp    LF4E           ; next word from input

; ------ HEX  ( -- )  set BASE to sixteen

.LF79    defb  $03
         defm  "HEX"
         defw  LF86-LF79

.LF7F    ld    hl,16          ; hexadecimal
         ld    (BASE),hl      ; is new base
         ret

; ------ AND  ( n1 n2 -- n3 )  bitwise AND

.LF86    defb  $03
         defm  "AND"
         defw  LF97-LF86

         rst   $10            ; pop HL from IY stack
         ex    de,hl          ; n2
         rst   $10            ; pop HL from IY stack
         ld    a,l
         and   e              ; low byte
         ld    l,a
         ld    a,h
         and   d              ; high byte
         ld    h,a
         rst   $08            ; push HL onto IY stack
         ret

; ------ OR  ( n1 n2 -- n3 )  bitwise OR

.LF97    defb  $02
         defm  "OR"
         defw  LFA7-LF97

         rst   $10            ; pop HL from IY stack
         ex    de,hl          ; n2
         rst   $10            ; pop HL from IY stack
         ld    a,l
         or    e              ; low byte
         ld    l,a
         ld    a,h
         or    d              ; high byte
         ld    h,a
         rst   $08            ; push HL onto IY stack
         ret

; ------ XOR  ( n1 n2 -- n3 )  bitwise XOR

.LFA7    defb  $03
         defm  "XOR"
         defw  LFB8-LFA7

         rst   $10            ; pop HL from IY stack
         ex    de,hl          ; n2
         rst   $10            ; pop HL from IY stack
         ld    a,l
         xor   e              ; low byte
         ld    l,a
         ld    a,h
         xor   d              ; high byte
         ld    h,a
         rst   $08            ; push HL onto IY stack
         ret

; ------ 2VAR <name>  ( d -- )     define double variable and assign value d
;             <name>  ( -- addr )  return address of double variable

.LFB8    defb  $04
         defm  "2VAR"
         defw  LFC6-LFB8

         call  LFD1           ; perform VARIABLE
         rst   $10            ; pop HL from IY stack
         jp    L7C4           ; put <HL> into dictionary

; ------ VARIABLE <name>  ( n -- )     define variable and assign value n
;                 <name>  ( -- addr )  return address of variable

.LFC6    defb  $08
         defm  "VARIABLE"
         defw  LFE1-LFC6

.LFD1    call  L866           ; set up new dictionary entry with <name>
         ld    hl,LFDE        ; execution vector
         call  L7D3           ; put CALL <HL> into dictionary
         rst   $10            ; pop HL from IY stack
         jp    L7C4           ; put <HL> into dictionary

; entry point at execution, return on IY stack variable address in dictionary

.LFDE    pop   hl             ; address of variable
         rst   $08            ; push HL onto IY stack
         ret

; ------ TASK <name> <word>  ( -- )  set up task with associated action
;             <taskname>  ( un d  -- )   perform action un at time d

.LFE1    defb  $04
         defm  "TASK"
         defw  L1029-LFE1

         call  L866           ; set up new dictionary entry with <name>
         ld    hl,(DICT)      ; pointer to latest dictionary entry
         set   $06,(hl)       ; flag this is a task entry
         ld    hl,(HERE)
         ld    de,5           ; offset to parameters
         add   hl,de          ; form pfa pointer
         push  hl             ; preserve and also
         call  L7C4           ; put <HL> into dictionary
         ld    hl,L1190       ; execution vector
         call  L7D3           ; put CALL <HL> into dictionary
         ld    hl,13          ; parameter area and xv of <word>
         rst   $08            ; push HL onto IY stack
         call  LF35           ; perform ALLOT
         ld    bc,$0B00       ; counter / clear byte
         pop   hl
         push  hl             ; copy of pfa pointer
.L100D   ld    (hl),c         ; clear 11 bytes
         inc   hl
         djnz  L100D          ; until done
         ex    de,hl
         call  L6E3           ; execution vector of next word to HL
         ex    de,hl
         ld    (hl),e         ; store xv of <word>
         inc   hl
         ld    (hl),d
         pop   de             ; retrieve pfa pointer
         ld    hl,(TSKE)      ; end of linked list of tasks
         out   ($FD),a        ; NMI generator off
         ld    (hl),e         ; add this task to end of list
         inc   hl
         ld    (hl),d
         out   ($FE),a        ; NMI generator on
         ex    de,hl
         ld    (TSKE),hl      ; new end of linked list of tasks
         ret

; ------ AUTO  ( -- )  only display when waiting for input

.L1029   defb  $04
         defm  "AUTO"
         defw  L1054-L1029

         ld    hl,SFLG        ; system flags
         res   $07,(hl)       ; AUTO mode
         ret

; check idle counter and if required change display mode

.L1036   ld    hl,SFLG        ; system flags
         bit   $07,(hl)       ; display mode ?
         ret   nz             ; is not AUTO, return
         ld    de,0
         ld    hl,(ICTR)      ; idle counter
         ex    de,hl
         ld    (ICTR),hl      ; zero counter
         ex    de,hl
         call  LD98           ; HL zero ?
         jr    z,L1060        ; yes were busy, perform FAST
         ld    de,256         ; else this figure rates as
         sbc   hl,de          ; idle enough ?
         jr    nc,L1073       ; yes, perform SLOW
         ret                  ; else not so quite idle, return

; ------ FAST  ( -- )  turn off display

.L1054   defb  $04
         defm  "FAST"
         defw  L1067-L1054

         ld    hl,SFLG        ; system flags
         set   $07,(hl)       ; cancel AUTO mode
.L1060   ld    hl,L57         ; FAST routine
         ld    (DISR), hl     ; display routine
         ret

; ------ SLOW  ( -- )  turn on display

.L1067   defb  $04
         defm  "SLOW"
         defw  L107A-L1067

.L106E   ld    hl,SFLG        ; system flags
         set   $07,(hl)       ; cancel AUTO mode
.L1073   ld    hl,L71         ; SLOW routine
         ld    (DISR),hl      ; display routine
         ret

; ------ D@  ( addr -- d )  fetch double from address

.L107A   defb  $02
         defm  "D@"
         defw  L109B-L107A

         push  ix
         rst   $10            ; pop HL from IY stack
         push  hl
         pop   ix             ; addr
         call  L108E          ; fetch double DEHL from (IX)
         call  LB39           ; push DE then HL onto IY stack
         pop   ix
         ret

.L108E   ld    l,(ix+$00)     ; fetch double to DEHL
         ld    h,(ix+$01)
         ld    e,(ix+$02)
         ld    d,(ix+$03)
         ret

; ------ D!  ( d addr -- )  store double at address

.L109B   defb  $02
         defm  "D!"
         defw  L10BC-L109B

         push  ix
         rst   $10            ; pop HL from IY stack
         push  hl
         pop   ix             ; addr
         call  LB9B           ; pop HL then DE from IY stack
         call  L10AF          ; store double DEHL to (IX)
         pop   ix
         ret

.L10AF   ld    (ix+$00),l     ; store double from DEHL
         ld    (ix+$01),h
         ld    (ix+$02),e
         ld    (ix+$03),d
         ret

; ------ IF  ( t/f -- )  start of IF construct

.L10BC   defb  $82
         defm  "IF"
         defw  L10DB-L10BC

         ld    hl,L10D7       ; execution vector
         call  L7D3           ; put CALL <HL> into dictionary
         ld    hl,(HERE)
         inc   hl             ; point to nn
         push  hl             ; to assigned by ELSE or THEN to
         ld    a,$CA          ; opcode JP Z,nn
         call  L7D5           ; put <A> and <HL> into dictionary
         call  L881           ; parse input loop during construct
         jp    L878           ; report last word in error and do warm reset

; entry point at execution, get t/f from IY stack and set Z flag accordingly

.L10D7   rst   $10            ; pop HL from IY stack
         jp    LD98           ; HL zero ?

; ------ ELSE  ( -- )  intermediate in IF ... ELSE ... THEN construct

.L10DB   defb  $84
         defm  "ELSE"
         defw  L10F9-L10DB

         pop   hl             ; discard return to parse routine
         pop   hl             ;    "    error routine
         pop   hl             ;    "    return to IF
         pop   de             ; address of nn in JP Z,nn (IF)
         call  L7CF           ; put JP <HL> into dictionary
         ld    hl,(HERE)
         ex    de,hl
         ld    (hl),e
         inc   hl
         ld    (hl),d         ; nn to point here in dictionary
         dec   de
         dec   de             ; point to nn of JP nn
         push  de             ; to be assigned by THEN
         call  L881           ; parse input loop
         jp    L878           ; report last word in error and do warm reset

; ------ THEN  ( -- )  end of IF ... ( ELSE ... ) THEN construct

.L10F9   defb  $84
         defm  "THEN"
         defw  L110C-L10F9

         pop   hl             ; discard return to parse routine
         pop   hl             ;    "    error routine
         pop   hl             ;    "    return to IF or ELSE
         pop   de             ; pointer to nn of JP Z,nn (IF) or JP nn (ELSE)
         ld    hl,(HERE)
         ex    de,hl
         ld    (hl),e
         inc   hl
         ld    (hl),d         ; nn to point here in dictionary
         ret

; ------ REV  ( waddr -- )  reverse screen attributes of window

.L110C   defb  $03
         defm  "REV"
         defw  L1121-L110C

         ld    de,$0006       ; offset to attributes
         rst   $10            ; pop HL from IY stack
         add   hl,de
         ld    a,$80          ; invert b7
         xor   (hl)
         ld    (hl),a         ; window XOR byte
         inc   hl
         ld    a,$80          ; invert b7
         xor   (hl)
         ld    (hl),a         ; window CLS byte
         ret

; ------ LOCK  ( -- )  lock out task list execution (performed on warm reset)

.L1121   defb  $04
         defm  "LOCK"
         defw  L1132-L1121

.L1128   ld    hl,(TSKI)      ; internal task
         ld    de,10          ; offset to task flags
         add   hl,de          ; (ROM address if no added tasks)
         set   $07,(hl)       ; lock list
         ret

; ------ UNLOCK  ( -- )  unlock task list execution

.L1132   defb  $06
         defm  "UNLOCK"
         defw  L1145-L1132

         ld    hl,(TSKI)      ; internal task
         ld    de,10          ; offset to tasks flags
         add   hl,de          ; (ROM address if no added tasks)
         res   $07,(hl)       ; unlock list
         ret

; ------ STOP  ( -- 3 3 3 )

.L1145   defb  $04
         defm  "STOP"
         defw  L1153-L1145

         ld    hl,3
         rst   $08            ; push HL onto IY stack
         rst   $08            ; push HL onto IY stack
         rst   $08            ; push HL onto IY stack
         ret

; ------ START  ( -- 4 4 4 )

.L1153   defb  $05
         defm  "START"
         defw  L1162-L1153

         ld    hl,4
         rst   $08            ; push HL onto IY stack
         rst   $08            ; push HL onto IY stack
         rst   $08            ; push HL onto IY stack
         ret

; ------ IN  ( -- 0 )

.L1162   defb  $02
         defm  "IN"
         defw  L116C-L1162

         ld    hl,0
         rst   $08            ; push HL onto IY stack
         ret

; ------ EVERY  ( -- 1 )

.L116C   defb  $05
         defm  "EVERY"
         defw  L1179-L116C

         ld    hl,1
         rst   $08            ; push HL onto IY stack
         ret

; ------ AT  ( -- 2 )

.L1179   defb  $02
         defm  "AT"
         defw  L1183-L1179

         ld    hl,2
         rst   $08            ; push HL onto IY stack
         ret

; ------ RUN  ( -- 5 5 5 )

.L1183   defb  $03
         defm  "RUN"
         defw  L11F5-L1183

         ld    hl,5
         rst   $08            ; push HL onto IY stack
         rst   $08            ; push HL onto IY stack
         rst   $08            ; push HL onto IY stack
         ret

; entry point at execution (assigned by TASK), call routine for token on IY stack

.L1190   ex    (sp),ix        ; preserve / pointer to pfa
         call  LB9B           ; pop HL then DE from IY stack
         call  LB2D           ; two's complement double DEHL
         push  de             ; stack negative event time
         push  hl             ; double
         rst   $10            ; pop HL from IY stack
         ld    de,-6
         add   hl,de          ; task token
         jr    c,L11A9        ; invalid, >5
         add   hl,hl          ; form offset into
         ld    de,L11BA       ; jump table
         add   hl,de          ; routine for task token
         jp    L365           ; jump to address held at (HL)

.L11A9   pop   hl             ; discard stacked event time
         pop   hl             ; double
         pop   ix             ; restore
         ret

; task token jump table

         defw  L11C2          ; 0  IN
         defw  L11CE          ; 1  EVERY
         defw  L12A8          ; 2  AT
         defw  L11D5          ; 3  STOP
         defw  L11DE          ; 4  START
         defw  L11E7          ; 5  RUN

; update task parameter

.L11BA   out   ($FD),a        ; NMI generator off
         call  L10AF          ; store double DEHL to (IX)
         out   ($FE),a        ; NMI generator on
         ret

; task token 0 - IN, set task timeout double

.L11C2   inc   ix
         inc   ix             ; point to task timeout double
.L11C6   pop   hl
         pop   de             ; negative event time
         call  L11BA          ; store it
         pop   ix             ; restore
         ret

; task token 1 - EVERY, set task period double

.L11CE   ld    de,6
         add   ix,de          ; point to task period double
         jr    L11C6          ; set task period double

; task token 3 - STOP, stop task

.L11D5   pop   hl
         pop   hl             ; discard event time
         set   $06,(ix+$0A)   ; stop task
         pop   ix             ; restore
         ret

; task token 4 - START, start task

.L11DE   pop   hl
         pop   hl             ; discard event time
         ld    (ix+$0A),0     ; start task
         pop   ix             ; restore
         ret

; task token 5 - RUN, run task

.L11E7   pop   hl
         pop   hl             ; discard event time
         bit   $06,(ix+$0A)   ; task stopped ?
         jr    nz,L11F2       ; yes
         inc   (ix+$0A)       ; else invoke run
.L11F2   pop   ix             ; restore
         ret

; ------ TT  ( n -- d )  make frames value a double

.L11F5   defb  $02
         defm  "TT"
         defw  L11FD-L11F5

         jp    LC13           ; perform S->D

; ------ TS  ( n -- d )  multiply by frames per second

.L11FD   defb  02
         defm  "TS"
         defw  L1209-L11FD

         ld    hl,FPS         ; frames per second
         rst   $08            ; push HL onto IY stack
         jp    LB00           ; perform M*

; ------ TM  ( n -- d )  multiply by frames per minute

.L1209   defb  $02
         defm  "TM"
         defw  L1215-L1209

         ld    hl,FPM         ; frames per minute
         rst   $08            ; push HL onto IY stack
         jp    LB00           ; perform M*

; ------ TH  ( n -- d )  multiply by frames per hour

.L1215   defb  $02
         defm  "TH"
         defw  L1229-L1215

         call  LC13           ; perform S->D
         ld    de,FPHH        ; frames per hour
         ld    hl,FPHL
.L1223   call  LB39           ; push DE then HL onto IY stack
         jp    LCD6           ; perform D*

; ------ TD  ( n -- d )  multiply by frames per day

.L1229   defb  $02
         defm  "TD"
         defw  L1239-L1229

         call  LC13           ; perform S->D
         ld    de,FPDH        ; frames per day
         ld    hl,FPDL
         jr    L1223

; ------ TW  ( n -- d )  multiply by frames per week

.L1239   defb  $02
         defm  "TW"
         defw  L1249-L1239

         call  LC13           ; perform S->D
         ld    de,FPWH        ; frames per week
         ld    hl,FPWL
         jr    L1223

; ------ TY  ( n -- d )  multiply by frames per year

.L1249   defb   $02
         defm   "TY"
         defw   L1259-L1249

         call  LC13           ; perform S->D
         ld    de,FPYH        ; frames per year
         ld    hl,FPYL
         jr    L1223

; ------ CASE <word>'s  ( n -- )  start of CASE construct

.L1259   defb  $84
         defm  "CASE"
         defw  L1300-L1259

         ld    hl,L128B       ; execution vector
         call  L7D3           ; put CALL <HL> into dictionary
         ld    hl,(HERE)
         push  hl             ; start of xv list
         call  L7C4           ; put <HL> into dictionary
         ld    de,LE68        ; execution vector of ; (semicolon)
.L1270   call  L6E3           ; execution vector of next word to HL
         call  L12A2          ; perform SUB HL,DE
         jr    z,L127D        ; end of list reached (semicolon)
         call  L7C4           ; put <HL> into dictionary
         jr    L1270          ; until semicolon encountered

; end of list of <word>'s at semicolon

.L127D   ld    hl,(HERE)      ; end of list of xv's
         dec   hl
         dec   hl             ; point to last entry
         pop   de             ; pointer to list head entry
         or    a              ; clear carry
         sbc   hl,de          ; size of list
         ex    de,hl
         ld    (hl),e
         inc   hl
         ld    (hl),d         ; to list head entry
         ret

; entry point at execution, 

.L128B   pop   hl             ; address of CASE list
         ld    e,(hl)
         inc   hl
         ld    d,(hl)         ; size of list
         inc   hl             ; point to start of list
         push  hl             ; preserve copy
         add   hl,de          ; point to end of list
         ex    (sp),hl        ; preserve copy
         push  hl             ; under start pointer
         rst   $10            ; pop HL from IY stack
         add   hl,hl          ; n word entries
         call  L12A2          ; perform SUB HL,DE
         jr    nc,L12A0       ; invalid index
         pop   de             ; start of list
         add   hl,de          ; point to required entry
         jp    L365           ; jump to address held at (HL)

.L12A0   pop   hl             ; discard start of list
         ret                  ; return past end of list

; perform SUB HL,DE

.L12A2   push  hl
         or    a              ; clear carry
         sbc   hl,de
         pop   hl
         ret

; task token 2 - AT, calculate and set task timeout double

.L12A8   ld    hl,0           ; point to negative event time double
         add   hl,sp          ; on stack
.L12AC   ld    de,TIME        ; point to current time
         or    a              ; clear carry
         call  L12D7          ; add doubles (HL)=(HL)+(DE)
         jp    nc,L11C2       ; event time not yet occured, perform IN
         call  L12E7          ; subtract doubles (HL)=(HL)-(DE)
         ld    e,$68          ; PER, period of TIME before wrap-around
         call  L12E7          ; subtract doubles (HL)=(HL)-(DE)
         jr    L12AC

; compare doubles on IY stack

.L12C0   call  L12F8          ; set up DE & HL

; compare doubles pointed to by HL and DE, return status of HL-DE

.L12C3   push  hl
         push  de
         ld    bc,$0004
         add   hl,bc          ; from MSB
         ex    de,hl
         add   hl,bc          ; from MSB
         ld    b,c            ; bytes to check
.L12CC   dec   hl
         dec   de
         ld    a,(de)
         cp    (hl)           ; if any comparison fails with
         jr    nz,L12D4       ; not equal, other flags set accordingly
         djnz  L12CC
.L12D4   pop   de
         pop   hl
         ret

; add doubles pointed to by HL and DE, result to HL location

.L12D7   push  hl
         push  de
         push  bc
         ld    b,$04          ; bytes to add
.L12DC   ld    a,(de)
         adc   a,(hl)
         ld    (hl),a
         inc   hl
         inc   de
         djnz  L12DC          ; until doubles done
         pop   bc
         pop   de
         pop   hl
         ret

; subtract double pointed to by DE from that pointed to by HL, result to HL location

.L12E7   push  hl
         push  de
         push  bc
         ld    b,$04          ; bytes to subtract
         ex    de,hl
.L12ED   ld    a,(de)
         sbc   a,(hl)         ; (carry from caller)
         ld    (de),a
         inc   hl
         inc   de
         djnz  L12ED          ; until double done
         pop   bc
         pop   de
         pop   hl
         ret

; set up DE & HL to point to doubles on IY stack

.L12F8   push  iy
         pop   de
         ld    hl,$0004
         add   hl,de
         ret

; ------ D+  ( d1 d2 -- d3 )  double addition

.L1300   defb  $02
         defm  "D+"
         defw  L1312-L1300

.L1305   push  hl
         push  de
         call  L12F8          ; set up DE & HL to
         call  L12D7          ; add doubles (HL)=(HL)+(DE)
         rst   $10            ; pop HL from IY stack
         rst   $10            ; pop HL from IY stack
         pop   de
         pop   hl
         ret

; ------ D-  ( d1 d2 -- d3 )  double subtraction

.L1312   defb  $02
         defm  "D-"
         defw  L1324-L1312

         push  hl
         push  de
         call  L12F8          ; set up DE & HL to
         call  L12E7          ; subtract doubles (HL)=(HL)-(DE)
         rst   $10            ; pop HL from IY stack
         rst   $10            ; pop HL from IY stack
         pop   de
         pop   hl
         ret

; ------ CLS  ( -- )  clear screen and home cursor

.L1324   defb  $03
         defm  "CLS"
         defw  L132F-L1324

         ld    a,$0C          ; control code for clear screen
         jp    L493           ; emit character in A register

; ------ ABS  ( n -- un )  absolute value

.L132F   defb  $03
         defm  "ABS"
         defw  L133B-L132F

         rst   $10            ; pop HL from IY stack
         call  LB19           ; make HL absolute, record sign if negative
         rst   $08            ; push HL onto IY stack
         ret

; ------ DMINUS  ( d1 -- d2 )  two's complement double

.L133B   defb  $06
         defm  "DMINUS"
         defw  L134D-L133B

         call  LB9B           ; pop HL then DE from IY stack
         call  LB2D           ; two's complement double DEHL
         jp    LB39           ; push DE then HL onto IY stack

; ------ DABS  ( d1 -- d2 )  absolute value double

.L134D   defb  $04
         defm  "DABS"
         defw  L135D-L134D

         call  LB9B           ; pop HL then DE from IY stack
         call  LB28           ; make DEHL absolute, record sign if negative
         jp    LB39           ; push DE then HL onto IY stack

; ------ C@  ( addr -- c )  fetch character from address

.L135D   defb  $02
         defm  "C@"
         defw  L1368-L135D

         rst   $10            ; pop HL from IY stack
         ld    l,(hl)         ; fetch character
         ld    h,$00
         rst   $08            ; push HL onto IY stack
         ret

; ------ C!  ( c addr -- )  store character at address

.L1368   defb  $02
         defm  "C!"
         defw  L1373-L1368

         rst   $10            ; pop HL from IY stack
         ex    de,hl          ; addr
         rst   $10            ; pop HL from IY stack
         ld    a,l            ; character
         ld    (de),a         ; to addr
         ret

; ------ DROP  ( n -- )  discard

.L1373   defb  $04
         defm  "DROP"
         defw  L137C-L1373

.L137A   rst   $10            ; pop HL from IY stack
         ret

; ------ BLANKS  ( addr un -- )  zero un bytes from address onwards ( un>0 )

.L137C   defb  $06
         defm  "BLANKS"
         defw  L138B-L137C

         rst   $10            ; pop HL from IY stack
         ex    de,hl          ; un bytes
         ld    c,$00          ; to zero
         jr    L1396          ; fill them

; ------ FILL  ( addr un c -- )  fill un bytes with c from address onwards ( un>0 )

.L138B   defb  $04
         defm  "FILL"
         defw  L139F-L138B

         rst   $10            ; pop HL from IY stack
         ld    c,l            ; fill with c
         rst   $10            ; pop HL from IY stack
         ex    de,hl          ; un bytes
.L1396   rst   $10            ; pop HL from IY stack
.L1397   ld    (hl),c         ; addr
         inc   hl
         dec   de             ; decrement un (no check for 0)
         ld    a,e
         or    d
         jr    nz,L1397       ; until done
         ret

; ------ COPY  ( addr1 addr2 -- )  copy 512 bytes from addr1 to addr2

.L139F   defb  $04
         defm  "COPY"
         defw  L13AC-L139F

         ld    hl,512         ; bytes to copy
         push  hl
         jr    L13B6          ; move them

; ------ MOVE  ( addr1 addr2 un -- )  move un cells from addr1 to addr2

.L13AC   defb  $04
         defm  "MOVE"
         defw  L13EF-L13AC

         rst   $10            ; pop HL from IY stack
         add   hl,hl          ; un cells to bytes
         push  hl
.L13B6   rst   $10            ; pop HL from IY stack
         ex    de,hl          ; addr2 destination
         rst   $10            ; pop HL from IY stack
         pop   bc             ; counter
         jp    L36A           ; move BC bytes from HL to DE

; internal task execution vector

.L13BD   call  L13D2          ; update TIME & UCTR
         call  L909           ; if TIB empty then
         call  z,L48D         ; flash cursor (request input)
         call  L18FA          ; check for memory corruption
         call  L1036          ; AUTO mode action
         ld    hl,SYSF        ; system set-up byte
         ld    (hl),$FF       ; make valid
         ret

; update TIME & UCTR, if TIME exceeds PER then wrap-around

.L13D2   push  hl
         push  de
         ld    hl,UCTR
         ld    de,UINC
         call  L12D7          ; UCTR = UCTR + UINC
         ld    l,$5C          ; point to TIME
         ld    e,$64          ; point to TINC
         call  L12D7          ; TIME = TIME + TINC
         ld    e,$68          ; point to PER
         call  L12C3          ; if TIME >= PER
         call  nc,L12E7       ; then TIME = TIME - PER
         pop   de
         pop   hl
         ret

; ------ SCREEN <name>  ( xo yo xl yl -- )  set up window parameter block
;               <name>  ( -- waddr )        return address of window parameter block

.L13EF   defb  $06
         defm  "SCREEN"
         defw  L141B-L13EF

         call  L866           ; set up new dictionary entry with <name>
         ld    hl,LFDE        ; execution vector (pop stack to IY)
         call  L7D3           ; put CALL <HL> into dictionary
         ld    hl,0           ; x,y cursor co-ords initialised
         push  hl
         call  L7C4           ; put <HL> into dictionary
         rst   $10            ; pop HL from IY stack
         ld    d,l            ; y co-ord limit
         rst   $10            ; pop HL from IY stack
         ld    e,l            ; x co-ord limit
         rst   $10            ; pop HL from IY stack
         ld    a,l            ; y co-ord origin
         rst   $10            ; pop HL from IY stack
         ld    h,a            ; x co-ord origin
         call  L7C4           ; put <HL> into dictionary
         ex    de,hl
         call  L7C4           ; put <HL> into dictionary
         pop   hl             ; XOR and CLS bytes initialised
         jp    L7C4           ; put <HL> into dictionary

; ------ .C  ( c waddr -- )  emit character c to window

.L141B   defb  $02
         defm  ".C"
         defw  L142C-L141B

         rst   $10            ; pop HL from IY stack
         push  hl
         ex    (sp),ix        ; address of screen parameters
         rst   $10            ; pop HL from IY stack
         ld    a,l            ; character
         call  L333           ; emit to display
         pop   ix
         ret

; ------ .W  ( c waddr -- )  emit c characters from CBUF to window

.L142C   defb  $02
         defm  ".W"
         defw  L1446-L142C

         rst   $10            ; pop HL from IY stack
         push  hl
         ex    (sp),ix        ; address of screen parameters
         rst   $10            ; pop HL from IY stack
         ld    a,l            ; count
         and   $3F            ; modulo 64
         jr    z,L1443        ; if zero then done
         ld    l,a            ; loop counter
.L143C   rst   $20            ; get character from CBUF
         call  L333           ; emit to display
         dec   l
         jr    nz,L143C       ; until done
.L1443   pop   ix
         ret

; ------ #  ( n -- c )  convert n to numeric string in CBUF of c characters

.L1446   defb  $01
         defm  "#"
         defw  L144D-L1446

         jp    LAD3           ; convert n to string in CBUF of c characters

; ------ D#  ( d -- c )  convert d to numeric string in CBUF of c characters

.L144D   defb  $02
         defm  "D#"
         defw  L1455-L144D

         jp    LA96           ; convert d to string in CBUF

; ------ ED  ( -- waddr )  return address of edit window parameters

.L1455   defb  $02
         defm  "ED"
         defw  L145F-L1455

         ld    hl,EDWP        ; edit window parameters
         rst   $08            ; push HL onto IY stack
         ret

; ------ CO  ( -- waddr )  return address of command window parameters

.L145F   defb  $02
         defm  "CO"
         defw  L1469-L145F

         ld    hl,COWP        ; command window parameters
         rst   $08            ; push HL onto IY stack
         ret

; ------ W@ ( addr -- c )  copy counted string to CBUF, return count

.L1469   defb  $02
         defm  "W@"
         defw  L1472-L1469

         rst   $10            ; pop HL from IY stack
         jp    L637           ; copy string to CBUF, return count on IY stack

; ------ C>N  ( n1 -- n2 c )  if n1 >0 increment n1 and fetch c from CBUF
;                             else if n1 =0 return n1 and c both as 0

.L1472   defb  $03
         defm  "C>N"
         defw  L148A-L1472

         ld    de,$FFFF
         xor   a              ; zero is default
         rst   $10            ; pop HL from IY stack
         add   hl,de          ; decrement (to carry)
         jr    c,L1483        ; n1 not zero
         inc   hl             ; return to zero
         jr    L1484

.L1483   rst   $20            ; get character from CBUF
.L1484   rst   $08            ; push HL onto IY stack
         ld    h,$00
         ld    l,a
         rst   $08            ; push HL onto IY stack
         ret

; ------ N>C  ( n1 c -- n2 )  store c in CBUF and increment n1

.L148A   defb  $03
         defm  "N>C"
         defw  L1497-L148A

         rst   $10            ; pop HL from IY stack
         ld    a,l            ; character
         rst   $18            ; put character into CBUF
         rst   $10            ; pop HL from IY stack
         inc   hl             ; increment counter
         rst   $08            ; push HL onto IY stack
         ret

; ------ "  ( -- c )  define string to CBUF, terminate with ", return count

.L1497   defb  $81
         defb  $22
         defw  L14D2-L1497

         ld    hl,L14BE       ; execution vector
.L149E   call  L7D3           ; put CALL <HL> into dictionary
         ld    hl,(HERE)      ; retain pointer to
         ld    a,-1           ; string count
.L14A6   call  L7B9           ; put <A> into dictionary
         inc   (hl)           ; increment string count
         or    a              ; clear carry
         call  L14C5          ; fetch and print character from input stream
         cp    '"'
         jr    nz,L14A6       ; until string terminator found
         ret

; return HL as start and DE as end+1 of counted string, address taken off stack

.L14B3   pop   de             ; return address
         pop   hl             ; pointer to counted string
         push  de             ; restore return address
         ld    e,(hl)         ; fetch count
         ld    d,$00
         ex    de,hl
         add   hl,de          ; pointer to end of string
         inc   hl             ; post end
         ex    de,hl          ; returned in DE, start in HL
         ret

; entry point at execution, stuff string into CBUF

.L14BE   call  L14B3          ; get string pointers
         push  de             ; return address is post string in dictionary
         jp    L637           ; copy string to CBUF, return count on IY stack

; fetch and print character from input stream

.L14C5   call  L621           ; fetch character from input stream
         and   $7F            ; (clear idle state tag)
         call  L493           ; emit character in A register
         cp    $1F            ; control code for get key input ?
         jr    z,L14C5        ; yes, ignore
         ret                  ; else return

; ------ ."  ( -- )  define string to be printed, terminate with "

.L14D2   defb  $82
         defb  '.',$22
         defw  L14E3-L14D2

         ld    hl,L14DC       ; exection vector
         jr    L149E          ; perform " (put string into dictionary)

; entry point at execution, print string

.L14DC   call  L14B3          ; get string pointers
         push  de             ; return address is post string in dictionary
         jp    L665           ; print string pointed to by HL

; ------ ABORT"  ( t/f -- )  define string, terminate with ", emit and WARM if true

.L14E3   defb  $86
         defm  "ABORT"
         defb  $22
         defw  L1569-L14E3

         ld    hl,L14F1       ; execution vector
         jr    L149E          ; perform " (put string into dictionary)

; entry point at execution, if truth on IY stack print string and WARM

.L14F1   call  L14B3          ; get string pointers
         push  de             ; return address is post string in dictionary
         push  hl             ; start of counted string
         rst   $10            ; pop HL from IY stack
         ld    a,l
         or    h              ; truth condition ?
         pop   hl
         ret   z              ; if false, return
         call  L665           ; print string pointed to by HL
         jp    L99A           ; do WARM restart

;--------------------------------------------------------------------;
; cassette saving routines (324 baud)                                ;
; ------------------------                                           ;
;                                                                    ;
; MIC output idle condition is high. Each byte preceded by one start ;
; bit (=1), bit 7 to bit 0 of data byte, and then one stop bit (=0). ;
;              ________________                ______     ______     ;
; bit = 0 |   |                |  bit = 1 |   |      |   |      |    ;
;         |___|                |          |___|      |___|      |    ;
;   T =    269       2820                  269  1261  269  1288      ;
;                                                                    ;
; Format : header of 125 bytes of $00                                ;
;          ident byte $A5                                            ;
;          block number (low then high byte)                         ;
;          data block 512 bytes                                      ;
;          trailer of 125 bytes of $00                               ;
;                                                                    ;
; Timing : from rising edge of stop bit (=0) low pulse to falling    ;
; edge of next start bit (=1) low pulse in T states =                ;
;                                                                    ;
;    2915  between header bytes                                      ;
;    2978  last header byte to ident byte                            ;
;    2960  ident byte to block number low byte                       ;
;    2908  between block number bytes                                ;
;    2981  block number high byte to first data byte                 ;
;    2943  between data bytes                                        ;
;    3024  last data byte to first trailer byte                      ;
;    2915  between trailer bytes                                     ;
;____________________________________________________________________;

; punctuate MIC high output with a low pulse 269T (82.77uS)

.L1501   push  af             ;                                    11
         in    a,($FE)        ; MIC output low     _               11
         ld    a,16           ;                     | LOW           7
.L1506   dec   a              ;                     |             ( 4 )
         jr    nz,L1506       ; delay for low       |     251  ( 7/12 )
         out   ($FF),a        ; MIC output high    _|              11
         ld    a,75           ;                                     7
.L150D   dec   a              ;                                   ( 4 )
         jr    nz,L150D       ; delay remain high        1195  ( 7/12 )
         pop   af             ;                                    10
         ret                  ;                                    10

; delay whilst MIC output high

.L1512   push  bc             ;                                    11
         ld    b,114          ;                                     7
.L1515   djnz  L1515          ; delay remain high        1477  ( 8/13 )
         pop   bc             ;                                    10
         ret                  ;                                    10

; save byte to MIC output

.L1519   push  af             ;                                    11
         push  bc             ;                                    11
         ld    b,10           ; bit counter (start+byte+stop)       7
         ld    a,(hl)         ; byte to save                        7
         scf                  ; start bit = 1                       4
.L151F   call  L1501          ; pulse MIC low            1530     (17)
         call  c,L1501        ; again if bit = 1    10 / 1530  (10/17)
         call  nc,L1512       ; else leave high     10 / 1532  (10/17)
         add   a,a            ; next bit to carry                   4
         djnz  L151F          ; until stop bit = 0               8/13
         pop   bc             ;                                    10
         pop   af             ;                                    10
         ret                  ;                                    10

; save header / trailer

.L152E   push  bc             ;                                    11
         push  hl             ;                                    11
         ld    hl,L153D       ; header / trailer byte              10
         ld    b,125          ; length                              7
.L1535   call  L1519          ; save byte                          17
         djnz  L1535          ; until done                       8/13
         pop   hl             ;                                    10
         pop   bc             ;                                    10
         ret                  ;                                    10

.L153D   defb  $00            ; header / trailer byte
.L153E   defb  $A5            ; ident byte

; save block

.L153F   ld    de,512         ; block size                         10
         push  bc             ;                                    11
         push  hl             ; block address                      11
         ld    bc,-1          ;                                    10
.L1547   call  L1519          ; save byte                          17
         inc   hl             ;                                     6
         ex    de,hl          ;                                     4
         add   hl,bc          ; less one                           15
         ex    de,hl          ;                                     4
         jr    c,L1547        ; until block done                 7/12
         pop   hl             ;                                    10
         pop   bc             ;                                    10
         ret                  ;                                    10

; save ident

.L1553   push  hl             ;                                    11
         ld    hl,L153E       ; ident byte                         10
         call  L1519          ; save byte                          17
         pop   hl             ;                                    10
         ret                  ;                                    10

; save block number

.L155C   push  hl             ;                                    11
         ld    hl,PAGE        ; holds current block number         10
         call  L1519          ; save low byte                      17
         inc   hl             ;                                     6
         call  L1519          ; save high byte                     17
         pop   hl             ;                                    10
         ret                  ;                                    10

; ------ STORE  ( n -- )  save block n to cassette port

.L1569   defb  $05
         defm  "STORE"
         defw  L1600-L1569

         ld    hl,DFLG
         bit   $06,(hl)       ; edit mode ?
         ret   z              ; is off, ignore command
         rst   $10            ; pop HL from IY stack
         ld    (PAGE),hl      ; save as block number n
.L157B   out   ($FD),a        ; NMI generator off
         ld    hl,(BLKB)      ; block base address
         call  L152E          ; save a header 125 bytes of $00
         call  L1553          ; save ident byte $A5                17
         call  L155C          ; save block number low/high bytes   17
         call  L153F          ; save 512 byte block                17
         call  L152E          ; save a trailer 125 bytes of $00    17
         out   ($FE),a        ; NMI generator on
         ret

;--------------------------------------------------------------------;
; cassette loading routines (324 baud)                               ;
; -------------------------                                          ;
; EAR input idle condition is low. Each bit is preceded by one start ;
; bit (=1), bit 7 to bit 0 of data byte, and then one stop bit (=0). ;
;          ___                             ___        ___            ;
; bit = 0 |   |                |  bit = 1 |   |      |   |      |    ;
;         |   |________________|          |   |______|   |______|    ;
;   T =    269       2820                  269  1261  269  1288      ;
;                                                                    ;
; Format : header lead in of 125 bytes of $00 (ignored)              ;
;          ident byte $A5                                            ;
;          block number (low then high byte)                         ;
;          data block 512 bytes                                      ;
;          trailer lead out of 125 bytes of $00 (ignored)            ;
;                                                                    ;
; Criterion : wait for rising edge of start bit, wait 857 T states,  ;
; sample level and then every 43T for a total of 37 times.  If 4 or  ;
; more samples are found high then bit =1 else =0.                   ;
;                                                                    ;
; Timing : maximum delay from reading last level sample for b0 of    ;
; previous byte to looking for rising edge of next start bit =       ;
;                                                                    ;
;     203  ident byte to block number low byte                       ;
;     169  between block number bytes                                ;
;     341  block number high byte to first data byte                 ;
;     212  between data bytes                                        ;
;                                                                    ;
; Note : bit width STORE'd is 3089 (=0) or 3087 (=1). Last sample of ;
; trailing bit 0 on LOAD is done at 2448 from start of bit. Stop bit ;
; (=0) is read and discarded as part of the next byte read since the ;
; start bit (=1) delimits the compile process (shifted into carry).  ;
;____________________________________________________________________;

; look for ident, abort if SPACE key pressed

.L1592   call  L15A5          ; fetch a byte                       17
         jr    c,L159C        ; SPACE key or EAR input too low   7/12
         cp    $A5            ; correct ident ?                     7
         jr    nz,L1592       ; no, continue trying              7/12
         ret                  ;                                    10

; confirm SPACE pressed and abort load operation or repeat attempt

.L159C   ld    a,$7F          ; A15 strobe low                      7
         in    a,($FE)        ; read keyboard port                 11
         rrca                 ; b0 to carry                         4
         ccf                  ; SPACE key pressed ?                 4
         jr    nc,L1592       ; no, EAR was too low, retry       7/12
         ret                  ; else abort LOAD                    10

; fetch a byte from the EAR input (reject if level low > 14592 T )

.L15A5   push  de             ;                                    11
         push  bc             ;                                    11
         ld    e,0            ; byte compiled                       7
.L15A9   ld    bc,$0000       ; loop counter / level counter       10
.L15AC   ld    a,$7F          ; A15 strobe low                      7
         in    a,($FE)        ; read keboard port & EAR input      11
         bit   $00,a          ; SPACE key pressed ?                 8
         jr    z,L15B9        ; yes, return                      7/12
         rlca                 ; EAR input high ?                    4
         jr    c,L15BD        ; yes, compile byte                7/12
         djnz  L15AC          ; loop 256 times whilst low        8/13
.L15B9   pop   bc             ; then give up                       10
         pop   de             ;                                    10
         scf                  ; SPACE key or EAR input too low     10
         ret                  ;

; rising egde of start bit found, compile byte

.L15BD   ld    b,62           ;                                     7
.L15BF   djnz  L15BF          ; delay                     801  ( 8/13 )
         ld    b,37           ; loop counter                        7
.L15C3   in    a,($FE)        ; EAR input level                    11
         rlca                 ; to carry                            4
         ld    a,c            ; whilst high                         4
         adc   a,0            ; increment                           7
         ld    c,a            ; level counter                       4
         djnz  L15C3          ; sample level 37 times            8/13
         cp    4              ; bit value to carry                  7
         ccf                  ; invert                              4
         rl    e              ; accumulate bit values               8
         jr    nc,L15A9       ; until start bit in carry         7/12
         pop   bc             ;                                    10
         ld    a,e            ; compiled byte                       4
         pop   de             ;                                    10
         or    a              ; clear carry, ok                     4
         ret                  ;                                    10

; clear block load buffer

.L15D8   ld    b,0            ; loop counter                        7
         call  L15DD          ; done twice                         17
.L15DD   ld    (hl),0         ; zero                               10
         inc   hl             ;                                     6
         djnz  L15DD          ; until done                       8/13
         ret                  ;                                    10

; load block

.L15E3   push  de             ;                                    11
         push  hl             ;                                    11
.L15E5   call  L15A5          ; fetch a byte                       17
         jr    c,L15F2        ; SPACE key or loading error       7/12
         ld    (hl),a         ; store byte in buffer                7
         inc   hl             ;                                     6
         dec   e              ;                                     4
         jr    nz,L15E5       ; until half complete              7/12
         dec   d              ; then                                4
         jr    nz,L15E5       ; until block complete             7/12
.L15F2   pop   hl             ;                                    10
         pop   de             ;                                    10
         ret                  ;                                    10

; discard block

.L15F5   ld    b,0            ; loop counter                        7
         call  L15FA          ; done twice                         17
.L15FA   call  L15A5          ; fetch and discard byte             17
         djnz  L15FA          ; until done                       8/13
         ret                  ;                                    10

; ------ LOAD  ( n -- )  load block n from cassette port

.L1600   defb  $04
         defm  "LOAD"
         defw  L1654-L1600

         ld    hl,DFLG
         bit   $06,(hl)       ; edit mode ?
         ret   z              ; is off, ignore command
         rst   $10            ; pop HL from IY stack
         ld    (PAGE),hl      ; block n to load
.L1611   out   ($FD),a        ; NMI generator off
         ld    hl,(BLKB)      ; block base address
         push  hl             ; preserve
         call  L15D8          ; clear block buffer
         call  L1592          ; look for ident
         jr    c,L1651        ; SPACE pressed, abort LOAD        7/12
         call  L15A5          ; get block number low byte          17
         ld    l,a            ;                                     4
         call  L15A5          ; get block number high byte         17
         ld    h,a            ;                                     4
         or    l              ; block zero ?                        4
         jr    nz,L162D       ; no                               7/12
         ld    (PAGE),hl      ; else usurp required block          20
.L162D   ex    de,hl          ; retain block number on tape         4
         ld    hl,(PAGE)      ; required block number              20
         ld    a,l            ; is it                               4
         or    h              ; zero (load next on tape) ?          4
         jr    nz,L163B       ; no, check against required       7/12
         ex    de,hl          ; block number on tape                4
         ld    (PAGE),hl      ; to be loaded                       20
         jr    L163D          ;                                    12

.L163B   sbc   hl,de          ; do block numbers concur ?          15
.L163D   pop   hl             ; retrieve block base address        10
         jr    nz,L164C       ; not required block               7/12
         ld    de,512         ; block size                         10
         call  L15E3          ; load block                         17
         call  L168C          ; auto-compile if CON activated
.L1649   out   ($FE),a        ; NMI generator on
         ret

.L164C   call  L15F5          ; discard block                      17
         jr    L1611          ; try next block

.L1651   pop   hl             ; discard block base address
         jr    L1649          ; and abort

; ------ -->  ( -- )  load next block number PAGE+1 or 0

.L1654   defb  $03
         defm  "-->"
         defw  L165F-L1654

         call  L166B          ; to next block
         jr    L1611          ; load it

; ------ <--  ( -- )  save as block number PAGE+1 or 0

.L165F   defb  $03
         defm  "<--"
         defw  L1684-L165F

         call  L166B          ; to next block
         jp    L157B          ; save it

.L166B   ld    hl,(PAGE)      ; current block number
         ld    a,l
         or    h
         ret   z              ; is zero
         inc   hl             ; else next block
         ld    (PAGE),hl      ; update
         ret

; signed comparison, return status of HL-DE

.L1676   push  hl
         push  de
         ld    a,h
         xor   d              ; (also clears carry)
         bit   $07,a          ; sign bits are
         jr    z,L167F        ; both same
         ex    de,hl          ; else differ, swap operands
.L167F   sbc   hl,de          ; set flags accordingly
         pop   de
         pop   hl
         ret

; ------ CPL  ( -- )  force compilation of edit window

.L1684   defb  $03
         defm  "CPL"
         defw  L1698-L1684

         jr    L1692

.L168C   ld    hl,SFLG        ; system flags
         bit   $02,(hl)       ; CON / COFF ?
         ret   z              ; is off
.L1692   ld    hl,SFLG        ; system flags
         set   $01,(hl)       ; do compilation
         ret

; ------ CON  ( -- )  activate compilation on LOAD

.L1698   defb  $03
         defm  "CON"
         defw  L16A4-L1698

         ld    hl,SFLG        ; system flags
         set   $02,(hl)       ; turn on
         ret

; ------ COFF  ( -- )  deactivate compilation on LOAD

.L16A4   defb  $04
         defm  "COFF"
         defw  L16B1-L16A4

         ld    hl,SFLG        ; system flags
         res   $02,(hl)       ; turn off
         ret

; ------ >  ( n1 n2 -- t/f )  greater than n1 > n2 ?

.L16B1   defb  $01
         defm  ">"
         defw  L16C2-L16B1

         rst   $10            ; pop HL from IY stack
         ex    de,hl          ; n2
         rst   $10            ; pop HL from IY stack
         ex    de,hl          ; n1,n2
.L16B9   call  L1676          ; signed comparison HL-DE
.L16BC   ccf                  ; carry is clear if true
.L16BD   sbc   hl,hl
         inc   hl             ; true = $01, false = $00
         rst   $08            ; push HL onto IY stack
         ret

; ------ <  ( n1 n2 -- t/f )  less than n1 < n2 ?

.L16C2   defb  $01
         defm  "<"
         defw  L16CB-L16C2

         rst   $10            ; pop HL from IY stack
         ex    de,hl          ; n2
         rst   $10            ; pop HL from IY stack
         jr    L16B9          ; n2,n1

; ------ =  ( n1 n2 -- t/f )  equals n1 = n2 ?

.L16CB   defb  $01
         defm  "="
         defw  L16DA-L16CB

         rst   $10            ; pop HL from IY stack
         ex    de,hl          ; n2
         rst   $10            ; pop HL from IY stack
         call  L1676          ; signed comparison HL-DE
.L16D5   jr    z,L16BD        ; true flag to IY stack
         or    a              ; clear carry
         jr    L16BC          ; false flag to IY stack

; ------ C=  ( c1 c2 -- t/f )  characters same ?

.L16DA   defb  $02
         defm  "C="
         defw  L16E5-L16DA

         rst   $10            ; pop HL from IY stack
         ld    a,l            ; c2
         rst   $10            ; pop HL from IY stack
         cp    l              ; result to zero flag
         jr    L16D5          ; true/false flag to IY stack

; ------ ?  ( addr -- )  emit word contents of address

.L16E5   defb  $01
         defm  "?"
         defw  L16EF-L16E5

         call  LEEB           ; perform @
         jp    LE7B           ; perform .

; ------ +!  ( n addr -- ) add n to word contents of address

.L16EF   defb  $02
         defm  "+!"
         defw  L1701-L16EF

         rst   $10            ; pop HL from IY stack
         push  hl             ; addr
         ld    e,(hl)
         inc   hl
         ld    d,(hl)         ; (addr)
         rst   $10            ; pop HL from IY stack
         add   hl,de          ; new value
         ex    de,hl
         pop   hl             ; addr
         ld    (hl),e
         inc   hl
         ld    (hl),d         ; update (addr)
         ret

; ------ +-  ( n1 n2 -- n3 )  apply sign of n2 to n1, result as n3

.L1701   defb  $02
         defm  "+-"
         defw  L1715-L1701

         rst   $10            ; pop HL from IY stack
         bit   $07,h          ; sign of n2
         push  af             ; preserve as flag
         rst   $10            ; pop HL from IY stack
         call  LB19           ; make HL absolute
         pop   af             ; retrieve sign of n2
         ret   z              ; was positive, done (error)   *** nop           !!!
         call  LB1E           ; two's complement HL          *** call nz,LB1E  !!!
         rst   $08            ; push HL onto IY stack
         ret

; ------ ,  ( n -- )  store n at next dictionary location

.L1715   defb  $01
         defm  ","
         defw  L171D-L1715

         rst   $10            ; pop HL from IY stack
         jp    L7C4           ; put <HL> into dictionary

; ------ C,  ( c -- )  store character at next dictionary location

.L171D   defb  $02
         defm  "C,"
         defw  L1727-L171D

         rst   $10            ; pop HL from IY stack
         ld    a,l            ; character
         jp    L7B9           ; put <A> into dictionary

; ------ SP@  ( -- addr )  return current IY stack pointer address

.L1727   defb  $03
         defm  "SP@"
         defw  L1732-L1727

         push  iy
         pop   hl             ; current stack pointer
         rst   $08            ; push HL onto IY stack
         ret

; ------ VLIST  ( -- )  list dictionary

.L1732   defb  $05
         defm  "VLIST"
         defw  L1751-L1732

         ld    hl,(DICT)      ; pointer to latest dictionary entry
         ld    de,8           ; print field width
.L1740   call  L637           ; copy string to CBUF, return count on IY stack
         call  L1758          ; pad with leading spaces to field width
         call  L795           ; update HL to point to next entry in dictionary
         call  L652           ; emit string from CBUF, count on IY stack
         xor   a              ; zero
         cp    h              ; test value
         jr    nz,L1740       ; until end of dictionary
         ret

; ------ W>  ( n1 n2 -- n3 )  put n2-n1 spaces in CBUF and return n2 or n1 if >n2

.L1751   defb  $02
         defm  "W>"
         defw  L1771-L1751

         rst   $10            ; pop HL from IY stack
         ex    de,hl          ; n2
.L1758   push  hl
         push  de
         rst   $10            ; pop HL from IY stack
         call  L1676          ; signed comparison HL-DE
         jr    nc,L176E       ; n1 >= n2 (no action)
         ex    de,hl          ; n1,n2
         rst   $08            ; push HL onto IY stack
         ccf                  ; clear carry flag
         sbc   hl,de          ; n2-n1
         ld    a,' '          ; fill with space/s
.L1767   rst   $18            ; put character into CBUF
         dec   l
         jr    nz,L1767       ; until done
.L176B   pop   de
         pop   hl
         ret

.L176E   rst   $08            ; push HL onto IY stack
         jr    L176B

; ------ IMM  ( -- )  make latest dictionary entry an immediate command

.L1771   defb  $83
         defm  "IMM"
         defw  L177D-L1771

.L1777   ld    hl,(XHERE)       ; latest accepted dictionary entry header
         set   $07,(hl)         ; make an immediate command
         ret

; ------ INTEGER <name>  ( n -- )  create <name> with value n   (aka VALUE)
;                <name>  ( -- n )  return value

.L177D   defb  $07
         defm  "INTEGER"
         defw  L17B5-L177D

         call  L866           ; set up new dictionary entry with <name>
         ld    hl,L1797       ; execution vector
         call  L7D3           ; put CALL <HL> into dictionary
         rst   $10            ; pop HL from IY stack
         call  L7C4           ; put <HL> into dictionary
         jp    L1777          ; perform IMM

; entry point at execution, return value or reassign if preceded with TO

.L1797   ld    hl,SFLG        ; system flags
         bit   $05,(hl)       ; preceded by TO ?
         res   $05,(hl)       ; dealt with
         pop   hl             ; address of value
         jr    z,L17AB        ; not prececed by TO
         ld    a,$D7          ; opcode RST $10 (pop HL from IY)
         call  L7B9           ; put <A> into dictionary
         ld    a,$22          ; opcode LD (nn),HL
         jp    L7D5           ; put <A> and <HL> into dictionary

.L17AB   ld    a,$2A          ; opcode LD HL,(nn)
         call  L7D5           ; put <A> and <HL> into dictionary
         ld    a,$CF          ; opcode RST $08 (push HL to IY)
         jp    L7B9           ; put <A> into dictionary

; ------ TO <word>  ( n -- )  assign new value n to <word> (defined by INTEGER)

.L17B5   defb  $82
         defm  "TO"
         defw  L17C0-L17B5

         ld    hl,SFLG        ; system flags
         set   $05,(hl)       ; <word> preceded by TO
         ret

; ------ BEGIN  ( -- )  start of BEGIN construct

.L17C0   defb  $85
         defm  "BEGIN"
         defw  L17CF-L17C0

         ld    hl,(HERE)
         push  hl             ; BEGIN again here
         call  L881           ; parse input loop during construct

; ------ AGAIN  ( -- )  end of BEGIN ... AGAIN construct

.L17CF   defb  $85
         defm  "AGAIN"
         defw  L17DE-L17CF

         pop   hl             ; discard return to parse routine
         pop   hl             ;    "    error routine
         pop   hl             ;    "    return to BEGIN
         pop   hl             ; address of BEGIN
         jp    L7CF           ; put JP <HL> into dictionary

; ------ WHILE  ( t/f -- )  intermediate in BEGIN ... WHILE ... REPEAT construct

.L17DE   defb  $85
         defm  "WHILE"
         defw  L17FC-L17DE


         pop   de             ; retain return to parse routine
         pop   hl             ; discard error routine
         pop   hl             ;    "    return to BEGIN
         ld    hl,L182B       ; execution vector (test flag on IY)
         call  L7D3           ; put CALL <HL> into dictionary
         ld    hl,(HERE)
         ld    a,$C2          ; opcode JP NZ,nn
         call  L7D5           ; put <A> and <HL> into dictionary
         inc   hl             ; point to nn in JP NZ,nn
         push  hl             ; for REPEAT to fill
         push  hl             ; dummy error routine
         ex    de,hl          ; discarded when
         jp    (hl)           ; back to parse construct routine

; ------ REPEAT  ( -- )  end of BEGIN ... WHILE ... REPEAT construct

.L17FC   defb  $86
         defm  "REPEAT"
         defw  L1814-L17FC

         pop   hl             ; discard return to parse routine
         pop   hl             ;    "    error routine
         pop   de             ; pointer to nn of JP NZ, nn at WHILE
         pop   hl             ; address of BEGIN
         call  L7CF           ; put JP <HL> into dictionary
         ld    hl,(HERE)
         ex    de,hl
         ld    (hl),e
         inc   hl
         ld    (hl),d         ; set nn in JP NZ,nn at WHILE
         ret

; ------ UNTIL  ( t/f -- )  end of BEGIN ... UNTIL construct

.L1814   defb  $85
         defm  "UNTIL"
         defw  L1831-L1814

         pop   hl             ; discard return to parse routine
         pop   hl             ;    "    error reoutine
         pop   hl             ;    "    return to BEGIN
         ld    hl,L182B       ; execution vector
         call  L7D3           ; put CALL <HL> into dictionary
         pop   hl             ; address of BEGIN
         ld    a,$CA          ; opcode JP Z,nn
         jp    L7D5           ; put <A> and <HL> into dictionary

; entry point at execution, return status of flag on IY sstcak

.L182B   rst   $10            ; pop HL from IY stack
         xor   a              ; zero
         cp    l              ; test flag
         ret   nz             ; false
         cp    h              ; test flag
         ret

; ------ KEY  ( -- n )  get ascii input

.L1831   defb  $03
         defm  "KEY"
         defw  L184A-L1831

.L1837   ld    a,$1F          ; control code for get key input
         call  L493           ; emit character in A register
         call  L621           ; fetch character from input stream
         and   $7F            ; (clear idle state tag)
         cp    $1F            ; control code for get key input ?
         jr    z,L1837        ; reject and retry
         ld    h,0
         ld    l,a            ; ascii
         rst   $08            ; push HL onto IY stack
         ret

; ------ (  ( -- )  comment, terminate with )

.L184A   defb  $81
         defm  "("
         defw  L1858-L184A

.L184E   call  L14C5          ; fetch character from input stream
         and   $7F
         cp    ')'
         jr    nz,L184E       ; until comment terminator found
         ret

; ------ 0<  ( n -- t/f )  less than zero ?

.L1858   defb  $02
         defm  "0<"
         defw  L1863-L1858

         rst   $10            ; pop HL from IY stack
         rlc   h              ; sign bit to carry
         jp    L16BC          ; return t/f accordingly

; ------ 0>  ( n -- t/f )  greater than zero ? (or equal)

.L1863   defb  $02
         defm  "0>"
         defw  L186E-L1863

         rst   $10            ; pop HL from IY stack
         rlc   h              ; sign bit to carry
.L186B   jp    L16BD          ; return t/f accordingly

; ------ 0=  ( n -- t/f )  equal to zero ?

.L186E   defb  $02
         defm  "0="
         defw  L187B-L186E

         rst   $10            ; pop HL from IY stack
         xor   a              ; zero
         cp    l              ; test
         jr    c,L186B        ; return false
         cp    h              ; test
         jr    L186B          ; return t/f accordingly

; ------ 2*  ( n1 -- n2 )  twice value of n1

.L187B   defb  $02
         defm  "2*"
         defw  L1884-L187B

         rst   $10            ; pop HL from IY stack
         add   hl,hl          ; 2*
         rst   $08            ; push HL onto IY stack
         ret

; ------ 2/  ( n1 -- n2 )  halve value of n1 (quotient)

.L1884   defb  $02
         defm  "2/"
         defw  L1890-L1884

         rst   $10            ; pop HL from IY stack
         sra   h              ; b0 to carry, maintain sign
         rr    l              ; carry to b7, discard b0
         rst   $08            ; push HL onto IY stack
         ret

; ------ W!  ( c addr -- )  extract c characters from CBUF to address as counted string

.L1890   defb  $02
         defm  "W!"
         defw  L1899-L1890

         rst   $10            ; pop HL from IY stack
         jp    L687           ; extract from CBUF to HL as counted string

; ------ EOFF  ( -- )  turn edit mode off

.L1899   defb  $04
         defm  "EOFF"
         defw  L18AF-L1899

         ld    hl,DFLG
         res   $06,(hl)       ; edit mode off
         ld    l,$B9          ; command window row origin
         ld    a,(hl)         ; get copy of then
         ld    (hl),$00       ; reset to top of display
         ld    l,$B7          ; command window cursor row
         add   a,(hl)         ; is now redefined
         ld    (hl),a         ; within larger command window
.L18AE   ret

; ------ BASE  ( -- un )  return current base number (or reassign with TO)

.L18AF   defb  $84
         defm  "BASE"
         defw  L18BD-L18AF

         ld    hl,BASE        ; address of value
         push  hl
         jp    L1797          ; return value or reassign if preceded with TO

; ------ DECIMAL  ( -- )  set BASE to ten

.L18BD   defb  $07
         defm  "DECIMAL"
         defw  L18CD-L18BD

         ld    hl,BASE        ; base is now
         ld    (hl),10        ; decimal
         ret

; ------ BACK <word>  ( -- )  set new background task to be <word> (default NUL)

.L18CD   defb  $04
         defm  "BACK"
         defw  L18DB-L18CD

         call  L6E3           ; execution vector of next word to HL
         ld    (BTSK),hl      ; is background task
         ret

; ------ MEM  ( -- un )  return free memory

.L18DB   defb  $03
         defm  "MEM"
         defw  L1918-L18DB

         call  L18E6          ; free memory to HL
         rst   $08            ; push HL onto IY stack
         ret

; return HL as free memory = RAMB+RAMS-HERE+IY (IY due to use = -depth)

.L18E6   ld    hl,(HERE)      ; top of dictionary
         ex    de,hl
         ld    hl,(RAMS)      ; size of RAM
         push  iy             ; current stack level
         or    a              ; clear carry
         sbc   hl,de          ; RAMS - HERE
         pop   de
         add   hl,de          ; + IY
         ex    de,hl
         ld    hl,(RAMB)      ; base of RAM
         add   hl,de          ; + RAMB = free memory
         ret

; check for memory corruption ( HERE > IY or IY-HERE < 32 )

.L18FA   ld    hl,SFLG        ; system flags
         bit   $03,(hl)       ; previous memory test failed ?
         ret   nz             ; yes, error already reported
         call  L18E6          ; free memory to HL
         ld    de,-32
         add   hl,de          ; less a little headroom
         ex    de,hl
         ld    hl,(RAMS)      ; size of RAM (a true figure)
         sbc   hl,de          ; less free RAM (possibly false)
         ret   nc             ; ok, coherent result
         ld    hl,SFLG        ; system flags
         set   $03,(hl)       ; memory test has failed
         ld    a,'M'          ; error code, memory test fail
         jp    L6C9           ; print error report in A

; ------ FENCE <word>  ( -- )  set dictionary boundary for FORGET

.L1918   defb  $05
         defm  "FENCE"
         defw  L1936-L1918

         call  L1927
         ld    (FNCE),hl      ; set boundary point
         ret

.L1927   call  L782           ; next word found in dictionary ?
         push  af             ; preserve status
         call  LD6B           ; perform CDROP
         pop   af             ; was word found ?
         ret   nc             ; yes
         pop   hl             ; else discard return address
         ld    a,'U'          ; error code, undefined word
         jp    L6C9           ; print error report in A

; ------ FORGET <word>  ( -- )  remove <word> onwards from dictionary

.L1936   defb  $06
         defm  "FORGET"
         defw  L1981-L1936

         call  L1927          ; find word dictionary entry
         ex    de,hl          ; <word> address
         ld    hl,(FNCE)      ; fence boundary
         call  L12A2          ; perform SUB HL,DE
         ld    a,'F'          ; error code, fence violation
         jp    nc,L6C9        ; is above <word>, report error
         ld    hl,(BTSK)      ; background task
         call  L12A2          ; perform SUB HL,DE
         jr    c,L195C        ; is below <word>
         ld    hl,L19B4       ; else NUL is new
         ld    (BTSK),hl      ; background task

; scan linked list of tasks and remove any above <word> in dictionary

.L195C   ld    hl,(TSKS)      ; start of linked list of tasks
.L195F   push  hl             ; preserve copy of
         pop   bc             ; task entry being checked
         ld    (TSKE),hl      ; update end of linked list of tasks
         ld    a,(hl)
         inc   hl
         ld    h,(hl)
         ld    l,a            ; link to next task in list
         or    h              ; end of list ?
         jr    z,L1974        ; yes, no tasks found above <word>
         call  L12A2          ; perform SUB HL,DE
         jr    c,L195F        ; task entry is below <word>, continue
         xor   a              ; else zeroing
         ld    (bc),a         ; this task entry
         inc   bc             ; makes it the
         ld    (bc),a         ; new end of list

; update end of dictionary, effectively forgetting <word> onwards

.L1974   ex    de,hl
         ld    (XHERE),hl     ; update accepted end of dictionary
         call  L795           ; update HL to point to next entry in dictionary
         ld    (XDICT),hl     ; update accepted last entry in dictionary
         jp    L765           ; reset DICT and HERE to accepted counterparts

; ------ TOFF  ( -- )  turn off task protection from WARM rest

.L1981   defb  $04
         defm  "TOFF"
         defw  L198E-L1981

         ld    hl,SFLG        ; system flags
         res   $06,(hl)       ; turn off
         ret

; ------ TON  ( -- )  turn on task protection from WARM reset

.L198E   defb  $03
         defm  "TON"
         defw  L199A-L198E

         ld    hl,SFLG        ; system flags
         set   $06,(hl)       ; turn on
         ret

; ------ WARM  ( -- )  do warm reset

.L199A   defb  $04
         defm  "WARM"
         defw  L19A4-L199A

         jp    L99A           ; do WARM restart

; ------ COLD  ( -- )

.L19A4   defb  $04
         defm  "COLD"
         defw  L19AE-L19A4

         jp    L953           ; do COLD restart

; ------- NUL  ( -- )  do nothing (return)

.L19AE   defb  $03
         defm  "NUL"
         defw  L19B5-L19AE

.L19B4   ret

; ------ HERE  ( -- addr )  return address of end of dictionary

.L19B5   defb  $04
         defm  "HERE"
         defw  L19C1-L19B5

         ld    hl,(HERE)      ; end of dictionary
         rst   $08            ; push HL onto IY stack
         ret

; ------ H  ( -- addr )  return address holding HERE pointer

.L19C1   defb  $01
         defm  "H"
         defw  L19CA-L19C1

         ld    hl,HERE        ; address of pointer
         rst   $08            ; push HL onto IY stack
         ret

; ------ T  ( -- addr )   return address holding start of dictionary pointer

.L19CA   defb  $01
         defm  "T"
         defw  L19D3-L19CA

         ld    hl,DICT        ; pointer to latest dictionary entry
         rst   $08            ; push HL onto IY stack
         ret

; ------ S@ <name>  ( -- c )  copy <name> to CBUF, return count

.L19D3   defb  $02
         defm  "S@"
         defw  L19DB-L19D3

         jp    L6F1           ; copy next word from input to CBUF

; ------ >#  ( c -- n t | c f )  return true if CBUF numeric and value on stack
;

.L19DB   defb  $02
         defm  ">#"
         defw  L19E6-L19DB

         call  LA36           ; convert CBUF to number on IY stack
         jp    L16BD          ; return t/f accordingly

; ------ .CO  ( c -- )  stuff c characters from CBUF into TIB

.L19E6   defb  $03
         defm  ".CO"
         defw  L19F8-L19E6

         rst   $10            ; pop HL from IY stack
         xor   a              ; zero
         or    l              ; test counter value
         ret   z              ; nothing to do
.L19F0   rst   $20            ; get character from CBUF
         call  L8F8           ; put character into TIB
         dec   l
         jr    nz,L19F0       ; until done
         ret

; ------ .CPU  ( -- )  emit system/cpu string

.L19F8   defb  $04
         defm  ".CPU"
         defw  L1A05-L19F8

         ld    hl,L1D6E       ; string 'ZX-Z80'
         jp    L665           ; print string pointed to by HL

; ------ U*  ( un1 un2 -- ud )  unsigned multiply to double

.L1A05   defb  $02
         defm  "U*"
         defw  L1A13-L1A05

         call  LB9B           ; pop HL then DE from IY stack
         call  LB3E           ; unsigned multiplication DE * HL = DEHL
         jp    LB39           ; push DE then HL onto IY stack

; ------ U/MOD  ( ud un1 -- un2 un3 )  unsigned ud / un1, quotient un2 remainder un3

.L1A13   defb  $05
         defm  "U/MOD"
         defw  L1A27-L1A13

.L1A1B   rst   $10            ; pop HL from IY stack
         push  hl
         pop   bc             ; un1
         call  LB9B           ; pop HL then DE from IY stack
         call  LBA0           ; perform ud / un
         jp    LB39           ; push DE then HL onto IY stack

; ------ UMOD  ( ud un1 -- un2 )  unsigned ud / un1, remainder un2

.L1A27   defb  $04
         defm  "UMOD"
         defw  L1A36-L1A27

         call  L1A1B          ; perform U/MOD
         call  LE41           ; perform SWAP
         rst   $10            ; pop HL from IY stack
         ret

; ------ U<  ( u1 u2 -- t/f )  unsigned less than u1 < u2 ? (or equal)

.L1A36   defb  $02
         defm  "U<"
         defw  L1A44-L1A36

         call  LB9B           ; pop HL then DE from IY stack
         or    a              ; clear carry
         sbc   hl,de          ; u2 - u1
         jp    L16BD          ; return t/f accordingly

; ------ MIN  ( n1 n2 -- n3 )  return minimum

.L1A44   defb  $03
         defm  "MIN"
         defw  L1A55-L1A44

         call  LB9B           ; pop HL then DE from IY stack
         call  L1676          ; signed comparison HL-DE
.L1A50   jr    c,L1A53        ; carry if n1 > n2
         ex    de,hl
.L1A53   rst   $08            ; push HL onto IY stack
         ret

; ------ MAX  ( n1 n2 -- n3 )  return maximum

.L1A55   defb  $03
         defm  "MAX"
         defw  L1A64-L1A55

         call  LB9B           ; pop HL then DE from IY stack
         call  L1676          ; signed comparison HL-DE
         ccf                  ; invert carry to
         jr    L1A50          ; yield correct result

; ------ PAD  ( -- addr )  return address of scratchpad

.L1A64   defb  $03
         defm  "PAD"
         defw  L1A6F-L1A64

         ld    hl,PAD         ; location of PAD
         rst   $08            ; push HL onto IY stack
         ret

; ------ U#  ( un -- c )  convert un to numeric string in CBUF of c characters

.L1A6F   defb  $02
         defm  "U#"
         defw  L1A7F-L1A6F

.L1A74   rst   $10            ; pop HL from IY stack
         ld    de,0           ; unsigned
         ex    de,hl
         rst   $08            ; push HL onto IY stack
         ex    de,hl          ; make un an unsigned double value
         rst   $08            ; push HL onto IY stack
         jp    LA96           ; convert double to string in CBUF

; ------ U.  ( un -- )  emit un as numeric string

.L1A7F   defb  $02
         defm   "U."
         defw  L1A8A-L1A7F

         call  L1A74          ; perform U#
         jp    L652           ; emit string from CBUF, count on IY stack

; ------ D=  ( d1 d2 -- t/f )  doubles equal ?

.L1A8A   defb  $02
         defm  "D="
         defw  L1A9D-L1A8A

         call  L12C0          ; compare doubles on IY stack
         rst   $10            ; pop HL from IY stack
         rst   $10            ; pop HL from IY stack
         rst   $10            ; pop HL from IY stack
         rst   $10            ; pop HL from IY stack
.L1A96   scf                  ; carry set makes
         jr    nz,L1A9A       ; false
         ccf                  ; carry clear makes true
.L1A9A   jp    L16BD          ; return t/f accordingly

; ------ D0=  ( d -- t/f )  double equals zero ?

.L1A9D   defb  $03
         defm  "D0="
         defw  L1AAB-L1A9D

         call  LB9B           ; pop HL then DE from IY stack
         call  LD93           ; double DEHL zero ?
         jr    L1A96          ; return t/f accordingly

; ------ DMIN  ( d1 d2 -- d3 )  double minimum

.L1AAB   defb  $04
         defm  "DMIN"
         defw  L1ABB-L1AAB

         call  L12C0          ; compare doubles on IY stack
         call  nc,LDA3        ; if d1 > d2 perform DSWAP
         rst   $10            ; pop HL from IY stack
         rst   $10            ; pop HL from IY stack
         ret

; ------ DMAX  ( d1 d2 -- d3 ) double maximum

.L1ABB   defb  $04
         defm  "DMAX"
         defw  L1ACB-L1ABB

         call  L12C0          ; compare doubles on IY stack
         call  c,LDA3         ; if d1 < d2 perform DSWAP
         rst   $10            ; pop HL from IY stack
         rst   $10            ; pop HL from IY stack
         ret

; ------ W=  ( addr1 addr2 -- t/f )  compare counted strings

.L1ACB   defb  $02
         defm  "W="
         defw  L1AD5-L1ACB

         call  L69D           ; compare strings, Z flag set accordingly
         jr    L1A96          ; return t/f result

; ------ S=  ( c addr -- c t/f )  compare counted string with CBUF

.L1AD5   defb  $02
         defm  "S="
         defw  L1AE0-L1AD5

         rst   $10            ; pop HL from IY stack
         call  L6A7           ; compare string with CBUF
.L1ADE   jr    L1A96          ; return t/f result

; ------ TIME  ( -- addr )  return address of time counter double

.L1AE0   defb  $04
         defm  "TIME"
         defw  L1AEC-L1AE0

         ld    hl,TIME        ; address of counter double
         rst   $08            ; push HL onto IY stack
         ret

; ------ PER  ( -- addr )  return address of period for TIME before wrap-around

.L1AEC   defb  $03
         defm  "PER"
         defw  L1AF7-L1AEC

         ld    hl,PER         ; address of period double
         rst   $08            ; push HL onto IY stack
         ret

; ------ +ORG ( un -- addr )  return address of offset into origin parameter area

.L1AF7   defb  $04
         defm  "+ORG"
         defw  L1B05-L1AF7

         ld    de,ORGB        ; base address of origin parameter area
         rst   $10            ; pop HL from IY stack
         add   hl,de          ; un
         rst   $08            ; push HL onto IY stack
         ret

; ------ BLK  ( -- addr )  return block base address (or reassign with TO)

.L1B05   defb  $83
         defm  "BLK"
         defw  L1B12-L1B05

         ld    hl,BLKB        ; holds block base address
.L1B0E   push  hl
         jp    L1797          ; return value or reassign if preceded with TO

; ------ PAGE ( -- un )  return current block number (or reassign with TO)

.L1B12   defb  $84
         defm  "PAGE"
         defw  L1B1E-L1B12

         ld    hl,PAGE        ; holds current block number
         jr    L1B0E          ; return value or reassign if preceded with TO

; ------ [_] <word>  ( -- )  put call to <word> into dictionary

.L1B1E   defb  $83
         defm  "[_]"
         defw  L1B2A-L1B1E

         call  L6E3           ; execution vector of next word to HL
         jp    L7D3           ; put CALL <HL> into dictionary

; ------ H>A  ( un -- c )  convert hex value (0-15) to ascii (0-F)

.L1B2A   defb  $03
         defm  "H>A"
         defw  L1B3A-L1B2A

         rst   $10            ; pop HL from IY stack
         ld    a,l            ; legal is 0-15
         call  L91E           ; convert 0-15 to ascii hex character
         ld    h,0
         ld    l,a            ; ascii code
         rst   $08            ; push HL onto IY stack
         ret

; ------ A>H  ( c -- un )  convert ascii (0-F, a-f) to hex value (0-15)

.L1B3A   defb  $03
         defm  "A>H"
         defw  L1B4A-L1B3A

         rst   $10            ; pop HL from IY stack
         ld    a,l            ; ascii code
         call  L677           ; convert hex ascii character 0-F, a-f to 0-15
         ld    h,0
         ld    l,a            ; hex value
         rst   $08            ; push HL onto IY stack
         ret

; ------ HEAD <name>  ( -- )  create dictionary entry with <name>   (aka CREATE)

.L1B4A   defb  $04
         defm  "HEAD"
         defw  L1B54-L1B4A

         jp    L866           ; set up new dictionary entry with <name>

; ------ INIT <word>  ( -- )  initialise RAM at $2000 with <word> as boot routine

.L1B54   defb  $04
         defm  "INIT"
         defw  L1B67-L1B54

         call  L6E3           ; execution vector of next word to HL
         ld    (RBOOT),hl     ; boot routine
         ld    hl,RHEAD
         ld    (hl),$A4       ; valid ident (to call boot routine)
         ret

; ------ UPDATE  ( -- )  update system pointers in RAM at $2000

.L1B67   defb  $06
         defm  "UPDATE"
         defw  L1B83-L1B67

.L1B70   ld    hl,(HERE)      ; pointer to end of dictionary
         ld    (RHERE),hl
         ld    hl,(DICT)      ; pointer to latest dictionary entry
         ld    (RDICT),hl
         ld    hl,(TSKI)      ; internal task
         ld    (RTSKL),hl
         ret

; ------ FBUF  ( -- addr )  return display base address (don't reassign with TO)

.L1B83   defb  $84
         defm  "FBUF"
         defw  L1B90-L1B83

         ld    hl,DISB        ; display base address
         jp    L1B0E          ; return value or reassign if preceded with TO

; ------ PASS  ( -- )  warm reset without resetting SP, CBUF or TIB

.L1B90   defb  $04
         defm  "PASS"
         defw  L1B9A-L1B90

         jp    L9A9           ; partial warm reset

; ------ OVER  ( n1 n2 -- n1 n2 n1 )  copy NOS over TOS

.L1B9A   defb  $04
         defm  "OVER"
         defw  L1BAA-L1B9A

         rst   $10            ; pop HL from IY stack
         ex    de,hl          ; n2
         rst   $10            ; pop HL from IY stack
         rst   $08            ; push HL onto IY stack
         ex    de,hl          ; n1,n2
         rst   $08            ; push HL onto IY stack
         ex    de,hl          ; n1
         rst   $08            ; push HL onto IY stack
         ret

; ------ PRINT <word> ( -- )  set printer routine <word> (default PRTR)

.L1BAA   defb  $05
         defm  "PRINT"
         defw  L1BB9-L1BAA

         call  L6E3           ; execution vector of next word to HL
         ld    (LPTR),hl      ; update printer routine used when P is on
         ret

; ------ P  ( -- )  toggle printer output on/off

.L1BB9   defb  $01
         defm  "P"
         defw  L1BC8-L1BB9

         ld    hl,PFLG        ; print flags
         bit   $00,(hl)       ; printing on/off ?
         res   $00,(hl)       ; assume off
         ret   nz             ; was on, done
         set   $00,(hl)       ; else turn on
         ret

; ------ PRTR ( c -- )  put character in printer line buffer (PAD), print when full.
;                       Backspace $08 deletes last character in buffer, CR is ignored,
;                       LF or buffer full (32 characters) sends buffer to printer,
;                       other control characters printed as . (dot)
;                       Note: assumes PAD is treated as counted string

.L1BC8   defb  $04
         defm  "PRTR"
         defw  L1C04-L1BC8

.L1BCF   push  hl
         rst   $10            ; pop HL from IY stack
         ld    a,l            ; character
         ld    hl,PAD         ; printer line buffer
         and   $7F            ; limit to ascii
         cp    $60            ; upper case ?
         jr    c,L1BDD        ; yes
         add   a,-$20         ; else make upper case
.L1BDD   cp    $20            ; printable character ?
         jr    nc,L1BF6       ; yes
         cp    $08            ; backspace ?
         jr    nz,L1BEC       ; no
         xor   a              ; zero
         or    (hl)           ; count ?
         jr    z,L1BEA        ; zero, nothing to delete
         dec   (hl)           ; decrement byte count
.L1BEA   pop   hl
         ret

.L1BEC   cp    $0D            ; CR ?
         jr    z,L1BEA        ; yes, no action
         cp    $0A            ; LF ?
         ld    a,'.'          ; default for other control characters
         jr    z,L1BFD        ; on LF send line buffer to printer

; printable character

.L1BF6   call  L737           ; put character into PAD
         bit   $05,(hl)       ; 32 characters done ?
         jr    z,L1C02        ; not yet
.L1BFD   call  L1C09          ; perform .P
         ld    (hl),$00       ; reset PAD count byte
.L1C02   pop   hl
         ret

; ------ .P  ( -- )  send print line buffer (PAD) to ZX printer

.L1C04   defb  $02
         defm  ".P"
         defw  L1C90-L1C04

.L1C09   xor   a              ; zero
         out   ($FD),a        ; NMI generator off
         out   ($FB),a        ; printer motor on (b2=0), print bit off (b7=0)
         push  de
         push  hl
         ld    hl,PAD         ; line buffer
         ld    a,31
         sub   (hl)           ; full line ?
         jr    c,L1C26        ; yes
         push  hl             ; preserve PAD pointer
         inc   a
         ld    e,a            ; trailing characters to fill line
         ld    a,(hl)         ; characters in buffer
         inc   hl
         add   a,l
         ld    l,a            ; point past last character
.L1C1F   ld    (hl),' '       ; fill with spaces
         inc   hl
         dec   e
         jr    nz,L1C1F       ; until buffer full
         pop   hl             ; retrieve PAD pointer
.L1C26   inc   hl             ; skip count byte
         ex    de,hl          ; pointer to string
         ld    hl,(IYSP)      ; pointer to raster buffer
         call  L1C37          ; output to printer
         pop   hl
         pop   de
         ld    a,$04
         out   ($FB),a        ; printer motor off (b2=1), print bit off (b7=0)
         out   ($FE),a        ; NMI generator on
         ret

; DE points to 32 character string, HL points to raster buffer

.L1C37   push  bc
         ld    c,0            ; bitmap line index
.L1C3A   push  de
         push  hl             ; preserve raster buffer pointer
         ld    b,32           ; bytes per row
.L1C3E   push  hl             ; position in raster buffer
         ld    h,0
         ld    a,(de)         ; character ascii code
         sub   $20            ; convert to
         and   $3F            ; internal representation
         ld    l,a            ; use as index
         add   hl,hl
         add   hl,hl
         add   hl,hl          ; each entry is 8 bytes
         ld    a,i            ; $1E high address of rom bitmaps
         or    h
         ld    h,a            ; point to character bitmap in ROM
         ld    a,c            ; bitmap line index
         or    l
         ld    l,a
         ld    a,(hl)         ; get bitmap for character line
         pop   hl             ; position in raster buffer
         ld    (hl),a         ; add bitmap line to raster
         inc   hl             ; update raster buffer pointer
         inc   de             ; point to next character
         djnz  L1C3E          ; until raster line complete

         pop   hl             ; start of raster buffer
         ld    a,c            ; bitmap line index
         cp    $06            ; penultimate line ?
         sbc   a,a            ; $FF no, $00 yes
         inc   a              ; $00 no, $01 yes
         add   a,a            ; $00 no, $02 yes
         ld    e,a            ; slow motor control for ZX printer
         call  L1C6B          ; send raster buffer to printer
         pop   de             ; start of character string
         inc   c              ; next bitmap line index
         bit   3,c            ; completed 8 rows ?
         jr    z,L1C3A        ; not yet
         pop   bc
         ret

; send raster buffer to printer

.L1C6B   push  hl
         push  de
         push  bc
.L1C6E   in    a,($FB)        ; get printer status
         rlca                 ; busy ?
         jr    nc,L1C6E       ; yes, wait
         ld    c,32           ; characters to process
.L1C75   ld    b,8            ; bits per character
         ld    d,(hl)         ; get character line bitmap
         inc   hl             ; update pointer
.L1C79   in    a,($FB)        ; get printer staus
         rrca                 ; busy ?
         jr    nc,L1C79       ; yes, wait
         rlc   d              ; bitmap pixel to carry
         ld    a,e            ; slow motor OFF rows 0-5, ON rows 6-7
         jr    nc,L1C85       ; pixel is unset
         or    $80            ; else set
.L1C85   out   ($FB),a        ; send it to printer
         djnz  L1C79          ; until 8 bits done
         dec   c              ; 32 characters done ?
         jr    nz,L1C75       ; not yet
         pop   bc
         pop   de
         pop   hl
         ret

; ------ .CN  ( c -- )  emit c characters from CBUF

.L1C90   defb  $03
         defm  ".CN"
         defw  L1C99-L1C90

         jp    L652           ; emit string from CBUF, count on IY stack

; ------ D>  ( d1 d2 -- t/f )  more than d1 > d2 ?

.L1C99   defb  $02
         defm  "D>"
         defw  L1CA3-L1C99

         call  LDA3           ; perform DSWAP
         jr    L1CA8          ; perform D<

; ------ D<  ( d1 d2 -- t/f )  less than d1 < d2 ?

.L1CA3   defb  $02
         defm  "D<"
         defw  0-L1CA3

.L1CA8   call  L12C0          ; compare doubles on IY stack
         rst   $10            ; pop HL from IY stack
         rst   $10            ; pop HL from IY stack
         rst   $10            ; pop HL from IY stack
         rst   $10            ; pop HL from IY stack
         ccf                  ; carry is clear if true
         jp    L16BD          ; return t/f accordingly


; =============================== LOOK-UP DATA  ===============================


.X1CB3   defs  13             ; spare zeroed bytes

; table of routines for unprintable ascii characters $00-$1F

.L1CC0   defw   L50, L297,  L50,  L50,  L50,  L50,  L50,  L50
         defw  L308, L2E3, L2C6, L2FF, L283, L29F,  L50,  L50
         defw   L50, L3EE, L41D, L441, L44B,  L50,  L50,  L50
         defw   L50,  L50, L3A3, L385, L3DA, L3BD, L455, L47C

.L1D00   defb  $33,$0C
         defm  "ZX81-FORTH BY DAVID HUSBAND"      ; *** "TREE-FORTH BY TREE SYSTEMS "      
         defb  $0D,$0A
         defm  "COPYRIGHT (C) 1983"
         defb  $0D,$0A,$0A

.L1D34   defb  $02,$0D,$0A,$20                    ; string for CR with LF

; ascii keycodes for keyboard matrix

.L1D38   defb  $00,'A','Q','1','0','P',$0D,' '    ; KBD0 - A8 to A15
         defb  'Z','S','W','2','9','O','L','.'    ; KBD1 -  "    "
         defb  'X','D','E','3','8','I','K','M'    ; KBD2 -  "    "
         defb  'C','F','R','4','7','U','J','N'    ; KBD3 -  "    "
         defb  'V','G','T','5','6','Y','H','B'    ; KBD4 -  "    "

.L1D60   defb  $05
         defm  " OK"
         defb  $0D,$0A

.L1D66   defb  $07
         defm  " ERROR "

.L1D6E   defb  $09,$0D,$0A
         defm  "ZX-Z80 "

; ascii keycodes for shifted keyboard matrix (at $1D78, +$40 on L1D38)

         defb  $00,$0C,$18,$1E,$1B,'"',$01,$80    ; sh  A  Q  1  0  P n/l sp
         defb  ':','%','!',$14,$1C,')','=',','    ;  Z  S  W  2  9  O  L  .
         defb  ';',$27,'@',$11,$09,'(','+','>'    ;  X  D  E  3  8  I  K  M
         defb  '?','\','[',$1A,$0B,'$','-','<'    ;  C  F  R  4  7  U  J  N
         defb  '/','^','_',$08,$0A,']','#','*'    ;  V  G  T  5  6  Y  H  B

;      _______________________________________________________________________________
;     |       |       |       |       |       |       |       |       |       |       |
;     | <RS>  | <DC4> | <DC1> | <SUB> | <BS>  | <LF>  | <VT>  | <HT>  | <FS>  | <ESC> |
;     |       |       |       |       |       |       |       |       |       |       |
;     |   1   |   2   |   3   |   4   |   5   |   6   |   7   |   8   |   9   |   0   |
;     |_______|_______|_______|_______|_______|_______|_______|_______|_______|_______|
;     |       |       |       |       |       |       |       |       |       |       |
;     | <CAN> |   !   |   @   |   [   |   _   |   ]   |   $   |   (   |   )   |   "   |
;     |       |       |       |       |       |       |       |       |       |       |
;     |   Q   |   W   |   E   |   R   |   T   |   Y   |   U   |   I   |   O   |   P   |
;     |_______|_______|_______|_______|_______|_______|_______|_______|_______|_______|
;     |       |       |       |       |       |       |       |       |       |       |
;     | <FF>  |   %   |   '   |   \   |   ^   |   #   |   -   |   +   |   =   | <SCH> |
;     |       |       |       |       |       |       |       |       |       |       |
;     |   A   |   S   |   D   |   F   |   G   |   H   |   J   |   K   |   L   |  n/l  |
;     |_______|_______|_______|_______|_______|_______|_______|_______|_______|_______|
;     |       |       |       |       |       |       |       |       |       |       |
;     | <NUL> |   :   |   ;   |   ?   |   /   |   *   |   <   |   >   |   ,   | <$80> |
;     |       |       |       |       |       |       |       |       |       |       |
;     |  sh   |   Z   |   X   |   C   |   V   |   B   |   N   |   M   |   .   |  sp   |
;     |_______|_______|_______|_______|_______|_______|_______|_______|_______|_______|



;================================= MEMORY MAP =================================

                           ; 4000 - FA7F  free memory

                           ; FA80         IY data stack

                           ; FA80 - FA9F  32 byte printer buffer

                           ; FAA0 - FB7F  224 bytes

                           ; FB80         SPTOP  machine stack

                           ; FB80 - FBBF  TIB 64 byte FIFO buffer

                           ; FBC0 - FC35  PAD 118 bytes (& printer line buffer)

                           ; FC36            SPIDT Idle SP top, pops...
                           ;                       IX/BC/DE/HL
                           ;                       return address LA1B

                           ; FC40         .b SYSI  if not $00 do COLD start

                           ; FC41 - FC4F  15 byte  safety buffer ?

                           ; FC50         .w ORGB  origin parameter area base
                           ; FC52         .w PAGE  current block number
                           ; FC54         .w ICTR  idle counter
                           ; FC56         .w RAMS  size of RAM

                           ; FC58         .l UCTR  user counter double
                           ; FC5C         .l TIME  double


; data at $1DA0-$1DFF copied to $FC60-$FCBF (effectively $7F60-7CBF)

.L1DA0   defl  0           ; FC60  UINC  added to UCTR each internal task
         defl  15          ; FC64  TINC  added to TIME each internal task
         defl  $004F1A00   ; FC68  PER   period of TIME before wrap-around

         defw  SRAM        ; FC6C  RAMB  bottom of RAM
         defw  DFILE       ; FC6E  DISB  display base address (FBUF value)
         defw  TSKI        ; FC70  TSKS  start of linked list of tasks
         defw  $2000       ; FC72  FNCE  dictionary boundary set by FENCE
         defw  L1BCF       ; FC74  LPTR  printer routine (default PRTR)
         defw  DFILE       ; FC76  BLKB  block base address (BLK value)
         defb  $FF,$00     ; FC78  SYSF  system set-up byte ($FF for valid)
         defw  LA36        ; FC7A  NMBR  parse as number routine
         defw  L50         ; FC7C  BTSK  background task (default NUL)
         defw  $8080       ; FC7E  TIBn  Output/Input indexes
         defw  L98         ; FC80  NACT  next action when NMI counter zeroed
         defw  L71         ; FC82  DISR  display routine (default SLOW)
         defw  CBUF        ; FC84  CBFP  current pointer
         defw  $000A       ; FC86  BASE  current base
         defw  LAFB        ; FC88  DICT  dictionary top entry
         defw  SRAM        ; FC8A  HERE  dictionary end
         defw  LAFB        ; FC8C  XDICT accepted dictionary top entry
         defW  SRAM        ; FC8E  XHERE accepted dictionary end
         defw  $FA80       ; FC90  IYSP  IY stack pointer (also printer buffer)
         defw  TSKI        ; FC92  TSKE  end of linked list of tasks
         defw  $0000       ; FC94  STSP  stack pointer store
         defw  $FC36       ; FC96  IDSP  idle stack pointer

         ; -------------------------- internal first task --------------------------

         defw  $0000       ; FC98  TSKI  first link in list of tasks
         defl  0,-15       ;             timeout and task period (negative frames)
         defb  $00         ;             action byte
         defw  L13BD       ;             execution vector

         ; -------------------------------------------------------------------------

                           ; flag bits marked . are cleared on a warm reset

         defb  $00         ; FCA5  SFLG flags    set / res
                           ;           . b0      yes - no       CR parsed
                           ;           . b1       on - off      force compilation
                           ;           . b2      CON - COFF     compilation on LOAD
                           ;           . b3     fail - ok       memory test
                           ;             b4      RAM - no RAM   at $2000
                           ;           . b5      yes - no       preceded by TO
                           ;             b6      TON - TOFF     tasks warm protection
                           ;             b7      off - on       AUTO display

         defw  L4CF        ; FCA6  INPB  process byte as input routine
         defw  L8F8        ; FCA8  PUTB  put byte into TIB routine

         defb  $03         ; FCAA  KEYD  debounce / auto-delay & repeat
         defb  $00         ; FCAB  KEYL  last key read
         defb  $80         ; FCAC  KEYN  new key event

         defb  $80         ; FCAD  DFLG flags    set / res
                           ;           . b0     edit - command   current window
                           ;           . b1     busy - free      edit window
                           ;           . b2     busy - free      command window
                           ;           . b3     busy - free      return character input
                           ;           . b4     idle - main      stack
                           ;             b5      yes - no        atomic stack change
                           ;             b6       on - off       edit mode
                           ;             b7      yes - no        get key invert cursor


                           ; FCAE  EDWP - Edit window parameters

         defb  $00,$00     ; cursor col/row
         defb  $00,$00     ; origin col/row
         defb  $1F,$0F     ; limit col/row
         defb  $00         ; XOR byte
         defb  $00         ; CLS byte

                           ; FCB6  COWP - Command window parameters

         defb  $00,$00     ; cursor col/row
         defb  $00,$00     ; origin col/row
         defb  $1F,$17     ; limit col/row
         defb  $00         ; XOR byte
         defb  $00         ; CLS byte

         defb  $00         ; FCBE  PFLG flags    set / res
                           ;           . b0      yes - no    emit to printer (P toggles)
                           ;           . b1       no - yes   emit to display

         defb  $0F         ; FCBF         CCNT count for string in CBUF

                           ; FCC0 - FCFF  CBUF 64 character LIFO buffer (reverse filled)

                           ; FD00 - FFFF  DFILE fixed display file 24x32


;============================== CHARACTER BIMAPS ==============================

; order  !"=$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\]^_

; character 0

         defb  @00000000
         defb  @00000000
         defb  @00000000
         defb  @00000000
         defb  @00000000
         defb  @00000000
         defb  @00000000
         defb  @00000000

; character 1          !

         defb  @00000000
         defb  @00001000
         defb  @00001000
         defb  @00001000
         defb  @00001000
         defb  @00000000
         defb  @00001000
         defb  @00000000

; character 2          "

         defb  @00000000
         defb  @00010010
         defb  @00010010
         defb  @00000000
         defb  @00000000
         defb  @00000000
         defb  @00000000
         defb  @00000000

; character 3          =

         defb  @00000000
         defb  @00100100
         defb  @01111110
         defb  @00100100
         defb  @00100100
         defb  @01111110
         defb  @00100100
         defb  @00000000

; character 4          $

         defb  @00000000
         defb  @00001000
         defb  @00111110
         defb  @01001000
         defb  @00111100
         defb  @00010010
         defb  @01111100
         defb  @00010000

; character 5          %

         defb  @00000000
         defb  @01100010
         defb  @01100100
         defb  @00001000
         defb  @00010000
         defb  @00100110
         defb  @01000110
         defb  @00000000

; character 6          &

         defb  @00000000
         defb  @00011000
         defb  @00100100
         defb  @00011000
         defb  @00101000
         defb  @01000100
         defb  @00111110
         defb  @00000010

; character 7          '

         defb  @00000000
         defb  @00000100
         defb  @00001000
         defb  @00000000
         defb  @00000000
         defb  @00000000
         defb  @00000000
         defb  @00000000

; character 8          (

         defb  @00000000
         defb  @00000100
         defb  @00001000
         defb  @00001000
         defb  @00001000
         defb  @00001000
         defb  @00000100
         defb  @00000000

; character 9          )

         defb  @00000000
         defb  @00100000
         defb  @00010000
         defb  @00010000
         defb  @00010000
         defb  @00010000
         defb  @00100000
         defb  @00000000

; character 10         *

         defb  @00000000
         defb  @00000000
         defb  @00001000
         defb  @00101010
         defb  @00011100
         defb  @00101010
         defb  @00001000
         defb  @00000000

; character 11         +

         defb  @00000000
         defb  @00000000
         defb  @00001000
         defb  @00001000
         defb  @00111110
         defb  @00001000
         defb  @00001000
         defb  @00000000

; character 12         ,

         defb  @00000000
         defb  @00000000
         defb  @00000000
         defb  @00000000
         defb  @00000110
         defb  @00000110
         defb  @00000010
         defb  @00000000

; character 13         -

         defb  @00000000
         defb  @00000000
         defb  @00000000
         defb  @00000000
         defb  @00111100
         defb  @00000000
         defb  @00000000
         defb  @00000000

; character 14         .

         defb  @00000000
         defb  @00000000
         defb  @00000000
         defb  @00000000
         defb  @00000000
         defb  @00000110
         defb  @00000110
         defb  @00000000

; character 15         /

         defb  @00000000
         defb  @00000100
         defb  @00000100
         defb  @00001000
         defb  @00010000
         defb  @00100000
         defb  @00100000
         defb  @00000000

; character 16         0

         defb  @00000000
         defb  @00111100
         defb  @01000110
         defb  @01001010
         defb  @01010010
         defb  @01100010
         defb  @00111100
         defb  @00000000

; character 17         1

         defb  @00000000
         defb  @00000100
         defb  @00001100
         defb  @00000100
         defb  @00000100
         defb  @00000100
         defb  @00001110
         defb  @00000000

; character 18         2

         defb  @00000000
         defb  @00111100
         defb  @01000010
         defb  @00000010
         defb  @00111100
         defb  @01000000
         defb  @01111110
         defb  @00000000

; character 19         3

         defb  @00000000
         defb  @00111100
         defb  @01000010
         defb  @00000100
         defb  @00000010
         defb  @01000010
         defb  @00111100
         defb  @00000000

; character 20         4

         defb  @00000000
         defb  @00000100
         defb  @00001100
         defb  @00010100
         defb  @00100100
         defb  @01111110
         defb  @00000100
         defb  @00000000

; character 21         5

         defb  @00000000
         defb  @01111110
         defb  @01000000
         defb  @01111100
         defb  @00000010
         defb  @01000010
         defb  @00111100
         defb  @00000000

; character 22         6

         defb  @00000000
         defb  @00111100
         defb  @01000000
         defb  @01111100
         defb  @01000010
         defb  @01000010
         defb  @00111100
         defb  @00000000

; character 23         7

         defb  @00000000
         defb  @01111110
         defb  @00000010
         defb  @00000100
         defb  @00001000
         defb  @00010000
         defb  @00100000
         defb  @00000000

; character 24         8

         defb  @00000000
         defb  @00111100
         defb  @01000010
         defb  @00111100
         defb  @01000010
         defb  @01000010
         defb  @00111100
         defb  @00000000

; character 25         9

         defb  @00000000
         defb  @00111100
         defb  @01000010
         defb  @01000010
         defb  @00111110
         defb  @00000010
         defb  @00111100
         defb  @00000000

; character 26         :

         defb  @00000000
         defb  @00000110
         defb  @00000110
         defb  @00000000
         defb  @00000000
         defb  @00000110
         defb  @00000110
         defb  @00000000

; character 27         ;

         defb  @00000000
         defb  @00000110
         defb  @00000110
         defb  @00000000
         defb  @00000110
         defb  @00000110
         defb  @00000010
         defb  @00000000

; character 28         <

         defb  @00000000
         defb  @00000000
         defb  @00000100
         defb  @00001000
         defb  @00010000
         defb  @00001000
         defb  @00000100
         defb  @00000000

; character 29         =

         defb  @00000000
         defb  @00000000
         defb  @00000000
         defb  @00111100
         defb  @00000000
         defb  @00111100
         defb  @00000000
         defb  @00000000

; character 30         >

         defb  @00000000
         defb  @00000000
         defb  @00010000
         defb  @00001000
         defb  @00000100
         defb  @00001000
         defb  @00010000
         defb  @00000000

; character 31         ?

         defb  @00000000
         defb  @00111100
         defb  @01000010
         defb  @00000010
         defb  @00001100
         defb  @00000000
         defb  @00001000
         defb  @00000000

; character 32         @

         defb  @00000000
         defb  @00111100
         defb  @01000010
         defb  @01011100
         defb  @01010010
         defb  @01000100
         defb  @00111110
         defb  @00000000

; character 33         A

         defb  @00000000
         defb  @00111100
         defb  @01000010
         defb  @01000010
         defb  @01111110
         defb  @01000010
         defb  @01000010
         defb  @00000000

; character 34         B

         defb  @00000000
         defb  @01111100
         defb  @01000010
         defb  @01111100
         defb  @01000010
         defb  @01000010
         defb  @01111100
         defb  @00000000

; character 35         C

         defb  @00000000
         defb  @00111100
         defb  @01000010
         defb  @01000000
         defb  @01000000
         defb  @01000010
         defb  @00111100
         defb  @00000000

; character 36         D

         defb  @00000000
         defb  @01111100
         defb  @01000010
         defb  @01000010
         defb  @01000010
         defb  @01000010
         defb  @01111100
         defb  @00000000

; character 37         E

         defb  @00000000
         defb  @01111110
         defb  @01000000
         defb  @01111100
         defb  @01000000
         defb  @01000000
         defb  @01111110
         defb  @00000000

; character 38         F

         defb  @00000000
         defb  @01111110
         defb  @01000000
         defb  @01111100
         defb  @01000000
         defb  @01000000
         defb  @01000000
         defb  @00000000

; character 39         G

         defb  @00000000
         defb  @00111100
         defb  @01000010
         defb  @01000000
         defb  @01000110
         defb  @01000010
         defb  @00111100
         defb  @00000000

; character 40         H

         defb  @00000000
         defb  @01000010
         defb  @01000010
         defb  @01111110
         defb  @01000010
         defb  @01000010
         defb  @01000010
         defb  @00000000

; character 41         I

         defb  @00000000
         defb  @00011100
         defb  @00001000
         defb  @00001000
         defb  @00001000
         defb  @00001000
         defb  @00011100
         defb  @00000000

; character 42         J

         defb  @00000000
         defb  @00000010
         defb  @00000010
         defb  @00000010
         defb  @00000010
         defb  @01000010
         defb  @00111100
         defb  @00000000

; character 43         K

         defb  @00000000
         defb  @01000010
         defb  @01000100
         defb  @01111000
         defb  @01001000
         defb  @01000100
         defb  @01000010
         defb  @00000000

; character 44         L

         defb  @00000000
         defb  @01000000
         defb  @01000000
         defb  @01000000
         defb  @01000000
         defb  @01000000
         defb  @01111110
         defb  @00000000

; character 45         M

         defb  @00000000
         defb  @01000010
         defb  @01100110
         defb  @01011010
         defb  @01000010
         defb  @01000010
         defb  @01000010
         defb  @00000000

; character 46         N

         defb  @00000000
         defb  @01000010
         defb  @01100010
         defb  @01010010
         defb  @01001010
         defb  @01000110
         defb  @01000010
         defb  @00000000

; character 47         O

         defb  @00000000
         defb  @00111100
         defb  @01000010
         defb  @01000010
         defb  @01000010
         defb  @01000010
         defb  @00111100
         defb  @00000000

; character 48         P

         defb  @00000000
         defb  @01111100
         defb  @01000010
         defb  @01000010
         defb  @01111100
         defb  @01000000
         defb  @01000000
         defb  @00000000

; character 49         Q

         defb  @00000000
         defb  @00111100
         defb  @01000010
         defb  @01000010
         defb  @01000010
         defb  @01001010
         defb  @00111110
         defb  @00000000

; character 50         R

         defb  @00000000
         defb  @01111100
         defb  @01000010
         defb  @01000010
         defb  @01111100
         defb  @01000100
         defb  @01000010
         defb  @00000000

; character 51         S

         defb  @00000000
         defb  @00111100
         defb  @01000010
         defb  @00110000
         defb  @00001100
         defb  @01000010
         defb  @00111100
         defb  @00000000

; character 52         T

         defb  @00000000
         defb  @00111110
         defb  @00001000
         defb  @00001000
         defb  @00001000
         defb  @00001000
         defb  @00001000
         defb  @00000000

; character 53         U

         defb  @00000000
         defb  @01000010
         defb  @01000010
         defb  @01000010
         defb  @01000010
         defb  @01000010
         defb  @00111100
         defb  @00000000

; character 54         V

         defb  @00000000
         defb  @01000010
         defb  @01000010
         defb  @01000010
         defb  @01000010
         defb  @00100100
         defb  @00011000
         defb  @00000000

; character 55         W

         defb  @00000000
         defb  @01000010
         defb  @01000010
         defb  @01000010
         defb  @01011010
         defb  @01011010
         defb  @00100100
         defb  @00000000

; character 56         X

         defb  @00000000
         defb  @01000010
         defb  @00100100
         defb  @00011000
         defb  @00011000
         defb  @00100100
         defb  @01000010
         defb  @00000000

; character 57         Y

         defb  @00000000
         defb  @01000010
         defb  @00100100
         defb  @00011000
         defb  @00001000
         defb  @00001000
         defb  @00001000
         defb  @00000000

; character 58         Z

         defb  @00000000
         defb  @01111110
         defb  @00000100
         defb  @00001000
         defb  @00010000
         defb  @00100000
         defb  @01111110
         defb  @00000000

; character 59         [

         defb  @00000000
         defb  @00001100
         defb  @00001000
         defb  @00001000
         defb  @00001000
         defb  @00001000
         defb  @00001100
         defb  @00000000

; character 60         \

         defb  @00000000
         defb  @00100000
         defb  @00100000
         defb  @00010000
         defb  @00001000
         defb  @00000100
         defb  @00000100
         defb  @00000000

; character 61         ]

         defb  @00000000
         defb  @00110000
         defb  @00010000
         defb  @00010000
         defb  @00010000
         defb  @00010000
         defb  @00110000
         defb  @00000000

; character 62         ^

         defb  @00000000
         defb  @00001000
         defb  @00011100
         defb  @00001000
         defb  @00001000
         defb  @00001000
         defb  @00001000
         defb  @00000000

; character 63         _

         defb  @00000000
         defb  @00000000
         defb  @00000000
         defb  @00000000
         defb  @00000000
         defb  @00000000
         defb  @00000000
         defb  @11111111
                                                  
